""" This module converts YASARA Menu declarations to a Tkinter menu bar for PyMol. It  """
# pylint: disable=R0903,W1402,W0122,W0512

import re
import sys
import os
import traceback
from ... import __version__ as VERSION

# The YASARA menu declaration should remain largely static at this point.
# This is declared here separately in keeping with the "No importing from non-shared resources" rule
# for this repo.
YASARA_DECLARATION = """
MainMenu after Effects: 3DM
  PullDownMenu: Structures
    PopUpMenu: Load structure from 3DM
      Request: load_from_3dm
    PopUpMenu: Load model from 3DM
      Request: load_model_from_3dm
    PopUpMenu: Load corresponding compounds
      Request: load_corresponding_compounds
  PullDownMenu: Dataset
    PopUpMenu: Select a 3DM System
      Request: pick_system
    PopUpMenu: Select an alignment
      Request: pick_family
    PopUpMenu: Select a subset
      Request: pick_subset
    PopUpMenu: Select numbering scheme
      Request: renumber_objects
  PullDownMenu: Load 1A28A
    Request: cmd_test_load_1a28a
  PullDownMenu: Hotspot Baskets
    PopUpMenu: Load a Hotspot basket
      Request: load_hotspot_basket
    PopUpMenu: Add residues to a Hotspot basket
      Request: add_residues_to_basket
    PopUpMenu: Create a new Hotspot basket
      Request: create_new_basket
    PopUpMenu: Remove an item from a basket
      Request: delete_residues_from_basket
  PullDownMenu: Show superfamily data
    PopUpMenu: B-factor
      Request: show_superfamily_data_bvalue
    PopUpMenu: Conservation
      Request: show_superfamily_data_topconserved
    PopUpMenu: Correlations
      Request: show_superfamily_data_correlatedmutations
    PopUpMenu: Ligand contacts
      Request: show_superfamily_data_ligandcontacts
    PopUpMenu: Mutations
      Request: show_superfamily_data_mutations
    PopUpMenu: RMSD
      Request: show_superfamily_data_rmsd
    PopUpMenu: External protein-protein interactions
      Request: show_superfamily_data_pdbresiduecontactsexternal
    PopUpMenu: Intra-complex protein-protein interactions
      Request: show_superfamily_data_pdbresiduecontactsinternal
    PopUpMenu: DNA/RNA contacts
      Request: show_superfamily_data_dnarnacontacts
  PullDownMenu: Color molecules by...
    PopUpMenu: Cores
      Request: color_molecules_by_cores
    PopUpMenu: Hydrophobicity
      Request: color_molecules_by_hydrophobicity
    PopUpMenu: Sidechain accessibility
      Request: color_molecules_by_sidechainaccessibility
  PullDownMenu: Literature Hotspots
    Request: show_literature_hotspots

  PullDownMenu: Show scene content details
    Request: show_scene_details
  PullDownMenu: Show amino acid distribution
    Request: show_amino_acid_distribution

  PullDownMenu: Set proxy information
    Request: set_proxy_information

  PullDownMenu: About 3DM
    PopUpMenu: About Bio-Prodict 3DM
      Request: show_about
    PopUpMenu: Go to 3DM PyMol Plugin Help
      Request: cmd_go_to_help_page
    PopUpMenu: Go to 3DM website
      Request: cmd_go_to_3dm_website

AtomContextMenu: 3DM
  PopUpMenu: Add residue to last loaded basket
    Request: add_residue_to_current_basket
  PopUpMenu: Add residue to a specific basket
    Request: add_residue_to_specific_basket

ResidueContextMenu: 3DM
  PopUpMenu: Add residue to last loaded basket
    Request: add_residue_to_current_basket
  PopUpMenu: Add residue to a specific basket
    Request: add_residue_to_specific_basket

SelectionContextMenu: 3DM
  PopUpMenu: Add residue to last loaded basket
    Request: add_residue_to_current_basket
  PopUpMenu: Add residue to a specific basket
    Request: add_residue_to_specific_basket
"""


class Menu(object):
    """Generic menu object."""

    def __init__(self, name, parent=None):
        """Initializes menus with required attributes."""
        self.name = name
        self.parent = parent
        self.children = []
        self.request = None

    def get_var_name(self):
        """Returns the name of this menu as a valid variable name in Python."""
        var_name = self.name.lower().replace(" ", "_")
        # Remove invalid characters
        var_name = re.sub(r"[,.()\[\];\"\'<>?/\\]", "", var_name)
        # Remove invalid numbers at the start of the name (if any)
        var_name = re.sub(r"^[0-9]+", "", var_name)
        return var_name


class MainMenu(Menu):
    """A top level menu."""

    pass


class PullDownMenu(Menu):
    """A Menu that can contain PopUpMenu's."""

    pass


class PopUpMenu(Menu):
    """A menu that is generally used as the last layer."""

    pass


MENU_TYPES = [MainMenu, PullDownMenu, PopUpMenu]
MENU_TYPE_NAMES = [m.__name__ for m in MENU_TYPES]


class CreateGUICode(object):
    """This class contains methods to generate a Tkinter UI for the PyMol 3DM pluginself.
    It can base this UI on a YASARA menu declaration."""

    def __init__(self, declaration):
        """Initializes the UI creator with a YASARA declaration string.
        This is a requirement for now."""
        self.declaration = declaration

    @staticmethod
    def check_indent_level(line_n, line):
        """Find the indent level of the current line in from the YASARA menu declaration."""
        indent_level = 0
        for char in line:
            if char == " ":
                indent_level += 0.5  # Two spaces is one indent level
            else:
                break
        # Check if the indent level is an integer
        assert indent_level % 1 == 0, (
            "The indentation level on line %i is incorrect. " % line_n
        )
        return indent_level

    @staticmethod
    def get_menu_type(line_n, line):
        """Returns the type of menu declared on a line string.
        The line number is required for error handling."""
        try:
            menu_type = line.strip().split(":")[0]
            return menu_type.split(" ")[0]
        except (IndexError, ValueError):
            raise ValueError("Could not retrieve the menu type from line %s." % line_n)

    @staticmethod
    def get_request_function(line_n, line):
        """Returns the type of request function declared on a line string.
        The line number is required for error handling."""
        try:
            return line.split(": ")[1].strip()
        except (IndexError, ValueError):
            raise ValueError(
                "Could not retrieve the request function from line %s." % line_n
            )

    @staticmethod
    def get_menu_name(line_n, line):
        """Returns the menu name declared on a line string.
        The line number is required for error handling."""
        try:
            return line.split(": ")[1].strip()
        except (IndexError, ValueError):
            raise ValueError("Could not retrieve the menu name from line %s." % line_n)

    def parse_declaration(self):
        """Parses the YASARA declaration into a list of Menu objects, each containing a tree
        structure of the underlying Menu elements."""
        menu_type_valid = True
        current_parent = None
        menu = None
        current_indent_level = 0

        menus = []

        for line_n, line in enumerate(self.declaration.split("\n")):
            if len(line.strip()) == 0:
                continue
            indent_level = self.check_indent_level(line_n, line)

            # Skip invalid menu type trees
            if indent_level == 0:
                current_parent, menu_type_valid = None, True
            elif not menu_type_valid:
                continue

            menu_type = self.get_menu_type(line_n, line)

            if menu_type not in MENU_TYPE_NAMES + ["Request"]:
                menu_type_valid = False
                continue

            if menu_type in MENU_TYPE_NAMES:
                if indent_level < current_indent_level:
                    current_parent = current_parent.parent
                    menu.parent = current_parent

                if indent_level > current_indent_level:
                    current_parent = menu

                name = self.get_menu_name(line_n, line)
                menu = MENU_TYPES[MENU_TYPE_NAMES.index(menu_type)](
                    name, current_parent
                )
                menus.append(menu)

                if current_parent is not None:
                    current_parent.children.append(menu)

            if menu_type == "Request":
                menu.request = self.get_request_function(line_n, line)
                continue  # Skip changing indent level

            current_indent_level = indent_level

        return self.get_top_level_menus(menus)

    @staticmethod
    def get_top_level_menus(menu_list):
        """Returns only menus that have no parents."""
        top_level_menus = []
        for menu in menu_list:
            if menu.parent is None:
                top_level_menus.append(menu)
        return top_level_menus

    def show_menu_children(self, menu, level=0):
        """Recursive function that attempts to print all the elements in a menu tree.
        Will raise a RuntimeError if the recursion depth exceeds 10."""
        print(level * "    " + menu.name + "[%s]" % menu.request)
        if level > 10:
            raise RuntimeError("Recursion of the menu is too deep (>10).")
        for child in menu.children:
            self.show_menu_children(child, level + 1)

    def show_menu_tree(self, menu_list):
        """Prints trees for each top level menu in the given list."""
        top_level_menus = self.get_top_level_menus(menu_list)
        for tlmenu in top_level_menus:
            self.show_menu_children(tlmenu)

    @staticmethod
    def generate_frames_code(frame_names):
        """Generates Tkinter code that places a list of frames on top of each other based on the
        frame_names given. The final Frame placed is a default frame with a simple message.
        Returns this code as a list."""
        frames_code = [
            "    global frames",
            "    frames = {}",
            "    container = Frame(root)",
            "    container.pack(side='top', fill='both', expand=True)",
            "    container.grid_rowconfigure(0, weight=1)",
            "    container.grid_columnconfigure(0, weight=1)",
            "",
        ]
        for frame_name in frame_names:
            frame_fname = "frames/%s_gui.py" % frame_name
            frame_fname = os.path.join(os.path.dirname(__file__), frame_fname)
            if os.path.exists(frame_fname):
                frames_code.insert(
                    1,
                    "    from .frames.%s_gui import frame as frame_%s"
                    % (frame_name, frame_name),
                )
                frames_code += [
                    "    frames['%s'] = frame_%s(container, root)"
                    % (frame_name, frame_name),
                    "    frames['%s'].grid(row=0, column=0, sticky='nsew', padx=10, pady=10)"
                    % frame_name,
                    "",
                ]
            else:
                frames_code += [
                    "    frames['%s'] = Frame(container)" % (frame_name),
                    "    Label(frames['%s'], text='This is frame %s.').pack(side='top', anchor='nw')"
                    % (frame_name, frame_name),
                    "    frames['%s'].grid(row=0, column=0, sticky='nsew', padx=10, pady=10)"
                    % frame_name,
                    "",
                ]

        # Default frame last.
        frames_code += [
            "    from .frames.login_gui import frame as login_frame",
            "    default_frame = login_frame(container, root)",
            "    default_frame.grid(row=0, column=0, sticky='nsew', padx=10, pady=10)",
        ]

        return frames_code

    @staticmethod
    def generate_command_code(
        code, command_prefix, expected_frames, child, enable_frame_functions=False
    ):
        """Generates code that adds a command button to a Menu and adds it to the `code` list.

        :param code: The list where the code is added/inserted.
        :param command_prefix: A prefix that is inserted before each function call from the GUI.
        :param expected_frames: A list where the frames created for requests can be added.
        :param child: A child Menu object.
        :param enable_frame_functions: Optional, if this is set to True, the name of the child Menu
                                       is added to the list of frames that need to be drawn and a
                                       function to switch to that frame is added to the code.
        """

        if child.request.startswith("cmd_test_") and not LIVE_TESTING_MODE:
            return

        if child.request.startswith("cmd_"):
            command_prefix = "root.viewer_instance."

        code.append(
            "    threedm_menu.add_command(label='%s', command=%s)"
            % (child.name, command_prefix + child.request)
        )
        if enable_frame_functions and not child.request.startswith("cmd_"):
            code.insert(
                6,
                "\n\ndef %s():\n    global switch_to_frame\n    switch_to_frame('%s')"
                % (command_prefix + child.request, child.request),
            )
            expected_frames.append(child.request)

    @staticmethod
    def generate_pulldown_code(
        code, command_prefix, expected_frames, child, enable_frame_functions=False
    ):
        """Generates code that adds a pulldown menu to a Menu and adds it to the `code` list.

        :param code: The list where the code is added/inserted.
        :param command_prefix: A prefix that is inserted before each function call from the GUI.
        :param expected_frames: A list where the frames created for requests can be added.
        :param child: A child Menu object.
        :param enable_frame_functions: Optional, if this is set to True, the name of the child Menu
                                       is added to the list of frames that need to be drawn and a
                                       function to switch to that frame is added to the code.
        """

        if code[-1] != "":
            code.append("")
        submenu_var = child.get_var_name()
        code.append("    %s = Menu(threedm_menu, tearoff=False)" % submenu_var)
        for subchild in child.children:
            if subchild.request.startswith("cmd_"):
                command_prefix = "root.viewer_instance."

            code.append(
                "    %s.add_command(label='%s', command=%s)"
                % (submenu_var, subchild.name, command_prefix + subchild.request)
            )
            if enable_frame_functions and not subchild.request.startswith("cmd_"):
                code.insert(
                    6,
                    "\n\ndef %s():\n    global switch_to_frame\n    switch_to_frame('%s')"
                    % (command_prefix + subchild.request, subchild.request),
                )
                expected_frames.append(subchild.request)

        code.append(
            "    threedm_menu.add_cascade(label=u'%s', menu=%s)"
            % (child.name, submenu_var)
        )
        code.append("")

    def generate_code(self, menu, enable_frame_functions=False):
        """Generates a fully functional Tkinter script that shows the user a GUI with a menu bar,
        containing the `menu` tree. If enable_frame_functions is added, boilerplate code will be
        generated that adds frames to the GUI and allows the user to switch between these frames.
        """
        window_size = (640, 420)

        code = [
            '""" This code is generated by `plugin.pymol.viewer.CreateGUICode`.',
            "This code has been generated verbosely and plainly to generate the basic GUI for",
            "the 3DM PyMol plugin. Frame code can be found in `plugin.pymol.viewer.frames`.",
            '"""',
            "# pylint: disable=missing-docstring,global-variable-not-assigned,invalid-name,"
            "global-variable-undefined,line-too-long,W0403,C0413,R0903,W0611,W0603,R0914,R0915",
            "try:\n"
            "    try:\n"
            "        from Tkinter import Frame, Tk, Menu, Label, Toplevel, PhotoImage\n"
            "        from ttk import Frame, Label\n"
            "        import tkinter.font as tkFont\n"
            "    except ImportError:\n"
            "        from tkinter import Frame, Tk, Menu, Label, Toplevel, PhotoImage\n"
            "        from tkinter.ttk import Frame, Label\n"
            "        import tkinter.font as tkFont\n"
            "except ImportError:\n"
            "    raise RuntimeError('You need Tkinter to run this GUI.')\n"
            "import os",
            "import platform",
            "from functools import partial",
            "",
            "",
        ]

        code += [
            "def switch_to_frame(frame):",
            "    global frames",
            "    frames[frame].tkraise()",
            "    activate_func = getattr(frames[frame], 'activate', None)",
            "    if activate_func is None:",
            "        print('No activation function for this frame (%s).' % frame)",
            "    else:",
            "        activate_func()",
            "",
            "",
            "def start_gui(parent, check_auto_login=True):",
            "    global frames",
            "    root = Toplevel(parent)",
            "    try:",
            "        root.viewer_instance = parent.viewer_instance",
            "        root.pymol_instance = parent.pymol_instance",
            "    except AttributeError:",
            "        pass",
            "    root.geometry('%sx%s')" % window_size,
            "    menu_bar = Menu(root)",
            "    root.wm_title('Bio-Prodict 3DM Plugin')",
            "    root.switch_to_scene_details = partial(switch_to_frame, 'show_scene_details')",
            "    threedm_menu = Menu(menu_bar, tearoff=False)",
        ]

        # Only add this code to the GUI when we are testing the code by running it directly from
        # the terminal.
        if LIVE_TESTING_MODE:
            code += [
                "    from .PyMolViewer import PyMolViewer as pmv",
                "    root.viewer_instance = pmv()",
                "    from plugin.pymol.test.MockPyMol import PYMOL",
                "    root.pymol_instance = PYMOL",
            ]

        command_prefix = ""

        expected_frames = []

        for child in menu.children:
            if len(child.children) == 0:
                self.generate_command_code(
                    code, command_prefix, expected_frames, child, enable_frame_functions
                )
            else:
                self.generate_pulldown_code(
                    code, command_prefix, expected_frames, child, enable_frame_functions
                )

        code += self.generate_frames_code(expected_frames)
        # MacOS may break when a Unicode character is shown
        code.append("    menu_label = u'3DM Menu +' if platform.system() == 'Darwin' else u'3DM Menu \u25BE'")

        code.append(
            "    menu_bar.add_cascade(label=menu_label, menu=threedm_menu, state='disabled')"
        )
        code.append(
            "    root.enable_menu = lambda: menu_bar.entryconfig(menu_label, state='normal')"
        )
        code.append("\n    root.config(menu=menu_bar)")
        code.append("\n    if check_auto_login:")
        code.append("\n        default_frame.check_auto_login()")
        code.append("\n    return root")
        return "\n".join(code)


def neat_print(code, from_to=None):
    """Prints code nicely with line numbers.

    :param code: The code to be printed (string)
    :param from_to: (Optional) A 2-tuple containing a start and end point for the code to print."""
    if from_to is None:
        print(
            "\n".join(
                [
                    "\033[93m" + str(line_n + 1).rjust(4, " ") + "│\033[0m" + line
                    for line_n, line in enumerate(code.split("\n"))
                ]
            )
        )
    else:
        start, end = from_to
        print(
            "\n".join(
                [
                    "\033[93m" + str(line_n).rjust(4, " ") + "│\033[0m" + line
                    for line_n, line in zip(
                        range(start, end), code.split("\n")[start:end]
                    )
                ]
            )
        )


if __name__ == "__main__":
    # If the LIVE_TESTING_MODE flag is set, make sure the generated code can run directly from the
    # command line, instead of being called by PyMol. This allows for quicker iteration.
    LIVE_TESTING_MODE = "test" in sys.argv
    CONVERTER = CreateGUICode(YASARA_DECLARATION)
    CREATED_MENUS = CONVERTER.parse_declaration()
    GENERATED_CODE = CONVERTER.generate_code(
        CREATED_MENUS[0], enable_frame_functions=True
    )

    if LIVE_TESTING_MODE:
        GENERATED_CODE += "\n\nroot = start_gui(Tk())"
        GENERATED_CODE += "\nroot.mainloop()"

    if "create" in sys.argv:
        # Prints the script plainly
        print(GENERATED_CODE)

    if "inspect" in sys.argv:
        # Prints the script with nicer formatting
        neat_print(GENERATED_CODE)

    if LIVE_TESTING_MODE:
        try:
            exec(GENERATED_CODE, globals(), locals())
        except Exception:  # pylint: disable=W0703
            traceback.print_exc(file=sys.stdout)
            ERROR_LINE = traceback.extract_tb(sys.exc_info()[2])[-1][1]
            neat_print(GENERATED_CODE, (max(1, ERROR_LINE - 5), ERROR_LINE + 5))
