""" This is the the PyMol viewer interaction module. """
# pylint: disable=W0403,W0201,E0401,E0611,W0406,R0904,R0903,R0902,C0103,R0913
import os
import struct
import codecs
import math
import json
import sys
import webbrowser
from collections import defaultdict
from contextlib import contextmanager
from functools import partial

from ...shared.vendors import requests
from ...shared.vendors.six import string_types
from ...shared.api.API import API
from ...shared.api.Baskets import BasketsAPI
from ...shared.api.APIAuth import APIAuth
from ...shared.Config import Config
from ...shared.viewer.Selection import Selection, ObjectIdentifier
from ...shared.viewer.Storage import Storage
from ...shared.viewer.CompositeUtilities import CompositeUtilities
from ... import __version__ as VERSION

try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk

try:
    import pymol as PYMOL
except ImportError:
    from ..test.MockPyMol import PYMOL


class PyMolStorage(Storage):
    """Storage subclass that allows us to check the scene for new objects when an error is raised."""

    def __getattr__(self, name):
        """Defines a custom getattribute functionality that attempts to load scene metadata, if an
        attribute is missing. This means that downloaded scenes are integrated immediately."""
        if PyMolViewer.get_metadata_objects():
            PyMolViewer.load_scene_data(self)
            PyMolViewer.clear_metadata_from_scene()

            return getattr(self, name)
        else:
            raise AttributeError(name)


class PyMolViewer(object):
    """This is the the PyMol viewer interaction module."""

    def __init__(self):
        self.storage = PyMolStorage()
        self.storage.subset = 1
        self.storage.sfamid = None
        self.defined_colors = []  # Colors defined for PyMol
        self.storage.views = []
        self.storage.scene_features = {}
        self.storage.scene_color_scales = {}
        self.storage.scene_lh_articles = {}

        # Extend this object with shared methods
        self.load_accessible_systems = CompositeUtilities.load_accessible_systems
        self.load_pdbs_for_system = partial(
            CompositeUtilities.load_pdbs_for_system, self
        )
        self.load_models_for_system = CompositeUtilities.load_models_for_system
        self.load_model = CompositeUtilities.load_model
        self.load_alignment_families = CompositeUtilities.load_alignment_families
        self.load_alignment_subsets = CompositeUtilities.load_alignment_subsets
        self.get_custom_numbering_schemes = (
            CompositeUtilities.get_custom_numbering_schemes
        )
        self.renumber_pdb = CompositeUtilities.renumber_pdb
        self.get_amino_acid_distribution = (
            CompositeUtilities.get_amino_acid_distribution
        )
        self.convert_residue_number_from_scene_to_align = (
            CompositeUtilities.convert_residue_number_from_scene_to_align
        )

        # Initial PyMOL settings
        PYMOL.cmd.set("label_size", 18)
        self.VERSION = VERSION
        self.HUMAN_COLOR_NAMES = {
            "red": "FF0000",
            "blue": "0000FF",
            "green": "00FF00",
            "magenta": "FF00FF",
            "yellow": "FFFF00",
            "purple": "800080",
            "cyan": "00FFFF",
        }

    def start(self, show_gui=True):
        """Starts the Tk window showing the interface for PyMol."""
        gui_path = os.path.join(os.path.dirname(__file__), "gui.py")
        if os.path.exists(gui_path):
            try:
                from .gui import start_gui
            except RuntimeError:
                print(
                    "You need to run this module on a machine with a display and Tkinter."
                )
                return
            if show_gui:
                if os.name == "nt":
                    root = tk.Tk()
                else:
                    root = PYMOL.plugins.get_tk_root()

                root.viewer_instance = self
                root.pymol_instance = PYMOL

                start_gui(root, check_auto_login="DISABLE_AUTO_LOGIN" not in os.environ)
                if os.name == "nt":
                    root.withdraw()
                    root.mainloop()
        elif show_gui:
            raise RuntimeError("The PyMol GUI was not generated.")

    @staticmethod
    def structure_identifier(name, *args, **kwargs):  # pylint: disable=W0613,R0201
        """This method is used in other viewers to compensate for limited space. We have not yet
        encountered problems with name length in PyMol, as such this function returns any object
        name in full."""
        return name

    def download_pdb_file(self, *args, **kwargs):
        """Used as a redirect for the CompositeUtilities function."""
        return CompositeUtilities.download_pdb_file(self, *args, **kwargs)

    def load_pdb_batch(self, *args, **kwargs):
        """Used as a redirect for the CompositeUtilities function."""
        CompositeUtilities.load_pdb_batch(self, *args, **kwargs)

    def load_pdb(
        self,
        pdbid,
        sfamid=None,
        specifier="chain",
        api_endpoint=None,
        auto_renumber=True,
        *args,
        **kwargs
    ):
        """Downloads a PDB file from the 3DM API and loads it into PyMol.

        :param pdbid: The PDB ID.
        :param sfamid: The superfamily ID.
        :param specifier: The specifier argument for the API. This allows you to filter what parts
                          of the structure are to be retrieved. "chain" is the default, results in
                          the entire chain. "amino" only downloads the protein, without any ligands.
                          "hetatam" only loads ligands. Finally, "full" loads the entire structure
                          contained in the PDB, including nucleotides.
        :param api_endpoint: Optional, you can use this to use a custom api endpoint to load a PDB.
                             This argument should be a method of the API class that is logged in.
                             All other args and kwargs are pased to the api endpoint call.
        :param auto_renumber: Optional, whether or not the PDB should be automatically renumbered
                              to the align numbering scheme. Defaults to True.
        """
        if sfamid is None:
            sfamid = self.get_sfamid()

        filename = self.download_pdb_file(
            pdbid, sfamid, specifier, api_endpoint=api_endpoint, *args, **kwargs
        )
        self.load_pdb_file_into_viewer(
            filename, pdbid, sfamid, specifier, auto_renumber
        )

    def load_pdb_file_into_viewer(
        self, filename, pdbid, sfamid, specifier="chain", auto_renumber=True
    ):
        """Takes a path to a PDB file and loads it into PyMol. Removes the file when it is done.

        :param filename: The path to a PDB file.
        :param pdbid: The PDB ID.
        :param sfamid: The superfamily ID.
        :param specifier: The specifier argument for the API. This allows you to filter what parts
                          of the structure are to be retrieved. "chain" is the default, results in
                          the entire chain. "amino" only downloads the protein, without any ligands.
                          "hetatam" only loads ligands. Finally, "full" loads the entire structure
                          contained in the PDB, including nucleotides.
        :param auto_renumber: Optional, whether or not the PDB should be automatically renumbered
                              to the align numbering scheme. Defaults to True.
        """

        obj_name = pdbid
        if specifier == "hetatm":
            obj_name += "_drug"
        elif specifier == "amino":
            obj_name += "_prot"
        elif specifier == "full":
            obj_name += "_full"

        PYMOL.cmd.load(filename, obj_name)

        if specifier in ["amino", "chain", "full"]:
            # Show the structure with cartoon graphics, similar to YASARA defaults.
            PYMOL.cmd.show("cartoon", "not hetatm in object %s" % obj_name)
            PYMOL.cmd.hide("lines", "not hetatm in object %s" % obj_name)
        if specifier in ["hetatm", "chain", "full"]:
            PYMOL.cmd.show("sticks", "hetatm in object %s" % obj_name)
            PYMOL.cmd.hide("lines", "hetatm in object %s" % obj_name)

        os.remove(filename)

        self.create_main_scene()

        # Remove alternative locations
        PYMOL.cmd.remove(
            "(object %s and (altloc 'B' or altloc 'C' or altloc 'D'))" % obj_name
        )

        try:
            self.storage.numbering_schemes[obj_name] = "pdb"
        except AttributeError:
            self.storage.numbering_schemes = {obj_name: "pdb"}

        if auto_renumber:
            # Renumber PDB to alignment numbering
            try:
                family = CompositeUtilities.get_familyid(self)
            except RuntimeWarning:
                family = 1

            self.renumber_pdb(self, obj_name, sfamid, family, "align")

    def cmd_test_load_1a28a(self):
        """Development function, instantly loads 1A28A. This option should only show up when the
        GUI is launched in debug mode."""
        self.storage.system = "NR_demo_2008"
        self.storage.family = 1
        self.load_pdb("1A28A")

    @staticmethod
    def check_scene_objects():
        """Checks if objects in the scene have 3DM information associated with them.
        Not implemented for PyMol."""
        return False

    @staticmethod
    def object_in_scene(obj):
        """Returns whether or not an object with the given name exists in the scene."""
        return obj in PYMOL.cmd.get_object_list()

    def _get_api(self, api="persistence3dmdata"):
        """Tries to return the API() object by accessing the stored API authentication instance.
        If it does not exist, attempts to log in the user.

        :param api: The Bio-Prodict API one is trying to access. Defaults to `persistence3dmdata`.
        :returns: An authenticated API instance."""
        try:
            self.storage.api_auth
        except AttributeError:
            raise RuntimeError("You are not logged in.")

        return API(api, self.storage.api_auth, user_agent=self.get_version())

    def api_login(self, username, password, twofactor=None):
        """Logs the user into the 3DM API and stores the authentication instance for later.

        :param username: Username for 3DM access.
        :param password: Password for 3DM access.
        """
        try:
            self.storage.api_auth = APIAuth(
                username=username,
                password=password,
                twofactor=twofactor,
                export_auth_on_login=True,
                client_id="pymol-plugin-v2",  # Identify these requests as PyMol instead of YASARA
            )
        except requests.exceptions.SSLError:
            raise RuntimeError(
                "Unable to log in. Your OpenSSL version is not up to date. Please follow the "
                "instructions in the 3DM PyMol installation guide to update the software."
            )

    def color_molecules_by(self, prop, objects=None):
        """Colors molecules according to a property from the superfamily data, ie Cores or
        Hydrophobicity.

        :param prop: String, a property of the superfamily data.
        :param objects: Optional, the objects that need to be colored. Defaults to all.
        """
        try:
            data = CompositeUtilities.download_superfamily_data(self)
        except AttributeError:
            raise RuntimeError(
                "You need to load something from 3DM first, or pick a superfamily."
            )

        if prop == "cores":
            # "Cores isn't really a property, it's actually just the locations where we have 3DM
            # numbers assigned
            position_data = {
                pos["alignmentPosition"]: pos["alignmentPosition"] for pos in data
            }
        else:
            position_data = {pos["alignmentPosition"]: pos[prop] for pos in data}

        positions = list(position_data.keys())
        values = list(position_data.values())
        (
            selectors,
            renumbered_residues,
            new_properties,
        ) = CompositeUtilities.convert_residues_to_selections(self, positions, values)

        del renumbered_residues

        if objects is not None:
            selectors = [sel for sel in selectors if sel.in_object in objects]

        if prop == "cores":
            with self.new_view(prop):
                PYMOL.cmd.color("gray")  # Color everything grey.
                CompositeUtilities.color_residues_by_property(
                    self, selectors, "red", new_properties, loop=True
                )
        else:
            with self.new_view(prop):
                PYMOL.cmd.color("gray")  # Color everything grey.
                CompositeUtilities.color_residues_by_property(
                    self, selectors, "blue", new_properties, start_color="green"
                )

    def show_superfamily_data(self, prop, maxvalue, cutoff, objects=None):
        """Shows the superfamily data specified by prop by utilizing the positionInfo API. Works
        in a new View.

        :param prop: Property name from the positionInfos API.
        :param maxvalue: The maximum amount of values to show (after sorting)
        :param cutoff:  The minimum cutoff value for the property.
        """
        try:
            self.get_sfamid()
        except AttributeError:
            raise RuntimeError(
                "You need to load something from 3DM first, or pick a superfamily."
            )

        data = CompositeUtilities.download_superfamily_data(self)

        position_data = CompositeUtilities.filter_position_data(
            data, prop, maxvalue=maxvalue, cutoff=cutoff
        )
        positions, values = zip(
            *sorted(position_data.items(), key=lambda x: (x[1], [0]), reverse=True)
        )

        (
            selectors,
            renumbered_residues,
            new_properties,
        ) = CompositeUtilities.convert_residues_to_selections(
            self, positions, values, objects=objects
        )

        if objects is not None:
            selectors = [sel for sel in selectors if sel.in_object in objects]

        with self.new_view(prop.capitalize()) as scene_name:
            PYMOL.cmd.color("gray")  # Color everything grey.
            self.show_and_label_residues(selectors, renumbered_residues)
            color_scale = CompositeUtilities.color_residues_by_property(
                self, selectors, "blue", new_properties, start_color="green"
            )

        self.storage.scene_features[scene_name] = dict(zip(selectors, new_properties))
        self.storage.scene_color_scales[scene_name] = color_scale

    @staticmethod
    def show_and_label_residues(selectors, residue_numbers):
        """Adds a label and shows the residue of each of the given Selection objects. The label
        will correspond to the first integer in the selection. If the label does not contain an int,
        it will be skipped.

        :param residue_positions: A list of YASARA selections."""

        # Hide all previously shown residues
        PYMOL.cmd.hide("sticks", Selection(in_protein=True).pymol())
        for selector, number in zip(selectors, residue_numbers):
            # Get the first integer from the string.
            PYMOL.cmd.show("sticks", selector.pymol())  # Shows the residue atoms

            # Enables the C-alpha selection for labeling, so we only get one label.
            selector.only_ca = True
            # The label name is actually a "Python expression that can be converted to a string."
            # In this case we always want the literal meaning.
            PYMOL.cmd.label(selector.pymol(), "'%s'" % number)

    def clear_numbering_sensitive_views(self, continue_response=None):
        """This method allows the user to remove any views that became invalid after renumbering.
        Some views are sensitive to numbering; they have labels applied containing an old numbering
        scheme. Because view changes cannot be detected, these labels may still be presented to the
        user when switching to those views. This gives the impression of incorrect or stale data.

        In order to prevent this disonance, this method asks the user to choose between removing
        the affected views, or to forego the renumbering operation. If the user elects to remove
        the views, they are removed by this function and True is returned. If the user prefers not
        to remove these views, the function returns False. If there are no known views that would
        be affected by the renumbering, True is returned automatically.

        :param continue_response: Optional, by default, the user is asked if they want to continue
                                  renumbering, even if it means losing current visualizations. If
                                  True or False is passed here, this dialog will be skipped and this
                                  response taken.
        :returns: True if there are no more invalid views and False if the user opted not to
                  remote invalid views."""
        if len(self.storage.views) > 0 and len(self.storage.scene_features) > 0:
            if continue_response is None:
                import tkMessageBox  # Only import this if necessary

                continue_response = tkMessageBox.askyesno(
                    "Continue renumbering?",
                    "You are currently viewing some visualizations that will be invalid after "
                    "renumbering. As such, these views will be removed if you choose to continue "
                    "with the renumbering procedure. Do you want to continue?",
                )

            if not continue_response:
                return False

            for view in self.storage.scene_features:
                PYMOL.cmd.scene(view, "delete")

            self.storage.views = []
            self.storage.scene_features = {}
            self.storage.scene_color_scales = {}
            PYMOL.cmd.scene("main")  # Return to the main scene.
            return True
        else:
            return True

    def _color_residue(self, residue, color):
        """Colors a residue.

        :param residue: Residue selector.
        :param color: A color in hex format."""

        if color in self.HUMAN_COLOR_NAMES:
            color = self.HUMAN_COLOR_NAMES[color]

        if len(color) != 6:
            raise ValueError(
                "Invalid color name '%s'. Please try the hexadecimal equivalent or one of the predefined colors: %s."
                % (color, list(self.HUMAN_COLOR_NAMES.keys()))
            )

        color_name = "#" + color.strip("#")
        if color_name not in self.defined_colors:
            # Convert hex color to RGB
            rgb = struct.unpack("BBB", codecs.getdecoder("hex_codec")(color)[0])
            PYMOL.cmd.set_color(color_name, rgb)
            self.defined_colors.append(color_name)

        residue.only_ca = False
        PYMOL.cmd.color(color_name, residue.pymol())

    # Methods for renumbering
    @staticmethod
    def get_stable_numbering_selections(obj):
        """Gets stable numbering Selections for renumbering purposes, numbering appears in the
        order that every residue is connected."""
        space = {"residues": []}
        full_selection = Selection(in_protein=True, in_object=obj, only_ca=True).pymol()

        PYMOL.cmd.iterate(full_selection, "residues.append(index)", space=space)

        visited = []
        indexes_in_obj = space["residues"]
        try:
            current_idx = indexes_in_obj[0]
        except IndexError:
            # No residues in object
            return []

        while set(indexes_in_obj).symmetric_difference(visited):
            visited.append(current_idx)
            space = {"adjacent_residues": []}
            PYMOL.cmd.iterate(
                "((name CA within 4.5 of (index %s and object %s)) and object %s)"
                % (current_idx, obj, obj),
                "adjacent_residues.append(index)",
                space=space,
            )
            for idx in space["adjacent_residues"]:
                if idx not in visited:
                    current_idx = idx
                    break
            else:
                try:
                    # Could not continue, pick up at the lowest unvisited point.
                    current_idx = sorted(list(set(indexes_in_obj).difference(visited)))[
                        0
                    ]
                except IndexError:
                    break

        # Use indexes for stability
        selections = []
        for idx in visited:
            # byres includes the entire residue where an atom (adressed by index) residues.
            sel = "(byres (index %s and object %s))" % (idx, obj)
            selections.append(sel)

        return selections

    def renumber_object_to_resnum(self, objects=None, negative=False):
        """This method will iterate over all the residues in the specified objects (defaults to
        all) and renumber them to residue numbering.

        :param objects: A list of ObjectIdentifiers where each residue will be catalogued and renumbered.
        :param negative: If set to True, renumbering will be negative (-1--N). Defaults to false."""

        if isinstance(objects, ObjectIdentifier):
            objects = [objects]

        selections = {}
        for obj in PYMOL.cmd.get_object_list():
            if objects is not None and obj not in objects:
                continue

            selections[obj] = self.get_stable_numbering_selections(obj)

        prev_index = 0
        for obj, selection_list in selections.items():
            for i, selection in enumerate(selection_list):
                n_i = i + 1
                index = int(str(selection).split(" ")[2])

                if abs(index - prev_index) < 4 and prev_index > 0:
                    raise RuntimeError(
                        "Renumbering error! %i %i %i"
                        % (index, prev_index, index - prev_index)
                    )

                prev_index = index
                if negative:
                    n_i *= -1
                self.renumber_residue(selection, n_i)

    @staticmethod
    def renumber_residue(_from, _to, extension=None):
        """Renumbers a single residue.

        :param _from: Either a Selection object (which will be used directly) or a residue number
        (which will be inserted into a PyMol selector)
        :param _to: The target number for the given residue.
        :param extension: Optional, adds a letter to the target number.
        """
        if extension is not None:
            _to = str(_to) + extension

        if isinstance(_from, Selection):
            selector = _from.pymol()
        elif isinstance(_from, string_types):
            selector = _from
        else:
            selector = Selection(numbers=_from).pymol()

        PYMOL.cmd.alter(selector, "resi='%s'" % _to)

    @staticmethod
    def unselect():
        """Discards the current selection."""
        PYMOL.cmd.delete("sele")

    @staticmethod
    def select_residue(selection):
        """Selects residues according to a Selection object."""
        PYMOL.cmd.select(selection.pymol())

    @staticmethod
    def group_residue(selection, group_name):
        """Groups residues into a selection according to a Selection object."""
        try:
            # First, try to add to the residues to the group, otherwise, create the group.
            PYMOL.cmd.select(group_name, group_name + " + " + selection.pymol())
        except PYMOL.CmdException:
            PYMOL.cmd.select(group_name, selection.pymol())

    @staticmethod
    def reverse_number_outside_group(group_name, pdbid):
        """Renumbers every number outside `group_name` in `pdbid` to a negative residue number."""
        selection = Selection(in_object=pdbid, not_in_group=group_name)
        PYMOL.cmd.alter(selection.pymol(), "resi=str(int(resi)*-1)")
        PYMOL.cmd.delete("Renumbered")

    # Methods for storing Scene data

    @staticmethod
    def create_metadata_object(num, data):
        """Create a new medata pseudoatom and store data in its title.

        :param num: Integer, numerical suffix. Make sure that, if more than one metadata object is
                    created, these suffixes are in the proper order, otherwise reading the data will
                    fail.
        :param data: The data to be stored inside the object title. Must be less than 255 bytes.
        """
        pseudoatom_name = "3dm_metadata_%i" % int(num)
        PYMOL.cmd.pseudoatom(pseudoatom_name)
        assert len(data) < 255
        PYMOL.cmd.set_title(pseudoatom_name, 0, data)

    @staticmethod
    def get_data_from_metadata_object(name):
        """Tries to get data from a metadata object."""
        data = PYMOL.cmd.get_title(name, 0)
        assert len(data) > 0
        return data

    def store_scene_data(self):
        """Stores all the data currently in storage inside the scene."""
        storage = self.storage.as_json()
        storage_length = len(storage)

        chunk_size = 254  # Maximum title storage size
        total_chunks = int(math.ceil(storage_length / float(chunk_size)))
        for i in range(0, total_chunks):
            start, end = i * chunk_size, (i + 1) * chunk_size
            data_chunk = storage[start:end]
            self.create_metadata_object(i, data_chunk)

    @staticmethod
    def retrieve_data_from_scene():
        """Finds metadata objects and combines their data into the original JSON string."""
        metadata_objects = PyMolViewer.get_metadata_objects()
        data = {}
        for obj in metadata_objects:
            data[obj] = PyMolViewer.get_data_from_metadata_object(obj)

        data = sorted(data.items(), key=lambda x: x[0])
        data = [v[1] for v in data]

        return "".join(data)

    @staticmethod
    def load_scene_data(storage):
        """Loads all the data from inside the scene and stores them into the Storage object."""
        json_data = PyMolViewer.retrieve_data_from_scene()
        storage.load_from_json(json_data)

    @staticmethod
    def clear_metadata_from_scene():
        """Removes all metadata storage objects from the scene."""
        metadata_objects = PyMolViewer.get_metadata_objects()
        for obj in metadata_objects:
            PYMOL.cmd.delete(obj)

    @staticmethod
    def get_metadata_objects():
        """Returns a list of object names that correspond to our metadata object name format."""
        objects = PYMOL.cmd.get_object_list()
        if objects is None:
            return []
        return [obj for obj in objects if obj.startswith("3dm_metadata_")]

    # Logging in with a PyMol scene
    def login_with_scene(self, filename):
        """Attempts to authenticate the user by means of a generated PyMol scene/session."""
        if len(PYMOL.cmd.get_object_list()) > 0:
            raise RuntimeError("Please make sure there are no loaded objects.")

        PYMOL.cmd.load(filename, format="pse")

        data = self.retrieve_data_from_scene()
        try:
            auth_details = json.loads(data)["auth_details"]
        except (ValueError, KeyError):
            raise RuntimeError(
                "The session you downloaded does not contain proper authentication information. "
                "Please try downloading and logging in again."
            )

        self.storage.api_auth = APIAuth()
        self.storage.api_auth.connect_with_token(
            access_token=auth_details["access_token"],
            token_expiry=3600,
            refresh_token=auth_details["refresh_token"],
        )
        self.storage.api_auth.export_api_auth()
        self.clear_metadata_from_scene()
        return True

    # Literature hotspots
    def show_literature_hotspots(self, keywords, cutoff=2, limit=25):
        """Shows literature hotspots to the user."""
        mutations_index, articles = CompositeUtilities.get_literature_hotspot_data(
            self, keywords, cutoff=cutoff, limit=limit
        )

        try:
            positions, values = zip(*mutations_index)
        except ValueError:
            raise ValueError("Keywords '%s' did not yield any results." % keywords)

        with self.new_view("Literature Hotspots (%s)" % keywords) as scene_name:
            (
                selectors,
                renumbered_residues,
                new_properties,
            ) = CompositeUtilities.convert_residues_to_selections(
                self, positions, values
            )

            renumbered_dict = dict(zip(positions, renumbered_residues))
            for article in articles:
                try:
                    article["position"] = renumbered_dict[article["position"]]
                except KeyError:
                    continue

            PYMOL.cmd.color("gray")
            self.show_and_label_residues(selectors, renumbered_residues)
            color_scale = CompositeUtilities.color_residues_by_property(
                self, selectors, "blue", new_properties, start_color="green"
            )

            self.storage.scene_features[scene_name] = dict(
                zip(selectors, new_properties)
            )
            self.storage.scene_color_scales[scene_name] = color_scale
            self.storage.scene_lh_articles[scene_name] = articles

    def load_article_abstracts_for_position(self, position, scene_name):
        """This method takes a position and a scene_name and uses them to retrieve stored articles,
        these are then combined and opened in the machine's default webbrowser. The user can use the
        NCBI interface to read abstracts and search through the documents.

        :param position: A residue position that is contained in the current numbering scheme.
        :param scene_name: The name of the scene containing the proper articles of the literature
        hotspots."""
        base_url = "https://www.ncbi.nlm.nih.gov/pubmed/"
        valid_ids = []
        for article in self.storage.scene_lh_articles[scene_name]:
            if str(article["position"]) == str(position):
                valid_ids.append(article["pubmed_id"])
            elif "." in str(article["position"]):  # Deal with GPCR numbering
                article_position = "%s%s" % CompositeUtilities.convert_gpcr_number(
                    article["position"]
                )
                if article_position == str(position):
                    valid_ids.append(article["pubmed_id"])

        if len(valid_ids) == 0:
            raise ValueError("No articles found for position '%s'." % position)
        if len(valid_ids) > 1:
            ids = ",".join(list(set(valid_ids)))
        else:
            ids = valid_ids[0]
        content = "?report=abstract"
        url = base_url + ids + content
        if "nosetests" in sys.argv[0]:
            # The pipe at the end of the error string does not need to be removed; it is the
            # separator between the error message and the URL.
            raise RuntimeError("Cannot open a browser during testing. |" + url)
        else:
            with self.clean_ld_lib_path():
                webbrowser.open(url)

    # Baskets
    def _get_baskets_api(self):
        """Returns a Baskets API interaction object."""
        try:
            self.storage.api_auth
        except AttributeError:
            raise RuntimeError("You are not logged in.")

        return BasketsAPI(self.storage.api_auth, user_agent=self.get_version())

    def create_new_basket(self, name):
        """Creates a new basket using user input."""
        basket_api = self._get_baskets_api()
        sfamid = self.get_sfamid()
        familyid = CompositeUtilities.get_familyid(self)

        basket = basket_api.new_basket(name, sfamid=sfamid, familyid=familyid)
        return basket

    def show_basket_positions(self, positions, annotations, basket_name):
        """Shows positions from a basket inside the scene and adds the positions to the scene
        details viewer.

        :param positions: A list of alignment positions from a basket.
        :param annotations: A list of annotations (descriptions) of the same length as the positions
                            list.
        :param basket_name: The name of the basket.
        """
        values = [1] * len(positions)
        (
            selectors,
            renumbered_residues,
            new_properties,
        ) = CompositeUtilities.convert_residues_to_selections(self, positions, values)

        with self.new_view(basket_name) as scene_name:
            PYMOL.cmd.color("gray")  # Color everything grey.
            self.show_and_label_residues(selectors, renumbered_residues)
            CompositeUtilities.color_residues_by_property(
                self, selectors, "green", new_properties, start_color="blue"
            )

        self.storage.scene_features[scene_name] = dict(zip(selectors, annotations))

    def delete_basket(self, basket_id):
        """Allows the user to select a basket, that will then be deleted."""
        basket_api = self._get_baskets_api()
        basket_api.delete_basket_native(basket_id)

    def delete_items_from_basket(self, positions, basket_id):
        """Deletes a list of positions from a basket.

        :param positions: A list of positions inside the basket.
        :param basket_id: The ID of the basket where the items need to be removed.
        """

        basket_api = self._get_baskets_api()
        basket = basket_api.load_basket(basket_id)
        for position in positions:
            basket.remove_item(where={"value": position})
        basket.save()

    def add_residues_to_basket(self, pdb, residues, description, basket):
        """Adds a list of residues from an object into a basket and saves the basket.

        :param pdb: The name of the object where the residues are selected.
        :param residues: A list of residue numbers from the scene.
        :param description: A string describing the significance of the residues.
        :param basket: A Basket object where the residues need to be added.
        """

        for residue in residues:
            new_number = CompositeUtilities.convert_residue_number_from_scene_to_align(
                self, pdb, residue
            )
            # This negativity test will handle strings and numbers
            if str(new_number)[0] == "-":
                raise RuntimeError(
                    "Could not add position %s to basket, this position exists outside of an "
                    "aligned core. Canceling all additions to this basket." % new_number
                )
            basket.add_residue_position(position=new_number, comment=description)

        self.show_and_label_residues(
            [Selection(numbers=res) for res in residues], residues
        )

        basket.save()
        if len(residues) == 1:
            return "Position %s sucessfully added to basket %s" % (
                residues[0],
                basket.id,
            )
        else:
            return "Positions sucessfully added to basket %s" % basket.id

    # Utilities
    @staticmethod
    def get_selected_residues():
        """This method can be called to retrieve the currently selected residue. This supports
        both singular and multi-selections using the Select dialog or selection tools.

        Returns a list of dictionaries containing the residue object from the pymol selection, the
        residue number and the object (PDB) this residue belongs to."""
        selected_residues_properties = {"numbers": defaultdict(list)}

        for obj in PYMOL.cmd.get_object_list():
            PYMOL.cmd.iterate(
                Selection(in_selected=True, only_ca=True).pymol(),
                "numbers['%s'].append(resi)" % obj,
                space=selected_residues_properties,
            )

        position_list = {}
        for obj, positions in selected_residues_properties["numbers"].items():
            for pos in positions:
                position_list[pos] = {
                    "residue": None,
                    "number": pos,
                    "obj": obj,
                }

        return list(position_list.values())

    @staticmethod
    def rename_object(oldname, newname):
        """Renames an object."""
        PYMOL.cmd.set_name(oldname, newname)

    def create_main_scene(self):
        """Creates/maintains a Main scene and stores it."""
        PYMOL.cmd.scene("main", "store", view=False)
        if "Main" not in self.storage.views:
            self.storage.views.append("Main")

    def get_sfamid(self):
        """Wraps around the CompositeUtilities function in order to provide auto scene loading
        functionality."""
        try:
            return CompositeUtilities.get_sfamid(self)
        except Exception as err:  # pylint: disable=W0703
            if not self.get_metadata_objects:
                raise err

            self.create_main_scene()
            self.load_scene_data(self.storage)
            self.clear_metadata_from_scene()

            return CompositeUtilities.get_sfamid(self)

    def _check_user_access(self):
        """Checks if the user is logged in, returns the log in state (boolean)."""
        return hasattr(self.storage, "api_auth") and self.storage.api_auth is not None

    @staticmethod
    @contextmanager
    def clean_ld_lib_path():
        """Makes sure the LD_LIBRARY_PATH is a "clean" path, without PyMol additions. Restores the
        old PyMol-tainted LD_LIBRARY_PATH on exit."""
        current_ld_lib_path = os.getenv("LD_LIBRARY_PATH", "")
        split_paths = current_ld_lib_path.split(":")
        clean_split_paths = [path for path in split_paths if "pymol" not in path]
        clean_ld_lib_path = ":".join(clean_split_paths)

        os.environ["LD_LIBRARY_PATH"] = clean_ld_lib_path
        yield

        # Upon exit, restore the old LD_LIBRARY_PATH
        os.environ["LD_LIBRARY_PATH"] = current_ld_lib_path

    @contextmanager
    def new_view(self, scene_name="auto"):
        """This scene code attempts to mimic the view functionality from YASARA. Use this as a
        context manager when running a method that needs to happen in a new view.

            with new_view():
                color_residues(...)

        Every attribute is stored inside the scene except the specific rotation of the protein. This
        allows the user to switch between different labels and colors, while maintaining the same
        rotation.
        """

        if scene_name != "auto":
            scene_name = CompositeUtilities.get_valid_view_name(self, scene_name)

        self.storage.views.append(scene_name)
        # Switch to the main scene before hand, this should yield a vanilla scene.
        PYMOL.cmd.scene("main")
        PYMOL.cmd.scene(scene_name, "store", view=False)
        yield scene_name  # Yield the validated scene_name, wait for the wrapped function to exit.

        # Update the scene with the current information
        PYMOL.cmd.scene(scene_name, "update", view=False)

    def cmd_go_to_3dm_website(self):
        """Uses the default webbrowser to open the 3DM Website."""
        with self.clean_ld_lib_path():
            webbrowser.open(Config().website_base_url)  # pylint: disable=E1101

    def cmd_go_to_help_page(self):
        """Uses the default webbrowser to open the PyMol help page."""
        with self.clean_ld_lib_path():
            # This is a public URL specifically for PyMol
            webbrowser.open(
                "https://plugin.bio-prodict.nl/pymol/pymol_installation.html"
            )

    @staticmethod
    def get_version():
        """Returns the version of the plugin and other dependencies."""
        return "bioprodict-pymol-plugin/%s pymol/%s python-requests/%s" % (
            VERSION,
            PYMOL.cmd.get_version(),
            requests.__version__,
        )
