""" Provides a GUI that allows the user to add residues to a basket. """
# pylint: disable=C0103,R0901,R0902,R0914,W0403,W0212,W0201,W0613
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
from ....shared.viewer.CompositeUtilities import CompositeUtilities


class AddResiduesToHotspotBasketGUI(tk.Frame):
    """Provides a GUI that allows the user to add residues to a basket."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.root = root
        self.selected_residues_list = None
        self.selected_basket = None
        self.fill_frame()

    def activate(self):
        """Loads available baskets when the frame is activated."""
        self.load_available_baskets()
        self.fill_selected_residue_listbox()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""

        description_entry = tk.Entry(self)
        description_entry.insert(0, "Description")
        description_entry.pack(side="top", anchor="nw", fill=tk.X)
        self.description_entry = description_entry

        # List type indicator
        header_frame = tk.Frame(self)

        tk.Label(header_frame, text="Baskets").pack(side="left", anchor="nw")
        tk.Label(header_frame, text="Currently selected residues").pack(
            side="right", anchor="ne"
        )

        header_frame.pack(side="top", anchor="nw", fill=tk.X, pady=3)

        two_lists_frame = tk.Frame(self)
        LISTBOX_HEIGHT = 18
        LISTBOX_WIDTH = 36

        # Basket List box with scrollbar
        basket_list_frame = tk.Frame(two_lists_frame)
        basket_list_scrollbar = tk.Scrollbar(basket_list_frame, orient=tk.VERTICAL)
        basket_listbox = tk.Listbox(
            basket_list_frame,
            yscrollcommand=basket_list_scrollbar.set,
            height=LISTBOX_HEIGHT,
            width=LISTBOX_WIDTH,
            selectmode=tk.SINGLE,
        )
        self.basket_listbox = basket_listbox

        self.basket_list = ["No baskets available."]  # Placeholder message

        basket_list_scrollbar.config(command=basket_listbox.yview)
        basket_list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        basket_listbox.pack(side="top", anchor="nw", fill=tk.X, expand=True)

        basket_list_frame.pack(side="left", anchor="nw", fill=tk.X)
        basket_listbox.bind("<<ListboxSelect>>", self.select_basket)

        # Item List box with scrollbar
        selected_residue_list_frame = tk.Frame(two_lists_frame)
        selected_residue_list_scrollbar = tk.Scrollbar(
            selected_residue_list_frame, orient=tk.VERTICAL
        )

        selected_residue_listbox = tk.Listbox(
            selected_residue_list_frame,
            yscrollcommand=selected_residue_list_scrollbar.set,
            height=LISTBOX_HEIGHT,
            width=LISTBOX_WIDTH,
            state=tk.DISABLED,
        )

        self.selected_residue_listbox = selected_residue_listbox
        self.selected_residues_list = [
            "No selected residues available."
        ]  # Placeholder message

        selected_residue_list_scrollbar.config(command=basket_listbox.yview)
        selected_residue_list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        selected_residue_listbox.pack(side="top", anchor="nw", fill=tk.X, expand=True)

        selected_residue_list_frame.pack(side="right", anchor="ne", fill=tk.X)

        two_lists_frame.pack(side="top", anchor="nw", fill=tk.X)

        # Button bar
        button_bar_frame = tk.Frame(self)

        refresh_selection_button = tk.Button(
            button_bar_frame,
            text="Refresh selection",
            command=self.fill_selected_residue_listbox,
        )
        refresh_selection_button.pack(side="left", anchor="nw", pady=10)

        refresh_hotspots_button = tk.Button(
            button_bar_frame,
            text="Refresh hotspot baskets",
            command=self.refresh_baskets,
        )
        refresh_hotspots_button.pack(side="left", anchor="nw", pady=10, padx=10)

        add_to_basket_button = tk.Button(
            button_bar_frame, text="Add residues to basket", command=self.add_to_basket
        )
        add_to_basket_button.pack(side="right", anchor="nw", pady=10)

        button_bar_frame.pack(side="top", anchor="nw", fill=tk.X)

    def select_basket(self, event=None):  # pylint: disable=W0613
        """Takes the basket that is currently selected inside the baskets list and based on this,
        sets the self.selected_basket value to a Basket object with the corresponding ID.

        Returns True if the selection action was sucecesful or False if no basket was selected. """
        selection = self.basket_listbox.curselection()
        if len(selection) == 0:
            self.selected_basket = None
            return False
        basket_id = self.basket_ids[selection[0]]
        baskets_api = self.root.viewer_instance._get_baskets_api()
        self.selected_basket = baskets_api.load_basket(basket_id)
        return True

    def refresh_baskets(self, event=None):
        """Refreshes the list of baskets."""
        self.load_available_baskets()
        self.fill_basket_listbox()

    def fill_basket_listbox(self):
        """Fills the basket listbox with baskets listed in self.basket_list."""
        self.basket_listbox.delete(0, tk.END)
        for basket in self.basket_list:
            self.basket_listbox.insert(tk.END, basket)

    def fill_selected_residue_listbox(self):
        """Fills the selected_residues listbox with selected residues from the PyMol viewer."""
        # You can't alter the contents of a disabled element
        self.selected_residue_listbox.config(state=tk.NORMAL)
        self.selected_residue_listbox.delete(0, tk.END)
        selected_residues_list = self.root.viewer_instance.get_selected_residues()
        print(selected_residues_list)
        selected_residues_list = [
            "%s: %s" % (r["obj"], r["number"]) for r in selected_residues_list
        ]
        print("~", selected_residues_list)
        if len(selected_residues_list) == 0:
            selected_residues_list = [
                "No residues selected or selection is out of date."
            ]
        for selected_residue in selected_residues_list:
            self.selected_residue_listbox.insert(tk.END, selected_residue)
        self.selected_residue_listbox.config(
            state=tk.DISABLED
        )  # Restore disabled state

        self.selected_residues_list = selected_residues_list

    def load_available_baskets(self):
        """Loads all the baskets that are available in the current situation and assigns them to
        the self.basket_ids and self.basket_list lists."""

        baskets = CompositeUtilities.load_available_hotspot_baskets(
            self.root.viewer_instance
        )

        self.basket_ids = [basket.id for basket in baskets]
        self.basket_list = [basket.label for basket in baskets]
        self.fill_basket_listbox()

    def add_to_basket(self):
        """Adds the residues that are currently selected to the basket that is currently selected.
        Will also pull the description from the description entry, if one has been filled in."""
        if (
            self.selected_residues_list is None
            or "No residues" in self.selected_residues_list[0]
        ):
            raise ValueError("No residues selected.")

        if self.selected_basket is None:
            raise ValueError("No basket selected.")

        description = self.description_entry.get().strip()
        if description in ["", "Description"]:
            description = ""

        objects, numbers = zip(
            *[sel.split(": ", 1) for sel in self.selected_residues_list]
        )
        numbers = [n.strip() for n in numbers]
        object_name = objects[0]  # We can only handle one object.

        try:
            self.root.viewer_instance.add_residues_to_basket(
                object_name, numbers, description, self.selected_basket
            )
        except AssertionError as err:
            if "duplicate key value" in str(err):
                raise ValueError("A selected residue already exists in this basket.")
            else:
                raise err


frame = AddResiduesToHotspotBasketGUI
