""" Provides a GUI Frame for coloring residues by properties, like hydrophobicity and sidechain
accessibility. """
# pylint: disable=C0103,W0201,W0231
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox


class ColorResiduesByPropertyGUIFactory(tk.Frame):
    """A Factory for GUI frames, used for coloring residues by properties."""

    def __init__(self, property_name, nice_name):
        self.property_name = property_name
        self.nice_name = nice_name

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirity the first time it is opened."""
        if len(self.winfo_children()) < 2:
            self.fill_frame()
        self.get_objects_in_scene()
        self.fill_listbox()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        text = (
            "This functionality will color objects based on their %s." % self.nice_name
        )
        tk.Label(self, text=text).pack(side="top", anchor="nw", fill=tk.X)
        tk.Label(
            self,
            text="Select objects from the list below to color them by their %s."
            % self.nice_name.lower(),
        ).pack(side="top", anchor="nw", fill=tk.X)

        # List box with scrollbar containing objects
        object_list_frame = tk.Frame(self)
        list_scrollbar = tk.Scrollbar(object_list_frame, orient=tk.VERTICAL)
        listbox = tk.Listbox(
            object_list_frame,
            yscrollcommand=list_scrollbar.set,
            height=18,
            selectmode=tk.EXTENDED,
        )
        self.object_list = [
            "There are no objects in the current scene"
        ]  # Placeholder message

        list_scrollbar.config(command=listbox.yview)
        list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        listbox.pack(side="top", anchor="nw", fill=tk.X)

        object_list_frame.pack(side="top", anchor="nw", fill=tk.X)
        self.listbox = listbox

        # Submit button
        tk.Button(
            self,
            text="Color",
            command=self.color_selected_objects,
        ).pack(side="top", anchor="ne", pady=10)

    def color_selected_objects(self):
        """Colors the selected objects by the selected property name."""

        selected_objects = [
            self.listbox.get(idx) for idx in self.listbox.curselection()
        ]
        if len(selected_objects) == 0:
            selected_objects = None
        try:
            self.root.viewer_instance.color_molecules_by(
                self.property_name, objects=selected_objects
            )
        except RuntimeError as err:
            tkMessageBox.showerror("RuntimeError", str(err))
        except ValueError as err:
            error = str(err)
            if "no numbering scheme was returned" in error:
                obj = error.split("(")[1].split(")")[
                    0
                ]  # Extract the PDB from the error message
                tkMessageBox.showwarning(
                    "Could not renumber %s" % obj,
                    "Warning: Could not renumber %s. This object cannot be found in the current "
                    "system. Please remove this structure to avoid further conflicts."
                    % obj,
                )
            else:
                tkMessageBox.showerror("ValueError", str(err))

    def get_objects_in_scene(self):
        """Sets the object list to a list of all the names of objects in the scene."""
        self.object_list = [
            obj
            for obj in self.root.pymol_instance.cmd.get_object_list()
            if not obj.startswith("3dm_meta")
        ]

    def fill_listbox(self):
        """Fills the listbox with Object ids."""
        self.listbox.delete(0, tk.END)
        for obj in self.object_list:
            self.listbox.insert(tk.END, obj)

    def __call__(self, parent, root, *args, **kwargs):
        """This method takes the place of the normal __init__ method. This way the GUI can be
        created with some small variations in features, while still functioning inside the larger
        framework the same way other frames do."""
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.root = root
        self.sfamid = None
        return self
