""" Provides a GUI Frame for coloring residues by properties, like hydrophobicity and sidechain
accessibility. """
# pylint: disable=C0103,W0201,W0231,R0902,W0403
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
from .tkinter_utilities import is_int
from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox


class ColorResiduesBySuperFamilyDataGUIFactory(tk.Frame):
    """A Factory for GUI frames, used for coloring residues by properties."""

    def __init__(self, property_name, nice_name, maxvalue, cutoff):
        self.property_name = property_name
        self.nice_name = nice_name
        self.maxvalue = maxvalue
        self.cutoff = cutoff

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirity the first time it is opened."""
        if len(self.winfo_children()) < 2:
            self.fill_frame()
        self.get_objects_in_scene()
        self.fill_listbox()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        text = (
            "This functionality will color objects based on their %s." % self.nice_name
        )
        tk.Label(self, text=text).pack(side="top", anchor="nw", fill=tk.X)
        tk.Label(
            self,
            text="Select objects from the list below to color them by their %s."
            % self.nice_name,
        ).pack(side="top", anchor="nw", fill=tk.X)

        # List box with scrollbar containing objects
        object_list_frame = tk.Frame(self)
        list_scrollbar = tk.Scrollbar(object_list_frame, orient=tk.VERTICAL)
        listbox = tk.Listbox(
            object_list_frame,
            yscrollcommand=list_scrollbar.set,
            height=15,
            selectmode=tk.EXTENDED,
        )
        self.object_list = [
            "There are no objects in the current scene"
        ]  # Placeholder message

        list_scrollbar.config(command=listbox.yview)
        list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        listbox.pack(side="top", anchor="nw", fill=tk.X)

        object_list_frame.pack(side="top", anchor="nw", fill=tk.X)
        self.listbox = listbox

        # Controls
        controls_frame = tk.Frame(self)

        # Maxvalue entry
        maxvalue_variable = tk.IntVar(self)
        maxvalue_variable.set(self.maxvalue)  # Default maxvalue value

        is_int_cmd = (self.register(is_int), "%P")
        maxvalue_entrybox = tk.Entry(
            controls_frame,
            text=maxvalue_variable,
            validate="key",
            width=10,
            validatecommand=is_int_cmd,
        )
        tk.Label(controls_frame, text="Maximum amount of values:").pack(
            side="left", anchor="nw", fill=tk.X, pady=10
        )
        maxvalue_entrybox.pack(side="left", anchor="nw", pady=10)

        # Cutoff entry
        cutoff_variable = tk.IntVar(self)
        cutoff_variable.set(self.cutoff)  # Default maxvalue value

        cutoff_entrybox = tk.Entry(
            controls_frame,
            text=cutoff_variable,
            validate="key",
            width=10,
            validatecommand=is_int_cmd,
        )

        cutoff_entrybox.pack(side="right", anchor="ne", pady=10)
        tk.Label(controls_frame, text="Cutoff value:").pack(
            side="right", anchor="ne", fill=tk.X, pady=10
        )

        controls_frame.pack(side="top", anchor="nw", fill=tk.X)

        self.maxvalue_variable = maxvalue_variable
        self.cutoff_variable = cutoff_variable

        # Submit button
        tk.Button(
            self,
            text="Color",
            command=self.color_and_label_selected_objects,
        ).pack(side="top", anchor="ne", pady=10)

    def color_and_label_selected_objects(self):
        """Colors the selected objects by a superfamily property."""

        selected_objects = [
            self.listbox.get(idx) for idx in self.listbox.curselection()
        ]

        if len(selected_objects) == 0:
            selected_objects = None

        try:
            self.root.viewer_instance.show_superfamily_data(
                self.property_name,
                maxvalue=self.maxvalue_variable.get(),
                cutoff=self.cutoff_variable.get(),
                objects=selected_objects,
            )
            self.root.switch_to_scene_details()
        except ValueError as err:
            error = str(err)
            if "no numbering scheme was returned" in error:
                obj = error.split("(")[1].split(")")[
                    0
                ]  # Extract the PDB from the error message
                tkMessageBox.showwarning(
                    "Could not renumber %s" % obj,
                    "Warning: Could not renumber %s. This object cannot be found in the current "
                    "system. Please remove this structure to avoid further conflicts."
                    % obj,
                )
            else:
                tkMessageBox.showerror("ValueError", str(err))

    def get_objects_in_scene(self):
        """Sets the object list to a list of all the names of objects in the scene."""
        self.object_list = [
            obj
            for obj in self.root.pymol_instance.cmd.get_object_list()
            if not obj.startswith("3dm_meta")
        ]

    def fill_listbox(self):
        """Fills the listbox with Object ids."""
        self.listbox.delete(0, tk.END)
        for obj in self.object_list:
            self.listbox.insert(tk.END, obj)

    def __call__(self, parent, root, *args, **kwargs):
        """This method takes the place of the normal __init__ method. This way the GUI can be
        created with some small variations in features, while still functioning inside the larger
        framework the same way other frames do."""
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.root = root
        self.sfamid = None
        return self
