""" Provides a GUI frame for downloading corresponding compounds in PyMol. """
# pylint: disable=C0103,W0201,W0231
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk


class LoadCorrespondingCompoundsGUI(tk.Frame):
    """Provides a GUI frame for downloading corresponding compounds in PyMol."""

    def __init__(self, parent, root, show_messages=True, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.show_messages = show_messages
        self.root = root

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirity the first time it is opened."""
        if len(self.winfo_children()) < 2:
            self.fill_frame()
        self.get_objects_in_scene()
        self.fill_listbox()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        tk.Label(
            self,
            text="You can load corresponding compounds in the current scene with this menu.",
        ).pack(side="top", anchor="nw", fill=tk.X)

        # List box with scrollbar containing objects
        object_list_frame = tk.Frame(self)
        list_scrollbar = tk.Scrollbar(object_list_frame, orient=tk.VERTICAL)
        listbox = tk.Listbox(
            object_list_frame,
            yscrollcommand=list_scrollbar.set,
            height=19,
            selectmode=tk.EXTENDED,
        )
        self.object_list = ["There are no elligible objects in the current scene"]

        list_scrollbar.config(command=listbox.yview)
        list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        listbox.pack(side="top", anchor="nw", fill=tk.X)

        object_list_frame.pack(side="top", anchor="nw", fill=tk.X)
        self.listbox = listbox

        # Drug/protein loading checkboxes
        self.load_type_var = tk.IntVar(self)

        load_drug_butt = tk.Radiobutton(
            self, text="Download compounds", variable=self.load_type_var, value=1
        )
        load_prot_butt = tk.Radiobutton(
            self, text="Download protein", variable=self.load_type_var, value=2
        )
        load_full_butt = tk.Radiobutton(
            self, text="Download full", variable=self.load_type_var, value=3
        )

        load_drug_butt.pack(side="left", anchor="nw", pady=10)
        load_prot_butt.pack(side="left", anchor="nw", pady=10)
        load_full_butt.pack(side="left", anchor="nw", pady=10)

        load_drug_butt.select()

        # Submit button
        tk.Button(
            self, text="Download", command=self.load_corresponding_compounds
        ).pack(side="top", anchor="ne", pady=10)

    def load_corresponding_compounds(self):
        """Downloads partial objects based on a list of objects set by the user, either through
        the optional parameters, or a dialog.

        :param download_mode: String, either protein or ligand. Corresponds to which item type
                              must be downloaded.
        :param objects: A list of PDB ids where the missing item should be downloaded."""
        download_mode = self.load_type_var.get()
        selected_objects = [
            self.listbox.get(idx) for idx in self.listbox.curselection()
        ]

        specifier = {1: "hetatm", 2: "amino", 3: "full"}[download_mode]
        for pdb_id in selected_objects:
            pdb_id = pdb_id[:5]  # Make sure we only get the PDB ID
            self.root.viewer_instance.load_pdb(
                pdb_id.split("_")[0], self.root.viewer_instance.get_sfamid(), specifier
            )

    def get_objects_in_scene(self):
        """Sets the object list to a list of all the names of objects in the scene."""
        self.object_list = self.root.pymol_instance.cmd.get_object_list()

    def fill_listbox(self):
        """Fills the listbox with Object ids."""
        self.listbox.delete(0, tk.END)
        for obj in self.object_list:
            self.listbox.insert(tk.END, obj)


frame = LoadCorrespondingCompoundsGUI
