""" Provides a GUI Frame for the loading PDB files from 3DM. """
# pylint: disable=W0201,C0103,W0102,R0902
try:
    import Tkinter as tk
    import ttk
except ImportError:
    import tkinter as tk
    import tkinter.ttk as ttk
import re
import traceback
from functools import partial
from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox
from ....shared.api.NumberingSchemes import SaveableAlignmentWarning, NumberingScheme


class AlternativeAlignmentDialog(object):
    """ " A dialog that enables users to select an alternative alignment when the current
    alignment does not support the requested PDB."""

    def __init__(self, parent, alignments, load_pdb_args):
        """Initializes the dialog that enables users to select an alternative alignment when
        the current alignment does not support the requested PDB.

        :param parent: Parent Tkinter element
        :param alignments: A dictionary containing alignment names as keys and alignment
        identifiers as values.
        :param load_pdb_args: The arguments that should be passed to load_pdb when the user
        has switched families.

        """
        self.root = parent
        top = self.top = tk.Toplevel(parent, pady=10, padx=10)
        # Set modal size.
        top.geometry("500x160")
        tk.Label(top, text=("No valid numbering scheme was found for this PDB.")).pack(
            side="top", anchor="nw", fill=tk.X
        )
        tk.Label(
            top,
            text=("The following alignments will result in a valid numbering scheme:"),
        ).pack(side="top", anchor="nw", fill=tk.X)

        alignments_nice = ["%i: %s" % (i, a) for a, i in alignments.items()]

        self.alignments_var = tk.StringVar(parent)
        self.alignments_var.set(alignments_nice[0])  # Default list value

        self.alignments = ttk.Combobox(
            top, textvariable=self.alignments_var, values=alignments_nice
        )
        self.alignments.current(0)
        self.alignments.pack(side="top", anchor="nw", fill=tk.X, pady=10)

        switch_and_load_cmd = partial(self.switch_to_family, load_pdb_args)

        tk.Button(
            top, text="Switch to family and load PDB", command=switch_and_load_cmd
        ).pack(side="left")
        tk.Button(top, text="Cancel", command=self.cancel).pack(side="right")

    def cancel(self):
        """Cancels the operation for this window."""
        self.top.destroy()

    def switch_to_family(self, load_pdb_args):
        """Switch to family using the alignments dropdown.

        :param load_pdb_args: Arguments for loading the PDB from 3DM.
        """
        try:
            family_id = self.alignments_var.get().strip()
            family_id = int(family_id.split(":")[0])
        except Exception:  # pylint: disable=W0703
            tkMessageBox.show_warning("Invalid family selected.")

        self.root.viewer_instance.storage.family = family_id
        self.root.viewer_instance.load_pdb(*load_pdb_args)

        self.top.destroy()


class LoadFrom3DMGUI(tk.Frame):
    """A GUI Frame for the loading PDB files from 3DM."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.default_search_message = "Search PDB IDs"
        self.root = root
        self.sfamid = None
        self.opened = False

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirity the first time it is opened."""
        if not self.opened:
            self.fill_frame()
            self.opened = True

    def load_pdbs_for_sfam(self, *args, **kwargs):
        """Takes the SFAMID entered by the user in the GUI and fills the listbox with PDBs that
        belong to this superfamily."""
        del args, kwargs
        full_sfam = self.sfams_variable.get()
        n_sfamid = full_sfam.split("(")[-1].strip(")").replace("-", "_")
        objects = self.root.pymol_instance.cmd.get_object_list()

        # Superfamily switch guard
        if self.sfamid is not None and self.sfamid != n_sfamid and len(objects) > 0:
            tkMessageBox.showwarning(
                "Family switch warning",
                "Warning: You selected a 3DM system (%s) that differs from the one that is"
                " currently active (%s). You also have structures from the previous system loaded. "
                " Please remove these structures to prevent conflicts from occuring: %s "
                % (n_sfamid, self.sfamid, (", ".join(objects))),
            )

        api = self.root.viewer_instance._get_api()  # pylint: disable=W0212
        self.sfamid = n_sfamid
        self.pdb_list = ["Loading PDBs for %s ..." % self.sfamid]
        self.fill_listbox()
        try:
            self.pdb_list = self.root.viewer_instance.load_pdbs_for_system(
                self.sfamid, api
            )
        except RuntimeWarning as w:
            # Display warning after 500ms to ensure that the combobox is closed.
            f = partial(tkMessageBox.showwarning, parent=self)
            f.__name__ = "showwarning_self"  # Required for after
            self.after(500, f, "Warning", str(w))

        self.fill_listbox()

    def load_structures(self):
        """Retreives the selected structures from the GUI, downloads the corresponding PDB files
        and opens them using PyMol. The files are deleted afterwards."""
        if self.sfamid is None:
            raise ValueError("You need to select a 3DM system first.")

        self.root.viewer_instance.storage.system = (
            self.sfamid
        )  # Store the SFAMID in PyMolViewer
        if not hasattr(self.root.viewer_instance.storage, "family"):
            self.root.viewer_instance.storage.family = (
                1  # Store the family id in PyMolViewer
            )

        show_drug = bool(self.load_drug_var.get())
        show_prot = bool(self.load_prot_var.get())
        show_full = bool(self.load_full_var.get())

        selected_structures = [
            self.listbox.get(idx) for idx in self.listbox.curselection()
        ]
        if len(selected_structures) == 0:
            raise IndexError("You did not select any structures.")

        if not show_drug and not show_prot:
            tkMessageBox.showwarning(
                "Select drug and/or protein",
                "You unselected both drug and protein. Please select either or both of these "
                "options and try again.",
            )
            return

        specifiers = {
            (True, True, False): "full",  # Whenever show_full is True, full is used.
            (True, False, True): "full",
            (True, True, True): "full",
            (True, False, False): "full",
            (False, True, False): "hetatm",  # When show_drug is True
            (False, False, True): "amino",  # When show_prot is True
        }
        specifier = specifiers.get((show_full, show_drug, show_prot), "chain")

        if len(selected_structures) > 1:
            self.root.viewer_instance.load_pdb_batch(
                selected_structures, self.sfamid, specifier
            )
        else:
            try:
                self.root.viewer_instance.load_pdb(
                    selected_structures[0], self.sfamid, specifier
                )
            except SaveableAlignmentWarning:
                # A numbering scheme was not found, but may be available in another alignment.
                self.show_alternative_alignment_dialog(
                    selected_structures[0], self.sfamid, specifier
                )

            except RuntimeWarning as e:
                # Shows a warning
                tkMessageBox.showwarning("Warning during the loading procedure", str(e))
                # Prints the traceback to the console
                traceback.print_exc()

    def show_alternative_alignment_dialog(self, pdbid, sfamid, specifier):
        """Shows the alternative alignment dialog, used in cases where no correct numbering scheme is
        found in the current alignment.

        :pdbid: The PDB identifier that is currently being loaded.
        :sfamid: Selected superfamily (system) identifier.
        :specifier: The type of PDB to be loaded, see `load_pdb` in the PyMolViewer class.
        """
        api = self.root.viewer_instance._get_api(
            "persistence3dm"
        )  # pylint: disable=W0212
        alignments = self.root.viewer_instance.load_alignment_families(
            self.root.viewer_instance
        )
        valid_alignments = {}
        for alignment_name, alignment_id in alignments.items():
            # Check every alignment to see if it returns a valid numbering scheme.
            ns = NumberingScheme.retrieve_numbering_scheme(
                api, pdbid, alignment_id, sfamid
            )
            print(alignment_name, len(ns))
            if len(ns) > 0:
                valid_alignments[alignment_name] = alignment_id

        AlternativeAlignmentDialog(self.root, alignments, (pdbid, sfamid, specifier))

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        tk.Label(self, text="Load a structure from 3DM.").pack(
            side="top", anchor="nw", fill=tk.X
        )

        # Superfamily selection box
        sfams_variable = tk.StringVar(self)
        sfams_variable.set("Select a 3DM system")  # Default list value
        sfams_variable.trace("w", self.load_pdbs_for_sfam)
        access_api = self.root.viewer_instance._get_api(
            "access"
        )  # pylint: disable=W0212
        systems = self.root.viewer_instance.load_accessible_systems(access_api).keys()
        sfams = ttk.Combobox(self, textvariable=sfams_variable, values=sorted(systems))
        sfams.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        self.sfams_variable = sfams_variable

        # PDB search box
        vcmd = (self.register(self.search), "%P")
        entry = tk.Entry(self, validate="key", validatecommand=vcmd)
        entry.pack(side="top", anchor="nw", fill=tk.X, pady=10)

        entry.delete(0, tk.END)
        entry.insert(0, self.default_search_message)

        # List box with scrollbar
        pdb_list_frame = tk.Frame(self)
        list_scrollbar = tk.Scrollbar(pdb_list_frame, orient=tk.VERTICAL)
        listbox = tk.Listbox(
            pdb_list_frame,
            yscrollcommand=list_scrollbar.set,
            height=14,
            selectmode=tk.EXTENDED,
        )

        self.pdb_list = ["<Please select a 3DM system first.>"]  # Placeholder message

        list_scrollbar.config(command=listbox.yview)
        list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        listbox.pack(side="top", anchor="nw", fill=tk.X)

        pdb_list_frame.pack(side="top", anchor="nw", fill=tk.X)

        # Drug/protein loading checkboxes
        self.load_drug_var = tk.BooleanVar(self)
        self.load_prot_var = tk.BooleanVar(self)
        self.load_full_var = tk.BooleanVar(self)

        load_drug_box = tk.Checkbutton(
            self, text="Download compounds", variable=self.load_drug_var
        )
        load_prot_box = tk.Checkbutton(
            self, text="Download protein", variable=self.load_prot_var
        )
        load_full_box = tk.Checkbutton(
            self, text="Download full structure", variable=self.load_full_var
        )

        load_drug_box.pack(side="left", anchor="nw", pady=10)
        load_prot_box.pack(side="left", anchor="nw", pady=10)
        load_full_box.pack(side="left", anchor="nw", pady=10)

        # Select only the load drug + prot boxes by default
        load_drug_box.select()
        load_prot_box.select()

        # Submit button
        tk.Button(self, text="Load Structures", command=self.load_structures).pack(
            side="top", anchor="ne", pady=10
        )

        self.listbox = listbox
        self.fill_listbox()

        # Check set sfams
        try:
            self.sfamid = self.root.viewer_instance.get_sfamid()
        except (AttributeError, RuntimeWarning, ValueError):
            self.sfamid = None

        if self.sfamid:
            sfams.set(
                [
                    system
                    for system in sorted(systems)
                    if self.sfamid.replace("_", "-") in system
                ][0]
            )
            self.load_pdbs_for_sfam()

    def search(self, _filter):
        """Searches the listbox with a filter. Always returns True."""
        if _filter == self.default_search_message:
            return True

        if len(_filter.strip()) == 0:
            self.fill_listbox()
            return True

        self.fill_listbox(_filter)
        return True

    def fill_listbox(self, _filter=None):
        """Fills the listbox with PDB ids, limited by a filter (Regex string)."""
        if _filter is not None:
            try:
                filter_re = re.compile(_filter, flags=re.IGNORECASE)
            except:  # pylint: disable=W0702
                filter_re = re.compile(".")
        else:
            filter_re = re.compile(".")

        self.listbox.delete(0, tk.END)
        for pdb in self.pdb_list:
            if not filter_re.match(pdb):
                continue
            self.listbox.insert(tk.END, pdb)


frame = LoadFrom3DMGUI
