""" Provides a GUI that allows the user to load a basket. """
# pylint: disable=C0103,R0901,R0902,W0403,W0212,W0201,W0613
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox
from ....shared.viewer.CompositeUtilities import CompositeUtilities


class LoadHotspotBasketGUI(tk.Frame):
    """Provides a GUI that allows the user to load a basket."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.selected_basket = None
        self.root = root
        self.fill_frame()

    def activate(self):
        """Loads available baskets when the frame is activated."""
        self.load_available_baskets()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""

        # List type indicator
        header_frame = tk.Frame(self)

        tk.Label(header_frame, text="Baskets").pack(side="left", anchor="nw")
        tk.Label(header_frame, text="Items").pack(side="right", anchor="ne")

        header_frame.pack(side="top", anchor="nw", fill=tk.X, pady=3)

        two_lists_frame = tk.Frame(self)
        LISTBOX_HEIGHT = 19
        LISTBOX_WIDTH = 36

        # Basket List box with scrollbar
        basket_list_frame = tk.Frame(two_lists_frame)
        basket_list_scrollbar = tk.Scrollbar(basket_list_frame, orient=tk.VERTICAL)
        basket_listbox = tk.Listbox(
            basket_list_frame,
            yscrollcommand=basket_list_scrollbar.set,
            height=LISTBOX_HEIGHT,
            width=LISTBOX_WIDTH,
            selectmode=tk.BROWSE,
        )
        basket_listbox.bind("<<ListboxSelect>>", self.show_basket_contents)
        self.basket_listbox = basket_listbox

        self.basket_list = ["No baskets available."]  # Placeholder message

        basket_list_scrollbar.config(command=basket_listbox.yview)
        basket_list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        basket_listbox.pack(side="top", anchor="nw", fill=tk.X, expand=True)

        basket_list_frame.pack(side="left", anchor="nw", fill=tk.X)

        # Item List box with scrollbar
        item_list_frame = tk.Frame(two_lists_frame)
        item_list_scrollbar = tk.Scrollbar(item_list_frame, orient=tk.VERTICAL)

        item_listbox = tk.Listbox(
            item_list_frame,
            yscrollcommand=item_list_scrollbar.set,
            height=LISTBOX_HEIGHT,
            width=LISTBOX_WIDTH,
            selectmode=tk.EXTENDED,
        )

        self.item_listbox = item_listbox

        self.items_list = ["No items available."]  # Placeholder message

        item_list_scrollbar.config(command=basket_listbox.yview)
        item_list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        item_listbox.pack(side="top", anchor="nw", fill=tk.X, expand=True)

        item_list_frame.pack(side="right", anchor="ne", fill=tk.X)

        two_lists_frame.pack(side="top", anchor="nw", fill=tk.X)

        # Button bar
        button_bar_frame = tk.Frame(self)

        load_basket_button = tk.Button(
            button_bar_frame, text="Load basket", command=self.load_basket_into_view
        )
        refresh_hotspots_button = tk.Button(
            button_bar_frame,
            text="Refresh hotspot baskets",
            command=self.refresh_baskets,
        )
        delete_basket_button = tk.Button(
            button_bar_frame, text="Delete basket", command=self.delete_basket
        )

        load_basket_button.pack(side="left", anchor="nw", pady=10)
        refresh_hotspots_button.pack(side="left", anchor="nw", pady=10, padx=10)
        delete_basket_button.pack(side="left", anchor="nw", pady=10)

        button_bar_frame.pack(side="top", anchor="nw", fill=tk.X)

    def refresh_baskets(self, event=None):
        """Refreshes the list of baskets."""
        self.load_available_baskets()
        self.fill_basket_listbox()

    def fill_basket_listbox(self):
        """Fills the basket listbox with baskets listed in self.basket_list."""
        self.basket_listbox.delete(0, tk.END)
        for basket in self.basket_list:
            self.basket_listbox.insert(tk.END, basket)

    def fill_item_listbox(self):
        """Fills the items listbox with items listed in self.items_list."""
        self.item_listbox.delete(0, tk.END)
        for item in self.items_list:
            self.item_listbox.insert(tk.END, item)

    def load_available_baskets(self):
        """Loads the baskets that are available in the current situation and assigns them to the
        self.basket_ids and self.basket_list lists."""
        baskets = CompositeUtilities.load_available_hotspot_baskets(
            self.root.viewer_instance
        )

        self.basket_ids = [basket.id for basket in baskets]
        self.basket_list = [basket.label for basket in baskets]
        self.fill_basket_listbox()

    def show_basket_contents(self, event=None):  # pylint: disable=W0613
        """Shows the contents of a selected basket inside the item listbox."""
        selection = self.basket_listbox.curselection()
        if len(selection) == 0:
            return
        selected_basket = self.basket_ids[selection[0]]
        self.selected_basket = self.basket_list[selection[0]]
        self.selected_basket_id = self.basket_ids[selection[0]]
        self.items_list = self.load_basket(selected_basket)
        self.fill_item_listbox()

    def load_basket(self, basket_id):
        """Loads and returns the items that exist inside a basket."""
        vi = self.root.viewer_instance
        baskets_api = vi._get_baskets_api()
        basket = baskets_api.load_basket(basket_id)
        items_list = [
            "%s: %s" % (item.get("value"), item.get("label")) for item in basket.items
        ]

        return items_list

    def load_basket_into_view(self, raise_error_on_none=False):
        """Loads a basket and visualizes them inside the PyMol scene.

        :param raise_error_on_none: Raise an error if no basket is selected. Otherwise, show error
                                    window and continue.
        """
        if self.selected_basket is None:
            if raise_error_on_none:
                raise ValueError("Select a basket to load.")
            else:
                tkMessageBox.showerror(
                    "Select a basket to load.", "Select a basket to load."
                )

        basket_name = self.selected_basket
        item_selection = self.item_listbox.curselection()
        if len(item_selection) == 0:
            selected_items = self.items_list
        else:
            selected_items = [self.items_list[s] for s in item_selection]

        aligned_positions, annotations = zip(
            *[item.split(":", 1) for item in selected_items]
        )
        self.root.viewer_instance.show_basket_positions(
            aligned_positions, annotations, basket_name
        )

    def delete_basket(self, skip_ask=False):
        """Deletes a basket. asks the user for permission first."""
        selection = self.basket_listbox.curselection()
        if len(selection) == 0:
            return
        selected_basket_id = self.basket_ids[selection[0]]

        if not skip_ask:
            user_response = tkMessageBox.askyesno(
                "Remove basket", "Are you sure you want to remove this basket?"
            )
            if not user_response:
                return False

        try:
            self.root.viewer_instance.delete_basket(selected_basket_id)

        except RuntimeError as err:
            tkMessageBox.showerror(
                "Could not remove basket.",
                "Could not remove this basket. Do you have permission to delete this basket?",
            )
            print("RuntimeError: ", str(err))
            return False

        self.load_available_baskets()
        return True


frame = LoadHotspotBasketGUI
