""" Provides a GUI Frame for the loading model files from 3DM. """
# pylint: disable=W0201,C0103,W0102,R0902
try:
    import Tkinter as tk
    import ttk
except ImportError:
    import tkinter as tk
    import tkinter.ttk as ttk
import re
from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox


class LoadModelFrom3DMGUI(tk.Frame):
    """A GUI Frame for the loading model files from 3DM."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.default_search_message = "Search model IDs"
        self.root = root
        self.family_dict = {}
        self.sfamid = None
        self.opened = False

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirity the first time it is opened."""
        if not self.opened:
            self.fill_frame()
            self.opened = True

    def _get_sfamid(self):
        """Gets the SFAMID from the GUI combobox."""
        full_sfam = self.sfams_variable.get()
        sfamid = full_sfam.split("(")[-1].strip(")").replace("-", "_")
        return sfamid

    def _get_famid(self):
        """Gets the FamilyID (int) from the GUI combobox. Defaults to 1."""
        full_family = self.families_variable.get()
        try:
            return self.family_dict[full_family]
        except KeyError:
            return 1

    def load_models_for_sfam(self, *args, **kwargs):
        """Takes the SFAMID entered by the user in the GUI and fills the listbox with models that
        belong to this superfamily."""
        del args, kwargs
        objects = self.root.pymol_instance.cmd.get_object_list()

        # Superfamily switch guard
        if (
            self.sfamid is not None
            and self._get_sfamid() != self.sfamid
            and len(objects) > 0
        ):
            tkMessageBox.showwarning(
                "Family switch warning",
                "Warning: You selected a family that differs from the one that is currently active."
                " You also have structures from the previous family loaded. Please remove these "
                "structures to prevent conflicts from occuring: "
                + (", ".join(objects)),
            )

        sfamid = self._get_sfamid()
        famid = self._get_famid()

        api = self.root.viewer_instance._get_api(
            "persistence3dm"
        )  # pylint: disable=W0212
        self.sfamid = sfamid
        self.famid = famid

        self.model_list = ["Loading models for %s ..." % sfamid]
        self.fill_listbox()
        self.model_list = self.root.viewer_instance.load_models_for_system(
            sfamid, api, famid
        )
        self.fill_listbox()

    def load_structures(self):
        """Retrieves the selected structures from the GUI, downloads the corresponding model files
        and opens them using PyMol. The files are deleted afterwards."""
        if self.sfamid is None:
            raise ValueError("You need to select a Superfamily first.")

        self.root.viewer_instance.storage.system = (
            self.sfamid
        )  # Store the SFAMID in PyMolViewer
        self.root.viewer_instance.storage.family = (
            self.famid
        )  # Store the family id in PyMolViewer

        selected_structures = [
            self.listbox.get(idx) for idx in self.listbox.curselection()
        ]
        if len(selected_structures) == 0:
            raise IndexError("You did not select any structures.")
        for model_id in selected_structures:
            try:
                self.root.viewer_instance.load_model(
                    self.root.viewer_instance, model_id, self.sfamid, self.famid
                )
            except ValueError:
                tkMessageBox.showerror(
                    "Could not load model",
                    "Could not load model %s. Is this an alias for another model?"
                    % model_id,
                )

    def fill_families(self, *args, **kwargs):
        """Fills the families combobox with family names belonging to the currently selected
        superfamily. Called automatically when switching superfamilies."""
        del args, kwargs
        families = self.root.viewer_instance.load_alignment_families(
            self.root.viewer_instance, self._get_sfamid()
        )
        self.family_dict = families
        self.families_cb["values"] = families.keys()
        self.families_variable.set("Select a family (Superfamily 3DM default)")

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        tk.Label(self, text="Load a structure from 3DM.").pack(
            side="top", anchor="nw", fill=tk.X
        )

        # Superfamily selection box
        sfams_variable = tk.StringVar(self)
        sfams_variable.set("Select a superfamily")  # Default list value

        # Update the list of models and the list of families every time the superfamily changes.
        sfams_variable.trace("w", self.load_models_for_sfam)
        sfams_variable.trace("w", self.fill_families)

        access_api = self.root.viewer_instance._get_api(
            "access"
        )  # pylint: disable=W0212
        systems = self.root.viewer_instance.load_accessible_systems(access_api).keys()
        sfams = ttk.Combobox(self, textvariable=sfams_variable, values=sorted(systems))
        sfams.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        self.sfams_variable = sfams_variable

        # Family selection box
        families_variable = tk.StringVar(self)
        families_variable.set(
            "Select a family (Superfamily 3DM default)"
        )  # Default list value

        # Update the list of models every time the family is changed
        families_variable.trace("w", self.load_models_for_sfam)
        families_cb = ttk.Combobox(self, textvariable=families_variable, values=[])
        families_cb.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        self.families_variable = families_variable
        self.families_cb = families_cb

        # model search box
        vcmd = (self.register(self.search), "%P")
        entry = tk.Entry(self, validate="key", validatecommand=vcmd)
        entry.pack(side="top", anchor="nw", fill=tk.X, pady=10)

        entry.delete(0, tk.END)
        entry.insert(0, self.default_search_message)

        # List box with scrollbar
        model_list_frame = tk.Frame(self)
        list_scrollbar = tk.Scrollbar(model_list_frame, orient=tk.VERTICAL)
        listbox = tk.Listbox(
            model_list_frame,
            yscrollcommand=list_scrollbar.set,
            height=12,
            selectmode=tk.EXTENDED,
        )

        self.model_list = [
            "<Please select a superfamily first.>"
        ]  # Placeholder message

        list_scrollbar.config(command=listbox.yview)
        list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        listbox.pack(side="top", anchor="nw", fill=tk.X)

        model_list_frame.pack(side="top", anchor="nw", fill=tk.X)

        # Submit button
        tk.Button(self, text="Load models", command=self.load_structures).pack(
            side="top", anchor="ne", pady=10
        )

        self.listbox = listbox
        self.fill_listbox()

    def search(self, _filter):
        """Searches the listbox with a filter. Always returns True."""
        if _filter == self.default_search_message:
            return True

        if len(_filter.strip()) == 0:
            self.fill_listbox()
            return True

        self.fill_listbox(_filter)
        return True

    def fill_listbox(self, _filter=None):
        """Fills the listbox with model ids, limited by a filter (Regex string)."""
        if _filter is not None:
            try:
                filter_re = re.compile(_filter, flags=re.IGNORECASE)
            except:  # pylint: disable=W0702
                filter_re = re.compile(".")
        else:
            filter_re = re.compile(".")

        self.listbox.delete(0, tk.END)
        for model in self.model_list:
            if not filter_re.match(model):
                continue
            self.listbox.insert(tk.END, model)


frame = LoadModelFrom3DMGUI
