""" A frame that allows the user to log in to 3DM. """
# pylint: disable=C0103

import json
import webbrowser

try:
    import Tkinter as tk
    import tkSimpleDialog
    import tkFileDialog
except ImportError:
    import tkinter as tk
    import tkinter.simpledialog as tkSimpleDialog
    import tkinter.filedialog as tkFileDialog

from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox
from ....shared.viewer.ThreeDMWebsite import ThreeDMWebsite
from ....shared.api.APIAuth import APIAuth
from ....shared.viewer.Updates import Updates


# The following import makes testing easier, as it instantly adds the testing API password to the
# input boxes, without having the user enter them. The `test` package is not copied with the
# distribution version of the plugin.
try:
    from plugin.shared.test.credentials import API_PASSWORD, API_USERNAME
except ImportError:
    API_USERNAME = "Username"
    API_PASSWORD = "Password"


class LogInGUI(tk.Frame):
    """A frame that allows the user to log in to 3DM."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.root = root
        self.fill_frame()

    def check_auto_login(self):
        """Automatically logs in and shows the success state if the user has an active session."""
        api_auth = APIAuth(try_auto_login=True, client_id="pymol-plugin-v2")

        if api_auth._check_token_validity() == 0:  # pylint: disable=W0212
            self.root.viewer_instance.storage.api_auth = api_auth
            self.login_success()
            return True

        return False

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        tk.Label(
            self,
            text="Welcome to the Bio-Prodict 3DM plugin for PyMol. Please log in using your "
            "username (email) and password.",
            wraplength=400,
        ).pack(side="top", anchor="nw", fill=tk.X, pady=20)

        tk.Label(self, text="Username").pack(side="top", anchor="nw", fill=tk.X)
        username = tk.Entry(self)
        username.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        username.delete(0, tk.END)
        username.insert(0, API_USERNAME)
        self.username_entry = username

        tk.Label(self, text="Password").pack(side="top", anchor="nw", fill=tk.X)
        password = tk.Entry(self, show="*")
        password.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        password.delete(0, tk.END)
        password.insert(0, API_PASSWORD)
        self.password_entry = password

        self.message_label = tk.Label(
            self, text="You need to log in to enable the menu."
        )
        self.message_label.pack(side="top", anchor="nw", fill=tk.X)

        current_version = self.root.viewer_instance.VERSION
        new_version_available = Updates.new_version_available(
            current_version, "pymol", check_updates_force=True
        )

        tk.Button(self, text="Log in", command=self.log_in).pack(
            side="top", anchor="center", pady=10
        )

        if new_version_available:
            tk.Label(
                self,
                text="A new version is available (v%s). Please update the plugin for optimal "
                "performance." % new_version_available,
                fg="blue",
            ).pack(side="top", anchor="nw", fill=tk.X)

        # Login with scene button
        tk.Button(
            self, text="Log in using Bio-Prodict scene", command=self.login_with_scene
        ).pack(side="bottom", anchor="e", pady=10)

    def login_success(self):
        """Present the user with the login success state."""
        tkMessageBox.showinfo(
            "Login successful.",
            "Login successful. You can now use the menu in the top left.",
            parent=self,
        )
        self.root.enable_menu()

    def login_error(self, err):
        """Shows an error explaining a failed login."""
        try:
            error = json.loads(str(err))["error_description"]
        except ValueError:
            error = str(err)
        tkMessageBox.showerror(
            "Error logging in.",
            "We were unable to log you in to this account. The server responded: %s"
            % error,
            parent=self,
        )

    def handle_twofactor(self):
        """Handles twofactor authentication."""
        twofactor = tkSimpleDialog.askstring(
            "Two factor authentication",
            "Your organization requires two-factor authentication.",
        )

        username = self.username_entry.get()
        password = self.password_entry.get()

        try:
            self.root.viewer_instance.api_login(username, password, twofactor)
        except AssertionError as err:
            self.login_error(err)
            return

        self.login_success()

    def log_in(self):
        """Attempts to log in the user using credentials from the login form. Enables the menu if
        the login was succcesful."""
        username = self.username_entry.get()
        password = self.password_entry.get()

        try:
            self.root.viewer_instance.api_login(username, password)
            self.login_success()
        except AssertionError as err:
            if "twofactor" in str(err):
                self.handle_twofactor()
            else:
                self.login_error(err)

    def login_with_scene(self):
        """Allows the user to log in with a Bio-Prodict scene."""
        options = {
            "defaultextension": ".pse",
            "filetypes": [("PyMol Sessions", ".pse")],
        }

        launch = tkMessageBox.askquestion(
            "Log in using Bio-Prodict session",
            "You can log in using a Bio-Prodict session. This will launch your browser and you will"
            " be able to download a login session after authentication. Use the file dialog to open"
            " this file and you will be logged in. Would you like to continue?",
        )

        if launch != "yes":
            return

        url = ThreeDMWebsite.get_login_scene_url("pymol")
        with self.root.viewer_instance.clean_ld_lib_path():
            webbrowser.open(url)

        filename = tkFileDialog.askopenfilename(**options)

        try:
            self.root.viewer_instance.login_with_scene(filename)
        except Exception as err:  # pylint: disable=W0703
            self.login_error(err)
            return

        self.login_success()


frame = LogInGUI
