""" Provides a GUI Frame for renumering objects. """
# pylint: disable=C0103,W0201,W0231
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox
from ....shared.viewer.CompositeUtilities import CompositeUtilities


class RenumberObjectsGUI(tk.Frame):
    """Provides a GUI Frame for renumering objects."""

    def __init__(self, parent, root, show_messages=True, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.show_messages = show_messages
        self.root = root

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirity the first time it is opened."""
        if len(self.winfo_children()) < 2:
            self.fill_frame()
        self.get_objects_in_scene()
        self.fill_listbox()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        tk.Label(
            self, text="You can renumber objects in the current scene with this menu."
        ).pack(side="top", anchor="nw", fill=tk.X)

        # List box with scrollbar containing objects
        object_list_frame = tk.Frame(self)
        list_scrollbar = tk.Scrollbar(object_list_frame, orient=tk.VERTICAL)
        listbox = tk.Listbox(
            object_list_frame,
            yscrollcommand=list_scrollbar.set,
            height=16,
            selectmode=tk.EXTENDED,
        )
        self.object_list = [
            "There are no objects in the current scene"
        ]  # Placeholder message

        list_scrollbar.config(command=listbox.yview)
        list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        listbox.pack(side="top", anchor="nw", fill=tk.X)

        object_list_frame.pack(side="top", anchor="nw", fill=tk.X)
        self.listbox = listbox

        # Renumbering selection box
        renumbering_scheme_var = tk.StringVar(self)
        renumbering_scheme_var.set(
            "Select a new numbering format."
        )  # Default list value
        numbering_schemes = self.get_numbering_schemes()
        num_scheme_selector = tk.OptionMenu(
            self, renumbering_scheme_var, *numbering_schemes
        )
        num_scheme_selector.pack(side="top", anchor="nw", fill=tk.X, pady=10)

        self.renumbering_scheme_var = renumbering_scheme_var

        # Submit button
        tk.Button(self, text="Renumber", command=self.renumber_objects).pack(
            side="top", anchor="ne", pady=10
        )

    def renumber_objects(self, clear_views_answer=None):
        """Takes the currently selected objects from the listbox and renumbers them to the selected
        numbering scheme.

        :param clear_views_answer: Optional, allows you to force and answer for the
                                   clear_numbering_sensitive_views dialog."""
        selected_objects = [
            self.listbox.get(idx) for idx in self.listbox.curselection()
        ]
        if len(selected_objects) == 0:
            tkMessageBox.showinfo(
                "You did not select any objects to renumber.",
                "You did not select any objects to renumber.",
            )
            return

        target_numbering = self.get_selected_numbering_scheme()
        viewer_instance = self.root.viewer_instance

        if not viewer_instance.clear_numbering_sensitive_views(clear_views_answer):
            # User declined removing numbering sensitive views
            return

        renumbered_objects = []

        for obj in selected_objects:
            accession, pdb = CompositeUtilities.split_structure_accession(obj)
            try:
                viewer_instance.renumber_pdb(
                    viewer_instance,
                    pdb,
                    viewer_instance.get_sfamid(),
                    CompositeUtilities.get_familyid(viewer_instance),
                    target_numbering,
                    accession=accession,
                )
                renumbered_objects.append(obj)
            except ValueError:
                tkMessageBox.showwarning(
                    "Could not renumber %s" % obj,
                    "Warning: Could not renumber %s. This object cannot be found in the current "
                    "system. This structure will not be renumbered. Other structures in the "
                    "scene will be. Please remove this structure to avoid further conflicts."
                    % obj,
                )
                continue
        if self.show_messages:
            tkMessageBox.showinfo(
                "Renumbering complete",
                "Succesfully renumbered the following objects: "
                + (", ".join(renumbered_objects)),
            )

    def get_objects_in_scene(self):
        """Sets the object list to a list of all the names of objects in the scene."""
        self.object_list = self.root.pymol_instance.cmd.get_object_list()

    def fill_listbox(self):
        """Fills the listbox with Object ids."""
        self.listbox.delete(0, tk.END)
        for obj in self.object_list:
            self.listbox.insert(tk.END, obj)

    def get_numbering_schemes(self):
        """Returns a list of all available numbering schemes for the current family."""

        custom = self.root.viewer_instance.get_custom_numbering_schemes(
            self.root.viewer_instance
        )

        normal = [
            "3DM alignment numbering (align)",
            "PDB numbering (pdb)",
            "Residue numbering (res)",
        ]
        if custom is not None:
            return normal + custom
        else:
            return normal

    def get_selected_numbering_scheme(self):
        """Returns the numbering scheme the user has selected."""
        selected_scheme = self.renumbering_scheme_var.get()
        if "(" in selected_scheme and "(+" not in selected_scheme:
            return selected_scheme.rsplit("(", 1)[1].strip(") ")
        else:
            return selected_scheme


frame = RenumberObjectsGUI
