""" A frame that shows information about the plugin. """
# pylint: disable=C0103,W0406,E1101

import json

try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox
from ....shared.viewer.Proxies import Proxies
from ....shared.Config import Config


class SetProxyInfoGUI(tk.Frame):
    """A frame that shows information about the plugin. Because this frame is specified second to
    last in the PyMol Menu, it will automatically be initialized when the About (final) frame fails,
    allowing the user to add proxy information and restart the plugin. Note that this does not
    happen if update checking has been turned off in the configuration."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.https_verification_state = "Disable" if Config().verify_https else "Enable"
        self.root = root
        self.fill_frame()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        tk.Label(
            self,
            text="Enter your proxy information, you may need to include authentication information:",
            wraplength=400,
        ).pack(side="top", anchor="nw", fill=tk.X, pady=20)

        default_proxies = Proxies.get_proxies()

        tk.Label(self, text="HTTP Proxy").pack(side="top", anchor="nw", fill=tk.X)
        http_proxy = tk.Entry(self)
        http_proxy.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        http_proxy.delete(0, tk.END)
        http_proxy.insert(0, default_proxies.get("http", ""))

        self.http_proxy_entry = http_proxy

        tk.Label(self, text="HTTPS Proxy").pack(side="top", anchor="nw", fill=tk.X)
        https_proxy = tk.Entry(self)
        https_proxy.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        https_proxy.delete(0, tk.END)
        https_proxy.insert(0, default_proxies.get("https", ""))
        self.https_proxy_entry = https_proxy

        tk.Button(self, text="Test proxies", command=self.test_proxies).pack(
            side="left", anchor="center", pady=10
        )

        tk.Button(
            self,
            text="%s HTTPS verification" % self.https_verification_state,
            command=self.toggle_https_verification,
        ).pack(side="left", anchor="center", pady=10)

        tk.Button(
            self, text="Apply proxy settings", command=self.apply_proxy_settings
        ).pack(side="right", anchor="center", pady=10)

    def get_proxies_from_entries(self):
        """Returns a valid proxies dictionary from the entry fields in the GUI."""
        http_proxy = self.http_proxy_entry.get().strip()
        https_proxy = self.https_proxy_entry.get().strip()

        if https_proxy == "":
            https_proxy = http_proxy

        proxies = {"http": http_proxy, "https": https_proxy}

        return proxies

    def test_proxies(self):
        """Tests the proxies specified by the user. Shows a messagebox containing the diagnosis."""
        proxies = self.get_proxies_from_entries()
        try:
            Proxies.test_proxies(proxies)
        except Exception as e:  # pylint: disable=W0703
            tkMessageBox.showerror(
                "Proxy error",
                "There was a problem using the listed proxies: " + str(e),
                parent=self,
            )
            return

        tkMessageBox.showinfo(
            "Proxies are in order.",
            "Succesfully connected to Bio-Prodict with the listed options.",
            parent=self,
        )

    def toggle_https_verification(self):
        """Toggles verify_https in the currently loaded config file."""
        if self.https_verification_state == "Disable":
            confirmation = tkMessageBox.askokcancel(
                "Confirm disabling HTTPS verification.",
                "This option disables verifying the secure connection with Bio-Prodict servers. It"
                " can be used to bypass errors that arise from custom Certificate Authorities and"
                " workplace proxies. Please only disable this if you are instructed to do so or"
                "  are aware of the repercussions yourself.",
                parent=self,
            )
            if not confirmation:
                return False

        with open(Config().path) as f:
            config_data = json.load(f)

        config_data[
            "verify_https"
        ] = not Config().verify_https  # Toggle the verify state

        with open(Config().path, "w") as f:
            json.dump(config_data, f, indent=4)

        tkMessageBox.showinfo(
            "Stored HTTPS verification information",
            "Successfully stored HTTPS verification information. You must restart PyMol"
            " to apply these settings.",
            parent=self,
        )

    def apply_proxy_settings(self):
        """Applies the proxy settings specified by the user by saving them to the .PROXIES file."""
        proxies = self.get_proxies_from_entries()
        Proxies.dump_proxies_to_file(proxies)

        tkMessageBox.showinfo(
            "Stored proxy information",
            "Successfully stored proxy information. You must restart PyMol to get your proxy"
            " settings to work.",
            parent=self,
        )


frame = SetProxyInfoGUI
