""" A frame the allows the user to get a detailed view of the contents in the scene. This
specifically includes details about superfamily visualizations the user requested. """
# pylint: disable=C0103,R0902,R0914
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
import webbrowser
from .messagebox_wrapper import tkMessageBoxWrapper as tkMessageBox
from ....shared.viewer.Selection import Selection
from ....shared.viewer.ThreeDMWebsite import ThreeDMWebsite
from ....shared.viewer.CompositeUtilities import CompositeUtilities


class AminoAcidDistributionGUI(tk.Frame):
    """A frame that allows the user to view the contents of the scene."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.root = root
        self.canvas = None
        # These colors are the actual colors used by 3DM.
        self.aa_colors = {
            "A": "#33ff00",
            "C": "#ffff00",
            "D": "#ff0000",
            "E": "#ff0000",
            "F": "#33ff00",
            "G": "#aaaaaa",
            "H": "#8800ff",
            "I": "#33ff00",
            "K": "#005fa9",
            "L": "#33ff00",
            "M": "#ffff00",
            "N": "#ff00ff",
            "P": "#aaaaaa",
            "Q": "#ff00ff",
            "R": "#005fa9",
            "S": "#33ff00",
            "T": "#33ff00",
            "V": "#33ff00",
            "W": "#33ff00",
            "Y": "#33ff00",
            "-": "#aaaaaa",
        }

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirity the first time it is opened."""
        if len(self.winfo_children()) == 0:
            self.fill_frame()

    def get_currently_selected_residue(self):
        """Returns the currently selected residue number from PyMol (align numbering)."""
        selected_space = {"selected_residues": []}
        self.root.pymol_instance.cmd.iterate(
            Selection(in_selected=True, only_ca=True).pymol(),
            "selected_residues.append((model, resi))",
            space=selected_space,
        )
        try:
            object_id, res_num = selected_space["selected_residues"][0]
        except IndexError:
            return None

        return self.root.viewer_instance.convert_residue_number_from_scene_to_align(
            self.root.viewer_instance, object_id, res_num
        )

    def fill_canvas(self, data, canvas):
        """Fills the canvas with a barchart, illustrating the relation between the different
        distributions of amino acids to each other.

        :param data: A dictionary containing amino acids as keys and their frequency percentage as
                     values (int, 0-100).
        :param canvas: The canvas element where this should be drawn.
        """
        c_height = int(canvas["height"])
        # This variable scales the bars to use to the entire height of the canvas
        max_data_point = max(data.values())
        if max_data_point > 0:
            y_scale = 2.5 * (max_data_point**-1)
        else:
            y_scale = 2.5  # Default scaling factor

        y_gap = 20
        x_stretch = 14
        x_width = 15
        x_gap = 2
        canvas.delete("all")  # Clear canvas of previous visualizations
        for x, (amino_acid, percentage) in enumerate(sorted(data.items())):
            color = self.aa_colors.get(amino_acid, "red")
            percentage *= 100
            x0 = x * x_stretch + x * x_width + x_gap
            y0 = c_height - (percentage * y_scale + y_gap)
            x1 = x * x_stretch + x * x_width + x_width + x_gap
            y1 = c_height - y_gap
            # Draw shadow
            canvas.create_rectangle(
                x0 + 2, y0 + 2, x1 + 2, y1 + 2, fill="gray", outline="silver"
            )
            # Draw actual bar
            canvas.create_rectangle(x0, y0, x1, y1, fill=color, outline=color)
            # Percentage above bar
            canvas.create_text(x0, y0, anchor=tk.SW, text="%.1f" % percentage)
            # Amino acid code under bar.
            canvas.create_text(x0 + 4, y1 + 18, anchor=tk.SW, text=str(amino_acid))

    def fill_canvas_with_currently_selected_residue(self):
        """Fills the canvas with the amino acid distribution for the currently selected residue."""
        residue_number = self.get_currently_selected_residue()
        if residue_number is not None:
            self.fill_canvas(
                self.root.viewer_instance.get_amino_acid_distribution(
                    self.root.viewer_instance, residue_number
                ),
                self.canvas,
            )

    def go_to_current_residue_page(self):
        """Gets the currently selected residue and opens a browser showing more information about
        this residue in 3DM (including amino acid distribution)."""
        residue_number = self.get_currently_selected_residue()
        if residue_number is None:
            tkMessageBox.showerror(
                "No residue selected",
                "Can't show more information if no residue is selected.",
            )
            return
        sfamid = self.root.viewer_instance.get_sfamid()
        familyid = CompositeUtilities.get_familyid(self.root.viewer_instance)
        url = ThreeDMWebsite().create_3dm_residue_link(
            residue_number, sfamid, familyid, 1, 1
        )

        with self.root.viewer_instance.clean_ld_lib_path():
            webbrowser.open(url)

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        header = tk.Label(
            self,
            text='Select a residue in the scene and click the "Show" button.\n'
            "The amino acid distribution will appear in here.",
        )
        header.pack(side="top", anchor="nw", fill=tk.X)

        c_width = 600
        c_height = 300
        canvas = tk.Canvas(self, width=c_width, height=c_height)
        canvas.pack(side="top", anchor="nw", fill=tk.X)
        self.canvas = canvas

        # Might as well fill the canvas if a residue is selected when the frame is activated.
        self.fill_canvas_with_currently_selected_residue()

        tk.Button(
            self,
            text="Show amino acid distribution",
            command=self.fill_canvas_with_currently_selected_residue,
        ).pack(side="left", anchor="center", pady=10)

        tk.Button(
            self,
            text="More information about this residue",
            command=self.go_to_current_residue_page,
        ).pack(side="right", anchor="center", pady=10)


frame = AminoAcidDistributionGUI
