""" A frame the allows the user to get a detailed view of the contents in the scene. This
specifically includes details about superfamily visualizations the user requested. """
# pylint: disable=C0103,R0902,R0914
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
from ....shared.viewer.Selection import Selection


class SceneContentDetails(tk.Frame):
    """A frame that allows the user to view the contents of the scene."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.root = root
        self.listbox = None
        self.scene_variable = None
        self.opened = False
        self.load_articles_button = None
        self.current_scene_selectors = []
        self.current_scene_colors = []
        self.features = []
        self.update_scene_timeout = 200

    def update_current_scene(self):
        """Checks if the user switched the scene. If they have, get the scene name and _properly_
        load that scene."""

        actual_scene_name = self.root.pymol_instance.cmd.get("scene_current_name")
        selected_scene_name = self.scene_variable.get()

        if actual_scene_name != selected_scene_name:
            self.scene_variable.set(actual_scene_name)
            self.change_scene()

    def update_current_scene_timer(self):
        """This timer makes sure that the scene is checked and updated on a regular interval."""
        self.update_current_scene()
        self.root.after(self.update_scene_timeout, self.update_current_scene_timer)

    def clear_frame(self):
        """Removes all the widgets from this current frame."""
        for widget in self.winfo_children():
            widget.destroy()

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirety the first time it is opened."""
        if not self.opened:
            self.fill_frame()
            self.update_current_scene_timer()
            self.opened = True
        else:
            self.clear_frame()
            self.fill_frame()
        self.fill_listbox()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        header = tk.Label(
            self,
            text="Select any of the following residues to highlight them inside the scene.",
        )
        header.pack(side="top", anchor="nw", fill=tk.X)

        # Scene selection box
        scene_variable = tk.StringVar(self)
        scene_variable.set("Select a scene")  # Default list value
        scene_variable.trace("w", self.change_scene)
        scene_names = self.root.viewer_instance.storage.scene_features.keys()

        if len(scene_names) == 0:
            scene_names = ["No views containing scene details."]

        scenes = tk.OptionMenu(self, scene_variable, *scene_names)
        scenes.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        self.scene_variable = scene_variable

        feature_list_frame = tk.Frame(self)
        list_scrollbar = tk.Scrollbar(feature_list_frame, orient=tk.VERTICAL)
        listbox = tk.Listbox(
            feature_list_frame,
            yscrollcommand=list_scrollbar.set,
            height=16,
            selectmode=tk.EXTENDED,
        )

        # Calls the on_select method whenever any option in the box is clicked.
        listbox.bind("<<ListboxSelect>>", self.on_select)
        self.features = ["<Select a scene from the dropdown.>"]

        list_scrollbar.config(command=listbox.yview)
        list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        listbox.pack(side="top", anchor="nw", fill=tk.X)

        feature_list_frame.pack(side="top", anchor="nw", fill=tk.X)

        self.load_articles_button = tk.Button(
            self, text="Load articles", command=self.load_articles
        )
        self.load_articles_button.pack(side="top", anchor="center", pady=10)
        self.load_articles_button.config(state="disabled")

        self.listbox = listbox

    def load_articles(self):
        """Opens a number of articles in the user's default browser."""
        current_scene = self.scene_variable.get()
        for i in self.listbox.curselection():
            pos = self.current_scene_selectors[int(i)].numbers
            self.root.viewer_instance.load_article_abstracts_for_position(
                pos, current_scene
            )

    def change_scene(self, *args, **kwargs):
        """Handles changing the scene."""
        del args, kwargs
        pretty_features = []
        current_scene = self.scene_variable.get()

        all_scene_features = self.root.viewer_instance.storage.scene_features
        color_scales = self.root.viewer_instance.storage.scene_color_scales
        lh_articles = self.root.viewer_instance.storage.scene_lh_articles

        scene_features = all_scene_features.get(current_scene)
        color_scale = color_scales.get(current_scene)
        articles = lh_articles.get(current_scene)

        # Disable the load articles button if there are no articles.
        if articles is None:
            self.load_articles_button.config(state="disabled")
        else:
            self.load_articles_button.config(state="normal")

        if scene_features is None:
            self.features = ["No features in this scene."]
            self.root.pymol_instance.cmd.scene(current_scene)
            self.fill_listbox()
            return None

        self.current_scene_selectors = []
        self.current_scene_colors = []

        # Remove all currently applied labels
        self.root.pymol_instance.cmd.label(Selection().pymol(), "")
        scene_features_generator = sorted(
            enumerate(scene_features.items()), key=lambda x: x[1][1], reverse=True
        )

        for _sf, (selector, _property) in scene_features_generator:
            num = selector.numbers
            obj = selector.in_object
            self.current_scene_selectors.append(selector)
            if color_scale is not None:
                self.current_scene_colors.append(color_scale(_property))

            if obj is not None:
                pretty_features.append(
                    "%s (%s) - %s: %s" % (num, obj, current_scene, _property)
                )
            else:
                pretty_features.append("%s - %s: %s" % (num, current_scene, _property))

            # Label everything every time a scene is changed from the 3DM window.
            selector.only_ca = True
            self.root.pymol_instance.cmd.label(
                selector.pymol(), "'%s'" % selector.numbers
            )
            selector.only_ca = False

        self.features = pretty_features
        self.root.pymol_instance.cmd.scene(current_scene)
        self.fill_listbox()

    def fill_listbox(self):
        """Fills the listbox with features from the current scene."""
        self.listbox.delete(0, tk.END)
        for i, feature in enumerate(self.features):
            self.listbox.insert(tk.END, feature)
            try:
                color = self.current_scene_colors[i]
                self.listbox.itemconfig(tk.END, {"fg": color})
            except IndexError:
                pass

    def on_select(self, event):
        """Handles what happens when a user selects a items from the listbox."""
        widget = event.widget
        # Unlabel everything
        self.root.pymol_instance.cmd.label(Selection().pymol(), "")

        try:
            selections = [
                self.current_scene_selectors[int(i)] for i in widget.curselection()
            ]
        except IndexError:
            return
        for selector in selections:
            self.root.pymol_instance.cmd.indicate(selector.pymol())
            selector.only_ca = True
            self.root.pymol_instance.cmd.label(
                selector.pymol(), "'%s'" % selector.numbers
            )
            selector.only_ca = False


frame = SceneContentDetails
