""" This class provides the user with a frame where they can select a viewer-wide variable,
like Superfamily/system, family, subset, etc. It includes a search box tha dynamically filters
the listed values. """
# pylint: disable=W0201,C0103,W0102
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
import re


class ViewerVariablePickerGUI(tk.Frame):
    """An abstract class that that provides a Tkinter frame for generic viewer-wide variable
    selection."""

    def __init__(self, parent, root, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.default_search_message = "Search 3DM system"
        self.label_message = "Select a 3DM system to work with."
        self.root = root
        self.opened = False
        self.values_list = None

    def activate(self):
        """Generic frame activation function. In this case, the frame will be filled in its
        entirity the first time it is opened."""
        if not self.opened:
            self.fill_frame()
            self.opened = True
        else:
            self.load_values_list()
            self.fill_listbox()

    def fill_frame(self):
        """Fills the GUI frame with the appropriate elements."""
        tk.Label(self, text=self.label_message).pack(side="top", anchor="nw", fill=tk.X)

        # System search box
        vcmd = (self.register(self.search), "%P")
        entry = tk.Entry(self, validate="key", validatecommand=vcmd)
        entry.pack(side="top", anchor="nw", fill=tk.X, pady=10)
        entry.delete(0, tk.END)
        entry.insert(0, self.default_search_message)

        # List box with scrollbar
        pdb_list_frame = tk.Frame(self)
        list_scrollbar = tk.Scrollbar(pdb_list_frame, orient=tk.VERTICAL)
        listbox = tk.Listbox(
            pdb_list_frame,
            yscrollcommand=list_scrollbar.set,
            height=17,
        )

        list_scrollbar.config(command=listbox.yview)
        list_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        listbox.pack(side="top", anchor="nw", fill=tk.X)

        pdb_list_frame.pack(side="top", anchor="nw", fill=tk.X)

        # Submit button
        tk.Button(self, text="Select", command=self.set_viewer_variable).pack(
            side="top", anchor="ne", pady=10
        )

        self.listbox = listbox
        self.load_values_list()
        self.fill_listbox()

    def load_values_list(self):
        """This is an abstract class, you need to implement the `load_values_list` function.
        It should return a list of strings containing the options you would to present to the user.

        If the strings end in a value surrounded by (parentheses), `get_selected_value` will return
        that value when one of these selected.
        """
        raise NotImplementedError(
            "This is an abstract class, you need to implement the `load_values_list` function."
        )

    def set_viewer_variable(self):
        """This is an abstract class, you need to implement the `set_viewer_variable` function.

        This function should set the required variable in `self.root.viewer_instance`. It is called
        when the 'Select' button is pressed.
        """
        raise NotImplementedError(
            "This is an abstract class, you need to implement the `set_viewer_variable` function."
        )

    def get_selected_value(self):
        """Returns the value that was selected by the user in the listbox. If the value ends with
        a word in parentheses ('(string)'), that value is returned without the parentheses. If the
        user did not select anything, this function returns None.
        """
        selected = [self.listbox.get(idx) for idx in self.listbox.curselection()]
        if len(selected) == 0:
            return None

        selection = selected[0].strip()
        parens_value = re.findall(r"\([^)]+\)$", selection)
        if len(parens_value) > 0:
            return parens_value[0].strip("()")
        else:
            return selection

    def search(self, _filter):
        """Searches the listbox with a filter. Always returns True."""
        if _filter == self.default_search_message:
            return True

        if len(_filter.strip()) == 0:
            self.fill_listbox()
            return True

        self.fill_listbox(_filter)
        return True

    def fill_listbox(self, _filter=None):
        """Fills the listbox with PDB ids, limited by a filter (Regex string)."""
        if _filter is not None:
            _filter = _filter.replace("*", ".*")
            filter_re = re.compile(_filter, flags=re.IGNORECASE)
        else:
            filter_re = re.compile(".")

        self.listbox.delete(0, tk.END)
        for pdb in self.values_list:
            if not filter_re.match(pdb):
                continue
            self.listbox.insert(tk.END, pdb)
