"""
API interaction module.
"""
import warnings

try:
    from urlparse import urlparse, parse_qs
except ImportError:
    from urllib.parse import urlparse, parse_qs
import sys
from functools import partial
from ..vendors import requests
from ..Config import Config
from ..viewer.Proxies import Proxies

if Config().enable_requests_caching:  # pylint: disable=E1101
    # Import requests_cache during testing, as it reduces server load and prevents
    # "Too many requests" errors.
    print("Installing disk backed testing cache.")
    SITE_PACKAGES_DIR = "/usr/local/lib/python2.7/dist-packages"
    sys.path.append(
        SITE_PACKAGES_DIR
    )  # Temporarily append system site packages directory to path
    try:
        import requests_cache
    except ImportError:
        raise ImportError(
            "Could not import requests_cache, because requests_cache is not installed in your "
            "system site packages directory: %s." % SITE_PACKAGES_DIR
        )
    sys.path.pop()  # And back off it comes.

    class _ConfiguredCachedSession(
        requests_cache.core.CachedSession
    ):  # pylint: disable=R0903
        def __init__(self):
            super(_ConfiguredCachedSession, self).__init__(
                cache_name="molviewer_plugin_cache",
                backend="sqlite",
                expire_after=300,
                allowable_codes=tuple(range(200, 300)),
            )

    requests.sessions.Session = requests.Session = _ConfiguredCachedSession

# Most of the methods in these classes are called on the fly, so this warning is useless.
# pylint: disable=R0903,R0913,W0201,E1101


class TooManyRequestsException(Exception):
    """A custom error that is raised when too many requests are made by a function. This is useful
    during testing, where it can be caught and tests skipped when it occurs, instead of failing."""

    pass


class API(object):
    """This class interacts with the Bio-Prodict API. Every method exposed by the API can be
    accessed by the user using a simple call, ie: api.bio-prodict.nl/persistence3dm/protein is
    accessible through `API().protein()`. This is done dynamically, which is also the reason why
    R0903 is disabled for this class in PyLint.

    When calling one of the API functions, one can thereafter navigate the API using
    TraversableResponse's `follow` method, or by calling a navigation link as a method (`.next()`).

    Refer to the TraversableResponse documentation for more details.
    """

    def __init__(
        self, name, auth, user_agent=None, base_url=None, use_system_proxies=True
    ):
        config = Config()  # Load configuration

        self.auth = auth
        self.name = name
        # Find the API url by name or with the base_url parameter
        self.base_url = config.api_urls.get(name, base_url)
        if self.base_url is None:
            raise ValueError(
                "You requested an API instance that has no known base url. Please add the url for "
                "this API to your configuration file, or pass it as the `base_url` argument here."
            )
        self.sfamid_required = config.api_sfamid_required
        self.user_agent = user_agent
        if use_system_proxies:
            self.proxies = Proxies.get_proxies()
        else:
            self.proxies = {}

    def __call__(
        self,
        path,
        suffix=None,
        sfamid_header=None,
        data=None,
        request_method="GET",
        custom_headers=None,
        **params
    ):
        """This is the function that is returned by __getattr__ on this class. For example, one can
        request API().protein and it will return this method prepared with the path set to protein.

        :param path: This is the first level of the API endpoint. It is supplied by __getattr__.
        :param suffix: A list of url parts that come after the first level. These will be joined
                       with slashes.
        :param sfamid_header: The ID of the 3DM superfamily of you want to access.
        :param data: Optional, any data you'd like to send to the server as part of a POST/PUT/...
                     request.
        :param request_method: Optional, if you specified data to be sent with the request, this
                               specifies the type of request that is made. Can be any of the
                               following: [POST, PATCH, DELETE, PUT]. Defaults to POST.
        :returns: The `requests` Response object, containing the full response, a JSON
        representation (where applicable) and the status code.
        :rtype: requests.Response
        """

        if suffix is None:
            suffix = []

        headers = self.auth.get_request_headers()

        if self.user_agent is not None:
            headers["User-Agent"] = self.user_agent

        if sfamid_header is not None:
            params["sfamid"] = sfamid_header

        if custom_headers is not None:
            headers.update(custom_headers)

        url_parts = [self.base_url, path]
        url_parts.extend(suffix)
        url = "/".join(url_parts)

        if "sfamid" not in params and self.name in self.sfamid_required:
            raise ValueError("You MUST specify an SFAMID!")

        # Params should be unique
        params = dict(params)

        # Create and execute request
        if request_method == "GET":
            response = requests.get(
                url,
                params=params,
                headers=headers,
                proxies=self.proxies,
                verify=Config().verify_https,
            )
        else:
            request_type = {
                "POST": requests.post,
                "DELETE": requests.delete,
                "PUT": requests.put,
                "PATCH": requests.patch,
            }.get(request_method.upper(), requests.post)

            return request_type(
                url,
                data=data,
                headers=headers,
                proxies=self.proxies,
                verify=Config().verify_https,
            )

        return TraversableResponse(response, self, sfamid=params.get("sfamid"))

    def __getstate__(self):
        """Returns the internal dictionary when pickling.
        Some objects (like Basket) internally include an API object, that makes it easier for them
        to perform API related calls. When storing such an object in YASARA storage, it becomes
        necessary to pickle this instance. Because of the way __getattr__ is overloaded in this
        class, we need to define the __getstate__ function ourselves.
        """
        return self.__dict__

    def __setstate__(self, state):
        """Sets the internal dictionary when pickling."""
        self.__dict__ = state

    def links(self, links_dict, sfamid_header=None, headers=None, **params):
        """This method creates a PreparedRequests instance from the navigation dictionary that can
        found in the response from most API calls. It is called by TraversableResponse to create
        itself with links built in, but this method can be used if the links require headers or
        parameters not included in the original request. If possible, however, using
        TraversableResponses is better and will yield far cleaner code.

        Note that the responses retrieved from PreparedRequests output are 'raw' `requests`
        responses, not TraversableResponses.

        :param links_dict: The dictionary retrieved from an API response.
        :param sfamid_header: Optional, but probably necessary. The superfamily ID of the system you
                              are accessing.
        :param headers: Optional, any headers you'd like to pass with these requests.
        :param **params: Optional, any GET parameters you'd like to pass with these requests.
        :returns: A filled PreparedRequests instance, equipped with all the methods from the _links.
        :rtype: PreparedRequests()
        """

        if headers is None:
            headers = self.auth.get_request_headers()
        else:
            headers.update(self.auth.get_request_headers())

        if sfamid_header is not None:
            params["sfamid"] = sfamid_header

        if "sfamid" not in params and self.name in self.sfamid_required:
            raise ValueError("You MUST specify an SFAMID!")

        prepared_requests = PreparedRequests(self.proxies)
        for link, url in links_dict.items():
            try:
                if isinstance(url, list):
                    prepared_requests.add(
                        link, url[0]["href"], headers=headers, **params
                    )
                if isinstance(url, dict):
                    prepared_requests.add(link, url["href"], headers=headers, **params)
            except TypeError:
                warnings.warn("Tried to add an invalid link: %s" % url)

        return prepared_requests

    def __getattr__(self, name):
        return partial(self, name)


class TraversableResponse(object):
    """This class provides a wrapper for the response object returned by API(). This wrapper adds
    functionality to the response by allowing access to navigation links supplied by the API. Ex:

    .. code-block:: python

        response = API().protein(sfamid_header="NR_demo_2008")  # Provides a TraversableResponse
        response.next()  # Gets the next page of the results
        response.last()  # Gets the last page of the results
        response.next().next()  # Chains the next calls
        response.follow(["next"])  # The same as `response.next()`, but with strings
        response.follow(["next", "next"])  # The same as `response.next().next()`, but with strings
        response.follow(["proteins.2", "protein"])  # Returns the page of the third protein

    These are just some examples, more can be found in tests/test_API.py.

    If one of the links in a path cannot be reached, the returned error will look like this:

    .. code-block:: python

        ValueError: The path you specified cannot be followed
        ['self', 'wrong-link', 'nope']
        ~~~~~~~~~~~~~^

    This informs you of the link that yielded the error, that is: page *self* did not include link
    *wrong-link*.
    """

    def __init__(self, response, api, sfamid):
        self.__response = response
        self.__api = api
        self.__sfamid = sfamid

        self.__links = self.__get_links()

        self.proxies = api.proxies

    def __get_links(self, custom_links=None):
        """This method is used to retrieve the links from a response and convert them into a
        PreparedRequests object.

        :param custom_links: Optional, pass a valid dict with urls to this function to convert it
                             into a custom PreparedRequests object. useful when the _links attribute
                             is not at the root of the response.
        :returns: A PreparedRequests object prepared with the links from the response, or the custom
                  links.
        :rtype: PreparedRequests
        """

        headers = self.__response.request.headers
        if custom_links is None:
            try:
                response_links = self.__response.json()["_links"]
            except KeyError:
                if "error" in self.__response.json():
                    if self.__response.json()["error"] == "Forbidden":
                        raise RuntimeWarning(
                            "You do not have access to this 3DM system."
                        )
                    if "Too Many Requests" in self.__response.json()["error"]:
                        raise TooManyRequestsException(self.__response.content)
                    if "ZuulException" in self.__response.json()["error"]:
                        raise RuntimeError(
                            "Our servers are temporarily not accessible for maintenance reasons. "
                            "Please try again in a minute."
                        )

                    # Big fallback error.
                    raise RuntimeError(
                        "Error occurred: %s. URL: %s Content: %s"
                        % (
                            self.__response.json()["error"],
                            self.__response.request.url,
                            self.__response.json(),
                        )
                    )
                else:
                    response_links = {}
            except ValueError:  # Couldn't decode a JSON object, no links are present
                response_links = {}
            except TypeError:  # No links are present in JSON object
                response_links = {}
        else:
            response_links = custom_links

        links = self.__api.links(
            response_links,
            headers=headers,
            sfamid=self.__sfamid,  # always include the SFAMID
        )
        return links

    @staticmethod
    def __throw_path_error(path, link, message="no reason was supplied."):
        """Throws a ValueError with a custom message and a nicely formatted list, showing which
        link in the path caused the error. Example:

            ['self', 'wrong-link', 'nope']
            ~~~~~~~~~~~~~~^

        :param path: List, the path that was being walked
        :param link: The index of the link where the error occurred.
        :param message: Optional, more information for the user. Defaults to:
                        "no reason was supplied."
        :raises: ValueError
        """
        offset = len("', ''".join(path[0:link])) + 3  # +3 for brackets and quotes
        offset += (
            len(path[link]) // 2 + 1
        )  # Point the indicator to the middle of the link
        indicator = ("~" * offset) + "^"

        raise ValueError(
            "The path you specified cannot be followed, %s \n%s\n%s"
            % (message, path, indicator)
        )

    def _parse_dot_link(self, link, path, l_n):
        """If we have a response that contains a dot, we need to inspect the response contents
        and traverse it according to the link.

        :param link: A link string that includes dot notation.
        :param path: The full path
        :param l_n: The link number
        :returns: The contents of the final piece of the link.
        """

        level = "_embedded"
        current_level = self.__response.json()[level]
        for key in link.split("."):
            if isinstance(current_level, dict):
                try:
                    current_level = current_level[key]
                except KeyError:
                    message = "'%s' is not a valid key." % key
                    self.__throw_path_error(path, l_n, message=message)
            else:
                try:
                    key = int(key)
                except ValueError:
                    message = "'%s' is not a valid index." % key
                    self.__throw_path_error(path, l_n, message=message)

                try:
                    current_level = current_level[key]
                except IndexError:
                    message = "Index %s does not exist." % key
                    self.__throw_path_error(path, l_n, message=message)

        return current_level

    def follow(self, path):
        """Follows a path across the API. This can be used to dynamically string together
        different links. Examples can be found in tests/test_API.py.

        :param path: A list of strings containing different actions (next/prev), attributes (genes,
                     proteinSubsets) or page elements (aminoAcids.n).
        :returns: This TraversableResponse object, at the final page in the path.
        :rtype: TraversableResponse
        """
        for l_n, link in enumerate(path):
            if link in self.__links:
                self.__response = getattr(self.__links, link)()
                self.__links = self.__get_links()
            elif "." in link:
                final_level = self._parse_dot_link(link, path, l_n)
                self.__links = self.__get_links(custom_links=final_level["_links"])
            else:
                message = (
                    "'%s' could not be found in the document or in the links." % link
                )
                self.__throw_path_error(path, l_n, message=message)

        return self

    def __call__(self):
        return self

    def __getattr__(self, name):
        if name in self.__links:
            return TraversableResponse(
                getattr(self.__links, name)(), self.__api, self.__sfamid
            )
        else:
            return getattr(self.__response, name)


class PreparedRequests(object):
    """This class allows developers to easily prepare a number of URLs for requesting, with names.
    The class is called in API.links() to prepare requests for navigating between API pages."""

    def __init__(self, proxies=None):
        self._links = {}
        if proxies is None:
            self.proxies = {}
        else:
            self.proxies = proxies

    def __contains__(self, key):
        return key in self._links

    def __str__(self):
        """Returns a list of the prepared links, handy for debugging."""
        prepared_links = "\n".join(
            [" - %s ~> %s" % (name, prop["url"]) for name, prop in self._links.items()]
        )
        return "The following links are prepared:\n" + prepared_links

    def add(self, name, url, headers=None, **params):
        """Ads a link to the internal dictionary, with appropiate parameters.

        :param name: The link name, for example `next` or `prev`.
        :param url: The full URL that will be requested.
        :param headers: Optional, any headers that will need to be sent with with the request.
        :param params: Optional, any parameters that will need to be sent with with the request.
        """
        properties = {"url": url}

        if headers is not None:
            properties["headers"] = headers
        if params is not None:
            properties["params"] = params

        self._links[name] = properties

    @staticmethod
    def _get_params_from_url(url):
        """Returns the GET parameters from a URL as a dictionary."""
        parsed = urlparse(url)
        return dict(parse_qs(parsed.query))

    def __call__(self, name):
        """Sends out the request that matches the given name. Raises a ValueError if the name is
        not found or valid. Raises a RuntimeError if no SFAMID was specified and the API requires
        the ID to function. This should never happen unless a PreparedRequests object is created or
        added to manually.

        :param name: The name of the request, for example: `next` or `prev`
        :returns: The response of a `requests` request.
        :rtype: requests.Request
        :raises: ValueError, RuntimeError
        """
        link = self._links[name]
        url = link["url"]

        url_params = self._get_params_from_url(url)
        params = link.get("params", {})
        params.update(url_params)

        # Strip old parameters
        url = url.split("?", 1)[0]

        response = requests.get(
            url,
            headers=link.get("headers", {}),
            params=params,
            proxies=self.proxies,
            verify=Config().verify_https,
        )

        if response.status_code == 500 and "SFAMID" not in link.get("headers", {}):
            raise RuntimeError(
                "SFAMID is not being sent with request '%s'. Do not create your own "
                "PreparedRequests without the proper headers. Heed this error lest @ttebeek is "
                "convered in error emails." % name
            )

        if response.status_code == 500 and "ZuulException" in response.content:
            print("ZuulException: ", response.content)
            raise RuntimeWarning(
                "The server is currently down for maintenance, please try again in 15 minutes."
            )

        return response

    def __getattr__(self, name):
        if name in self._links:
            return partial(self, name)
        raise AttributeError("This link action is not available: '%s'" % name)
