# pylint: disable=R0902,R0913,C0103,W0201
""" Communicates with the API to retrieve a numbering scheme for, among others, converting
PDB positions to 3D numbers and back. """
from ...shared.vendors.six import string_types


class SaveableAlignmentWarning(RuntimeWarning):
    """A warning related to alignment or families that can be saved via user intervention."""

    pass


class NumberingScheme(object):
    """This class loads a numbering scheme from the API, indexes on each of the three numbering
    systems and provides cheap methods to convert between the numbering schemes."""

    def __init__(self, api, data_api, pdb, family, sfamid, accession=None):
        """Initializes the Numbering Scheme class and uses the parameters to immediately download
        the numbering scheme and index it.

        :param api: A logged in API instance.
        :param data_api: A logged in data API instance.
        :param pdb: A PDB number to retrieve.
        :param family: The family identifier for the NumberingScheme you need. If your superfamily
                       does not have separate families (NR_demo_2008), use 1.
        :param sfamid: The superfamily ID.
        :param accession: (Optional) The accession ID of a protein. When this is specified, the
                          numbering scheme for a model will be retrieved. No custom numbering
                          schemes are available for models.
        :raises: ValueError if no numbering scheme can be found.
        """
        self.__pdb = pdb
        self.numbering_scheme = self.retrieve_numbering_scheme(
            api, pdb, family, sfamid, accession=accession
        )
        self._check_num_scheme_integrity()
        self.create_indexed_schemes()
        self._init_aliases()

        if accession is not None:
            self._init_model_scheme()
        else:
            self._init_normal_scheme(data_api, family, sfamid)

    def _check_num_scheme_integrity(self):
        assert isinstance(self.numbering_scheme, list), (
            "Incorrect numbering system response: %s" % self.numbering_scheme
        )
        assert isinstance(self.numbering_scheme[0], dict), (
            "Incorrect numbering system response: %s" % self.numbering_scheme
        )

    def _init_normal_scheme(self, data_api, family, sfamid):
        """Initializes a normal, non-model NumberingScheme."""
        custom_numbering_schemes = self.retrieve_custom_numbering_schemes(
            data_api, family, sfamid
        )
        (
            self.custom_numbering_scheme,
            self.numbering_scheme_nice_names,
        ) = custom_numbering_schemes
        self.create_custom_indexed_schemes()

    def _init_model_scheme(self):
        """Initializes a model NumberingScheme. Models do not support custom numbering schemes."""
        self.numbering_scheme_nice_names = {
            1: "3DM Alignment numbering",
            "align": "3DM Alignment numbering",
            2: "Residue numbering",
            "res": "Residue numbering",
            3: "PDB numbering",
            "pdb": "PDB numbering",
        }
        self.custom_numbering_scheme, self.custom_names = [], []
        self.create_custom_indexed_schemes()

    def _init_aliases(self):
        """Sets up a number of aliases for number based renumbering."""
        self.align_to_1 = lambda x: x
        self.align_to_2 = self.align_to_res
        self.align_to_3 = self.align_to_pdb

        self.res_to_1 = self.res_to_align
        self.res_to_2 = lambda x: x
        self.res_to_3 = self.res_to_pdb

    @staticmethod
    def list_numbering_schemes_for_sfamid(api, family, sfamid):
        """Returns the names of the numbering schemes available for a specific SFAMID. This includes
        only the names, and no side-effects occur from this method call. No indexes are made.

        :param api: A logged in API instance.
        :param family: The family identifier for the NumberingScheme you need. If you superfamily
                       does not have separate families (NR_demo_2008), use 1.
        :param sfamid: The superfamily ID.
        """
        numbering_schemes = api.numberingSchemes(
            suffix=["search", "findDistinctByCoreNumbersAlnFamilyFamilyId"],
            sfamid_header=sfamid,
            familyid=family,
        )
        names = []
        for scheme in numbering_schemes.json()["_embedded"]["numberingSchemes"]:
            names.append(scheme["schemename"])

        return names

    @staticmethod
    def retrieve_custom_numbering_schemes(api, family, sfamid):
        """Retrieves *custom* numbering schemes, that is, a numbering scheme outside the normal
        residue/alignment/pdb set, from the API and returns it as a Python list.

        :param api: A logged in API instance.
        :param family: The family identifier for the NumberingScheme you need. If you superfamily
                       does not have separate families (NR_demo_2008), use 1.
        :param sfamid: The superfamily ID.
        :returns: A python list filled with dictionaries, containing the mapping between different
                  numbering systems and a dictionary containing nice names
        :raises: ValueError if no numbering schemes can be found.
        """
        numbering_schemes = api.numberingSchemes(
            suffix=["search", "findDistinctByCoreNumbersAlnFamilyFamilyId"],
            sfamid_header=sfamid,
            familyid=family,
        )

        numbering_scheme_nice_names = {
            1: "3DM Alignment numbering",
            "align": "3DM Alignment numbering",
            2: "Residue numbering",
            "res": "Residue numbering",
            3: "PDB numbering",
            "pdb": "PDB numbering",
        }

        raw_numbering_scheme = {}
        for scheme in numbering_schemes.json()["_embedded"]["numberingSchemes"]:
            _id = int(scheme["schemeid"])
            # Offset custom numbering scheme names by 10, conforming to what the visualize service
            # sends us in structure.
            offset_id = _id + 10
            name = scheme["schemename"]

            raw_scheme = api.numberingSchemes(
                suffix=[str(_id), "coreNumbers"], sfamid_header=sfamid
            ).json()["_embedded"]["coreNumbers"]

            numbering_scheme_nice_names[offset_id] = name
            numbering_scheme_nice_names[str(offset_id)] = name
            numbering_scheme_nice_names[name] = name

            for core_number in raw_scheme:
                al_pos = core_number["alignmentposition"]
                core_num = core_number["corenumber"]
                try:
                    raw_numbering_scheme[al_pos][name] = core_num
                    raw_numbering_scheme[al_pos][offset_id] = core_num
                    raw_numbering_scheme[al_pos][str(offset_id)] = core_num
                except KeyError:
                    raw_numbering_scheme[al_pos] = {
                        name: core_num,
                        offset_id: core_num,
                        str(offset_id): core_num,
                        "align": al_pos,
                    }

        return raw_numbering_scheme.values(), numbering_scheme_nice_names

    @staticmethod
    def retrieve_numbering_scheme(api, pdb, family, sfamid, accession=None):
        """Retrieves a numbering scheme from the API and returns it as a Python list.

        :param api: A logged in API instance.
        :param pdb: The PDB ID would like to retrieve the numbering scheme for. If `accession` is
                    specified, this functions as the *template*.
        :param family: The family identifier for the NumberingScheme you need. If your superfamily
                       does not have separate families (NR_demo_2008), use 1.
        :param sfamid: The superfamily ID.
        :param accession: (Optional) The accession ID of a protein. When this is specified, the
                          numbering scheme for a model will be retrieved.
        :returns: A python list filled with dictionaries, containing the mapping between different
                  numbering systems.
        :raises: ValueError if no numbering scheme can be found.
        """
        if pdb.endswith("_prot") or pdb.endswith("_drug"):
            pdb = pdb.rsplit("_", 1)[0]

        suffix = [str(family), str(pdb)]
        if accession is not None:
            # If an accession number is specified, the numbering scheme will be for a model.
            suffix.append(accession)

        numbering_scheme = api.pdbnumbering(suffix=suffix, sfamid_header=sfamid)
        numbering_scheme = numbering_scheme.json()

        if len(numbering_scheme) == 0:
            raise ValueError(
                "Numbering scheme was not available for this PDB: %s in system %s. "
                "Please contact support@bio-prodict.nl with this message."
                % (pdb, sfamid)
            )

        return numbering_scheme

    def create_indexed_schemes(self):
        """Creates several dictionaries, eached indexed with one of the numbering schemes, to
        provide fast access to any value."""
        self.numbering_scheme_pdb_index = {
            num["pdbnumber"]: num
            for num in self.numbering_scheme
            if num["pdbnumber"] is not None
        }
        self.numbering_scheme_align_index = {
            num["alignmentposition"]: num
            for num in self.numbering_scheme
            if num["alignmentposition"] is not None
        }
        self.numbering_scheme_res_index = {
            num["residuenumber"]: num
            for num in self.numbering_scheme
            if num["residuenumber"] is not None
        }

        if not self.numbering_scheme_align_index:
            raise SaveableAlignmentWarning(
                "When retrieving the numbering scheme for %s, no valid alignment numbering was "
                "found. You may need to select a different alignment and reload the structure "
                "to continue." % self.__pdb
            )

    def create_custom_indexed_schemes(self):
        """Creates two conversion schemes for fast resolution of alignment <-> custom numbering
        and residue numbering -> custom numbering schemes. Also stores the names of all the
        custom numbering schemes in the object. Note that alignment/custom numbering is
        bidirectional, while residue to custom can only be performed in that specific direction."""
        custom_names = []
        for num_scheme in self.custom_numbering_scheme:
            custom_names += num_scheme.keys()
            custom_names += [str(nsk) for nsk in num_scheme.keys()]

        self.custom_names = list(set(custom_names))
        try:
            del self.custom_names[self.custom_names.index("align")]
        except ValueError:
            # Align was already removed.
            pass

        # Align -> Custom
        self.numbering_scheme_align_to_custom_index = {
            num["align"]: num
            for num in self.custom_numbering_scheme
            if num["align"] is not None
        }

        # Custom -> Align
        self.numbering_scheme_custom_to_align_index = {}
        for custom_name in self.custom_names:
            self.numbering_scheme_custom_to_align_index[custom_name] = {}

            for num in self.custom_numbering_scheme:
                self.numbering_scheme_custom_to_align_index[custom_name][
                    num.get(custom_name)
                ] = num["align"]

        # Residue number -> Custom
        align_index = self.numbering_scheme_align_index
        self.numbering_scheme_res_to_custom_index = {}
        for num in self.custom_numbering_scheme:
            if num["align"] is None:
                continue
            try:
                res_number = int(align_index[num["align"]]["residuenumber"])
            except KeyError:
                # Could not convert an alignment number to a residue number...
                continue
            # Stores the residue number as the key and the custom dictionary as the value.
            self.numbering_scheme_res_to_custom_index[res_number] = num

    def align_to_custom(self, align_number, custom):
        """Converts an alignment number to a custom numbering type. Inputs are converted to
        strings. Requires both the number and the custom scheme name."""
        try:
            return self.numbering_scheme_align_to_custom_index[int(align_number)].get(
                custom
            )
        except KeyError:
            return self.numbering_scheme_align_to_custom_index.get(
                str(align_number), {}
            ).get(custom)

    def res_to_custom(self, residue_number, custom):
        """Converts a residue number to a custom numbering type. Inputs are converted to
        integers. Requires both the number and the custom scheme name."""
        return self.numbering_scheme_res_to_custom_index[int(residue_number)].get(
            custom
        )

    def custom_to_align(self, custom_number, custom):
        """Converts a custom number to the corresponding alignment number. Requires both the number
        and the custom scheme name."""
        try:
            custom_scheme = self.numbering_scheme_custom_to_align_index[custom]
        except KeyError:
            return None
        try:
            return custom_scheme[int(custom_number)]
        except KeyError:
            return custom_scheme.get(str(custom_number))

    def pdb_to_align(self, pdb_number):
        """Converts a PDB number to a 3D number. Returns None if the position is not available.
        May return negative integers, these are valid outputs. Inputs are converted to string."""
        return self.numbering_scheme_pdb_index.get(str(pdb_number), {}).get(
            "alignmentposition"
        )

    def pdb_to_res(self, pdb_number):
        """Converts a PDB number to a residue number. Returns None if the position is not
        available. Inputs are converted to string."""
        return self.numbering_scheme_pdb_index.get(str(pdb_number), {}).get(
            "residuenumber"
        )

    def align_to_pdb(self, ali_number):
        """Converts a 3D number to a PDB number. Returns None if the position is not available.
        Accepts both strings and numerical values."""
        try:
            return self.numbering_scheme_align_index[ali_number].get("pdbnumber")
        except KeyError:
            if not isinstance(ali_number, string_types):
                return self.numbering_scheme_align_index.get(str(ali_number), {}).get(
                    "pdbnumber"
                )
            else:
                return self.numbering_scheme_align_index.get(int(ali_number), {}).get(
                    "pdbnumber"
                )

    def align_to_res(self, ali_number):
        """Converts a 3D number to a residue number. Returns None if the position is not
        available. Accepts both strings and numerical values."""

        try:
            return self.numbering_scheme_align_index[ali_number].get("residuenumber")
        except KeyError:
            return self.numbering_scheme_align_index.get(str(ali_number), {}).get(
                "residuenumber"
            )

    def res_to_align(self, res_number):
        """Converts a residue number to a 3D number. Returns None if the position is not available.
        May return negative integers, these are valid outputs. Inputs are always numerical."""
        return self.numbering_scheme_res_index.get(int(res_number), {}).get(
            "alignmentposition"
        )

    def res_to_pdb(self, res_number):
        """Converts a residue number to a pdb number. Returns None if the position is not
        available. Inputs are always numerical."""
        return self.numbering_scheme_res_index.get(int(res_number), {}).get("pdbnumber")
