""" Dealing with colors is required for all Viewers. This module can create color scales to display
properties of residues with sliding values. """
# pylint: disable=R1706
import colorsys
import math


def b16(num):
    """Converts a number from integer to base 16."""
    numerals = "0123456789abcdef"
    result = ((num == 0) and numerals[0]) or (
        b16(num // 16).lstrip(numerals[0]) + numerals[num % 16]
    )
    return result.rjust(2, "0").upper()


class Colors(object):
    """Handles color conversion and ColorScale creation. Uses HSV (Hue, Saturation, Value)
    internally, as it allows for easy color shifting."""

    @staticmethod
    def named_to_hsv(name):
        """A few basic named colors can be used."""
        try:
            return {
                "blue": (2.0 / 3.0, 1, 255),
                "red": (0, 1, 255),
                "green": (1.0 / 3.0, 1, 255),
                "yellow": (1.0 / 6.0, 1, 255),
                "cyan": (0.5, 1, 255),
                "magenta": (5.0 / 6.0, 1, 255),
            }[name]
        except KeyError:
            raise ValueError("Unknown color '%s'." % name)

    @staticmethod
    def desaturate(color):
        """Makes the color completely desatured, keeping all the original values.

        :param color: an HSV tuple.
        :returns: The desatured color.

        """
        color = list(color)
        color[2] = 0
        return tuple(color)

    @staticmethod
    def invert(color):
        """Inverts the color on the hue value, resulting in the opposite color on the spectrum.

        :param color: an HSV tuple.
        :returns: The inverted color.
        """
        color = list(color)
        color[0] += 0.5
        if color[0] > 1:
            color[0] -= 1
        return tuple(color)

    def parse_color(self, color):
        """Takes either a string or an RGB tuple and attempts to convert them to an HSV value."""
        if isinstance(color, str):
            return self.named_to_hsv(color)

        elif isinstance(color, tuple):
            return colorsys.rgb_to_hsv(*color)

    def create_scale(
        self, color_from="grey", color_to="red", _min=0, _max=1, hsv=False, loop=False
    ):
        """Creates a color scale based on two different colors and min and max values.
        For colors, this method accepts either a limited range of named colors (see `named_to_hsv`),
        any RGB color tuple or an HSV color tuple if `hsv` is set to True.

        :param color_from: The color associated with the lowest values. Defaults to grey.
        :param color_to: The color associated with the highest values. Defaults to red.
        :param _min: The lowest amount in your value range.
        :param _max: The highest amount in your value range.
        :param hsv: Optional, defaults to False. If this is set to True, color values are assumed
                    to be hsv tuples.
        :param loop: Defaults to False. Set this to True to create a color scale that loops from
                     `color_from` all the way around the color wheel to itself. `color_to` is ignored.
        :returns: A ColorScale object.
        """
        if loop:
            color = self.parse_color(color_from)
            return ColorScale(color, color, _min, _max)

        if color_from == color_to:
            # This is not a redundant exception. If "grey" is used as a color, the other color is
            # parsed first. If the other color is also grey, this will lead to problems.
            raise ValueError("Can't create a scale for colors that are the same.")

        if not hsv:
            # In HSV color space, even if color is grey, a hue is associated with it. Thus, we need
            # to know the other color first, to ensure that the hue remains constant in the scale,
            # and only saturation and value change.
            if color_from == "grey":
                color_to = self.parse_color(color_to)
                color_from = self.desaturate(color_to)
            elif color_to == "grey":
                color_from = self.parse_color(color_from)
                color_to = self.desaturate(color_from)
            else:
                color_from = self.parse_color(color_from)
                color_to = self.parse_color(color_to)

        return ColorScale(color_from, color_to, _min, _max)

    @staticmethod
    def hsv_to_hex(color, with_hash=True):
        """Converts an HSV color to a hex code.

        :param color: An HSV tuple (float, float, int)
        :param with_hash: Optional, defaults to true, whether or not to include a hash (#) in the
                          string.
        :returns: String, full RGB hex code, 7 characters (6 if with_hash = False)
        """
        red, green, blue = colorsys.hsv_to_rgb(*color)
        hex_str = ""
        if with_hash:
            hex_str = "#"
        hex_str += b16(int(red))
        hex_str += b16(int(green))
        hex_str += b16(int(blue))
        return hex_str


class ColorScale(object):
    """This class returns a hex color based on a value in a range, interpolated from two base
    colors."""

    def __init__(self, _from, _to, _min, _max, allow_spectrum_overflow=False):
        """Creates a color scale.

        :param _from: The color at the start of the scale in HSV format.
        :param _to: The color at the end of the scale in HSV format.
        :param _min: The minimum value that will be passed into the scale.
        :param _max: The maximum value that will be passed into the scale.
        :param allow_spectrum_overflow: Defaults to False. If this is set to True, the difference in
                                        hue is made absolute. This is useful when creating a
                                        color_loop, but may lead to problems when using normal color
                                        scales.

        """
        if _min > _max:
            raise ValueError(
                "Can't create a scale for inverted min/max values (%s/%s)."
                % (_min, _max)
            )
        if _min == _max:
            _min -= 1

        # Set loop to True if both colors are the same.
        self.loop = _from == _to
        self.allow_spectrum_overflow = allow_spectrum_overflow

        self._from, self._to, self.min, self.max = _from, _to, _min, _max

    def interpolate_color(
        self,
        value,
    ):
        """Interpolates the from and to colors based on a value from 0 to 1.

        :param value: Float from 0 to 1.
        :returns: A color between ColorScale._from and ColorScale._to.
        :raises: ValueError if value is outside the 0-1 range.
        """
        if value > 1 or value < 0:
            raise ValueError("Value needs to be between 0 and 1.")

        h_diff = self._to[0] - self._from[0]
        if self.allow_spectrum_overflow:
            h_diff = abs(h_diff)
        s_diff = self._to[1] - self._from[1]
        v_diff = self._to[2] - self._from[2]

        new_hsv = list(self._from)

        new_hue = self._from[0] + h_diff * value
        if new_hue > 1:  # Wrap the hue around.
            new_hue -= 1

        new_hsv[0] = new_hue
        new_hsv[1] = self._from[1] + s_diff * value
        new_hsv[2] = self._from[2] + int(math.ceil(v_diff * value))
        return new_hsv

    def interpolate_color_loop(self, value):
        """Interpolates the from color based on a value from 0 to 1. The color is based on looping
        around the spectrum. This is done by creating two new color scales, each representing one
        half of the spectrum. The second half is stretched somewhat, to make sure that the start
        end and color are not the same, resulting in more clarity.

        :param value: Float from 0 to 1.
        :returns: A color between ColorScale._from and ColorScale._from, going around the spectrum.
        :raises: ValueError if value is outside the 0-1 range.
        """
        first_half = ColorScale(self._from, Colors.invert(self._from), 0, 0.5, True)
        last_half = ColorScale(Colors.invert(self._from), self._from, 0.5, 1.2, True)
        if value < 0.5:
            return first_half(value, hsv=True)
        else:
            return last_half(value, hsv=True)

    def interpolate_value(self, value):
        """Interpolates a value using the min and max, returns a value from 0 to 1."""
        value = value - self.min
        value = value / float(self.max - self.min)
        return value

    def __call__(self, value, with_hash=True, hsv=False):
        """Returns a hex color based on the value given."""
        value = self.interpolate_value(value)
        if self.loop:
            color = self.interpolate_color_loop(value)
        else:
            color = self.interpolate_color(value)
        if hsv:
            return color
        else:
            return Colors.hsv_to_hex(color, with_hash=with_hash)
