""" This module contains utilities that can be used by all plugins in their respective
interfaces. """
# pylint: disable=W0212,R0914,W0640,W0403,R0912,R0913,R0915,C0103,R0904,C0302

import re
import os
import tarfile
import tempfile
from .Colors import Colors
from .Selection import Selection, ObjectIdentifier
from ..api.NumberingSchemes import NumberingScheme

try:
    from cStringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO


class CompositeUtilities(object):
    """This class contains versatile methods for use by all plugins, building on the basic API
    methods. Methods in this class may require an instance of the viewer that called them. This
    is to increase the amount of code that can be shared between viewers."""

    @staticmethod
    def get_valid_view_name(viewer_instance, name):
        """Takes a string and attempts to make sure that it is a valid name for a new View.
        For example, if there are no Cores views, this method will simply return Cores. If there
        already are Cores windows, a number will be appended to reflect that, yielding Cores_2.

        :param name: The name for the view you want to open.
        :returns: A valid, unique view name."""
        if hasattr(viewer_instance.storage, "views"):
            while name in viewer_instance.storage.views:
                try:
                    name, number = name.rsplit("_", 1)
                    number = int(number) + 1
                    name = "%s_%s" % (name, number)
                except (ValueError, TypeError):
                    name += "_2"

        try:
            viewer_instance.storage.views.append(name)
        except AttributeError:
            viewer_instance.storage.views = [name]
        return name

    @staticmethod
    def create_formatted_key(identifier, pretty_name):
        """Creates a nice representation of an identifier and nice name.
        Cleans any incompatible characters (non-ascii).

        :param identifier: String, an identifier of any type
        :param pretty_name: String,  Nice name for the system.
        :returns: String in the format '<pretty_name> (<sfamid>)' with non ascii chars
                  removed.
        """
        try:
            identifier = identifier.encode("ascii", "ignore").decode()
        except AttributeError:
            identifier = str(identifier)
        pretty_name = pretty_name.encode("ascii", "ignore").decode()
        n_key = "{} ({})".format(pretty_name, identifier)
        return n_key

    @staticmethod
    def load_accessible_systems(access_api, systems_db="postgres-3dm-systems"):
        """Loads a dictionary containing the systems that are accessible to the current user.
        Return these dictionary names as friendly strings for use in system listings.

        :param access_api: A logged in instance of a API object for the `access` API.
        :param systems_db: The database of systems that should be listed. Defaults to
                           postgres-3dm-systems, but may feature names like "scop_v2".
        """
        systems = access_api.public(
            suffix=["accessibleAlignmentDatabases", "cluster", systems_db]
        ).json()

        # To prevent collisions in the keys (since pretty system names are non-unique), the sfamid
        # is appended to each pretty name in brackets.
        systems_dict = {}
        for system in systems:
            sfamid = system["sfamid"].replace("_", "-")
            pretty_name = system["nicename"].replace("_", "-")
            n_key = CompositeUtilities.create_formatted_key(sfamid, pretty_name)
            systems_dict[n_key] = system["sfamid"]

        return systems_dict

    @staticmethod
    def load_alignment_families(viewer_instance, sfamid=None):
        """Loads a dictionary containing the alignment families that are accessible to the
        current user, in the current superfamily. Raises an eror if no sfamid could be found."""
        api = viewer_instance._get_api("persistence3dmdata")
        if sfamid is None:
            sfamid = CompositeUtilities.get_sfamid(viewer_instance)
        result = api.alnFamilies(sfamid_header=sfamid).json()["_embedded"][
            "alnFamilies"
        ]
        families = {fam["familyname"]: int(fam["familyId"]) for fam in result}
        return families

    @staticmethod
    def load_alignment_subsets(viewer_instance, sfamid=None):
        """Loads a dictionary containing the alignment subsets that are accessible to the
        current user, in the current superfamily. Raises an eror if no sfamid could be found."""
        api = viewer_instance._get_api()
        if sfamid is None:
            sfamid = CompositeUtilities.get_sfamid(viewer_instance)

        result = api.subsets(sfamid_header=sfamid).json()["_embedded"]["subsets"]
        families = {fam["subsetName"]: int(fam["subsetId"]) for fam in result}
        return families

    @staticmethod
    def load_pdbs_for_system(viewer_instance, sfamid, api, chain_request_size=1000):
        """Loads all the available PDB chains for a given system from the API. At most 1000 PDBs
        are sent back each request, so if the total amount of elements exceeds 1000, additional
        requests are sent to the API. This call may be cached.

        :param sfamid: The superfamily ID of the system.
        :param api: A logged in API instance.
        :param chain_request_size: The number of PDBs that should be retrieved with each call.
                                   Defaults to 1000 (max).
        :returns: An alphabetically sorted list of PDB codes with chain.
        :raises RuntimeWarning: When no PDBs are found, but the request did go through.
        """
        try:
            return viewer_instance.storage.pdb_list[sfamid]
        except (AttributeError, KeyError):
            pass

        response = api.pdbChains(
            suffix=["search", "findByTransformationfile"],
            sfamid_header=sfamid,
            size=chain_request_size,
            transformationfile=True,
        )

        pdbs = [
            pdbchain["pdbcode"]
            for pdbchain in response.json()["_embedded"]["pdbChains"]
        ]

        while 1:
            try:
                response = response.next()
            except (AttributeError, TypeError):
                break
            pdbs.extend(
                [
                    pdbchain["pdbcode"]
                    for pdbchain in response.json()["_embedded"]["pdbChains"]
                ]
            )

        if not pdbs:
            raise RuntimeWarning(
                "The server did not return any PDB identifiers in this system, this may be a "
                "draft system, which does not yet contain data."
            )

        pdbs.sort()

        try:
            viewer_instance.storage.pdb_list[sfamid] = pdbs
        except AttributeError:
            viewer_instance.storage.pdb_list = {sfamid: pdbs}

        return pdbs

    @staticmethod
    def load_models_for_system(sfamid, api, family):
        """Loads all the available models identifiers for a given system from the API. At most 1000
        model identifiers are sent back each request, so if the total amount of elements exceeds
        1000, additional requests are sent to the API.

        :param sfamid: The superfamily ID of the system.
        :param api: A logged in API instance.
        :param family: A family identifier.
        :returns: An alphabetically sorted list of model identifiers.
        """
        response = api.visualizedata(
            suffix=["models", str(family)], sfamid_header=sfamid
        )
        model_ids = response.json()
        return sorted(model_ids)

    @staticmethod
    def load_model(viewer_instance, model_id, sfamid, familyid):
        """Loads a model structure from 3DM using the load_pdb method and a special model
        endpoint.

        :param model_id: The model ID as returned by the 3DM server: <accession>_<template>. Will
                         raise a RuntimeError if the model ID is not valid.
        :param sfamid: Superfamily ID for the model you would like to load.
        :param familyid: Family ID for the model you would like to load.
        """
        api = viewer_instance._get_api("molviewer")

        try:
            accession, template = model_id.rsplit("_", 1)
        except ValueError:
            raise RuntimeError("Selected an invalid model identifier: '%s'" % model_id)

        viewer_instance.load_pdb(
            model_id,
            sfamid,
            api_endpoint=api.hmodel,
            # API suffix
            suffix=["model", str(familyid), template, accession],
            auto_renumber=False,
            # We want to renumber this ourselves
            numbering="RESIDUE",
            # Accept header for plain text PDB
            custom_headers={"Accept": "text/plain"},
        )

        # We will receive the model in residue numbering format and renumber it to align ourselves.
        accession, pdbid = model_id.rsplit("_", 1)
        num_scheme_nice_name = CompositeUtilities.renumber_pdb(
            viewer_instance,
            pdbid,
            sfamid,
            familyid,
            accession=accession,
            target_numbering="align",
        )
        return num_scheme_nice_name

    @staticmethod
    def load_available_hotspot_baskets(viewer_instance):
        """Loads the baskets that are available in the current situation and returns them."""

        baskets_api = viewer_instance._get_baskets_api()
        group_ids = CompositeUtilities.get_user_groups(viewer_instance)
        baskets = baskets_api.list_baskets(
            sfamid=CompositeUtilities.get_sfamid(viewer_instance),
            familyid=CompositeUtilities.get_familyid(viewer_instance),
            basket_type="hotspot",
        )

        if len(group_ids) == 0:
            raise RuntimeError(
                "You are not in any groups. Please contact your administrator."
            )

        for group_id in group_ids:
            baskets += baskets_api.list_baskets(
                shared_with=group_id,
                sfamid=CompositeUtilities.get_sfamid(viewer_instance),
                familyid=CompositeUtilities.get_familyid(viewer_instance),
                basket_type="hotspot",
            )
            
        unique_basket_ids = []
        unique_baskets = []
        for basket in baskets:
            if basket.id not in unique_baskets:
                unique_baskets.append(basket)
                unique_basket_ids.append(basket.id)

        return unique_baskets

    @staticmethod
    def convert_residue_number_from_scene_to_align(
        viewer_instance, object_id, residue_number
    ):
        """Uses the viewer instance to determine the current numbering of the object where a
        a residue number was selected, and converts the residue from this numbering to alignment
        numbering. Also supports custom numberings. Will simply return input residue number if
        alignment numbering is detected, without any processing.

        :param viewer_instance: An instance of the viewer object for caching and API calls.
        :param object_id: The object name (commonly a PDB ID) where the residue was selected.
        :param residue_number: The residue number.
        :returns: The number of the residue that was input, according to alignment number.
        """

        # First, if the current numbering scheme is 'align' we can skip any processing at all.
        suffixes = ["", "_prot", "chain", "_full"]
        current_numbering_scheme_name = None
        for suffix in suffixes:
            try:
                if len(suffix) > 0:
                    name = object_id[: -len(suffix)]
                else:
                    name = object_id
                current_numbering_scheme_name = (
                    viewer_instance.storage.numbering_schemes[name]
                )
                break
            except KeyError:
                continue

        if current_numbering_scheme_name is None:
            raise RuntimeError(
                "Invalid object name: '%s'. Could not convert residue number."
                % object_id
            )

        if current_numbering_scheme_name in [1, "1", "align"]:
            return residue_number

        api = viewer_instance._get_api("persistence3dm")
        data_api = viewer_instance._get_api("persistence3dmdata")

        accession, pdb = CompositeUtilities.split_structure_accession(object_id)

        num_scheme = NumberingScheme(
            api,
            data_api,
            pdb,
            CompositeUtilities.get_familyid(viewer_instance),
            CompositeUtilities.get_sfamid(viewer_instance),
            accession=accession,
        )

        if current_numbering_scheme_name == "res":
            return num_scheme.res_to_align(residue_number)
        elif current_numbering_scheme_name == "pdb":
            return num_scheme.pdb_to_align(residue_number)
        else:
            return num_scheme.custom_to_align(
                residue_number, current_numbering_scheme_name
            )

    @staticmethod
    def get_amino_acid_distribution_data(viewer_instance, residue_number):
        """Retrieves the superfamily data that is needed to calculate to the properties on the
        distribition of amino acids.

        :param viewer_instance: An instance of the viewer object for caching and API calls.
        :residue_number: The number of the residue *as it currently exists in your scene*.
        :returns: A dictionary containing the absolute amount of sequences where each amino acid on
                  the given position occurs, with amino acid one-letter codes as keys and counts as
                  values.
        """
        data = CompositeUtilities.download_superfamily_data(viewer_instance)

        conservations = {
            str(pos["alignmentPosition"]): pos["conservation"] for pos in data
        }
        conservation = conservations.get(str(residue_number))

        if conservation is None:
            raise RuntimeWarning(
                "Could not get conservation for this residue: '%s'. "
                "Have you selected a residue in a core? " % (residue_number)
            )

        conservation = conservation.strip(";")

        cons_values = {
            c.split(":")[0]: int(c.split(":")[1]) for c in conservation.split(";")
        }

        return cons_values

    @staticmethod
    def get_amino_acid_distribution_details(viewer_instance, residue_number):
        """Returns a dictionary containing metadata about the amino acid distribution.

        :param viewer_instance: An instance of the viewer object for caching and API calls.
        :param residue_number: The number of the residue *as it currently exists in your scene*.
        :returns: A dictionary containing the some meta data on aa distribution.
        """
        cons_values = CompositeUtilities.get_amino_acid_distribution_data(
            viewer_instance, residue_number
        )
        total = sum(cons_values.values())
        metadata = {
            "total": total,
            "total_no_gaps": total - cons_values["-"],
            "total_gaps": cons_values["-"],
        }

        return metadata

    @staticmethod
    def get_amino_acid_distribution(viewer_instance, residue_number):
        """Uses the superfamily data to determine the conservation numbers for each amino acid
        possible on a given position.

        The "true" amino acid percentage calculation has some quirkiness to it: The gaps percentage
        is calculated with the full total number of sequences, whereas the amino acid percentages
        are calculated with the total number of non-gap sequences. The total of all these
        percentages will thus exceed 100%. For the full discussion on the merits of this decision,
        consult `Bio-Prodict/bioprodict-persistence#47` in source control.

        :param viewer_instance: An instance of the viewer object for caching and API calls.
        :param residue_number: The number of the residue *as it currently exists in your scene*.
        :returns: A dictionary containing the conservation values for each amino acid on the given
                  position, with amino acid one-letter codes as keys and conservation as values.
        """

        cons_values = CompositeUtilities.get_amino_acid_distribution_data(
            viewer_instance, residue_number
        )

        total = sum(cons_values.values())
        total_no_gaps = total - cons_values["-"]

        cons_percentage = {
            k: float(v) / total_no_gaps for k, v in cons_values.items() if k != "-"
        }
        cons_percentage["-"] = cons_values["-"] / float(total)
        return cons_percentage

    @staticmethod
    def download_superfamily_data(viewer_instance):
        """Loads superfamily data based on the SFAMID and a subset id. Will cache the data inside
        the viewer's Storage object.

        The superfamily data consists of a list containing a large amount of information points
        on each position (3DM numbering) of the superfamily. Each item in the list is a dictionary
        containing information like the entropy, Hydrophobicity, and bvalue of a residue.
        The position is included in each dictionary under the `alignmentPosition` key.

        :param viewer_instance: An instance of the viewer object for caching and API calls.
        :returns: The loaded superfamily data (May be cached)
        """
        sfamid = CompositeUtilities.get_sfamid(viewer_instance)
        family_id = CompositeUtilities.get_familyid(viewer_instance)
        subset_id = CompositeUtilities.get_subset_id(viewer_instance)

        try:
            return viewer_instance.storage.superfamily_data[
                (sfamid, family_id, subset_id)
            ]
        except (AttributeError, KeyError):
            api = viewer_instance._get_api("persistence3dmdata")
            data = []

            response = api.positionInfos(
                suffix=["search", "findByAlnFamilyFamilyIdAndSubsetSubsetId"],
                sfamid_header=sfamid,
                familyid=family_id,
                subsetid=subset_id,
                sort="alignmentPosition",
                size=10,
            )

            # Get all the pages
            while 1:
                position_infos = response.json()["_embedded"]["positionInfos"]
                data += position_infos
                print([pos["alignmentPosition"] for pos in position_infos])
                try:
                    response = response.next()
                except (AttributeError, TypeError) as e:
                    print(e)
                    break

            data += response.json()["_embedded"]["positionInfos"]

        try:
            viewer_instance.storage.superfamily_data[
                (sfamid, family_id, subset_id)
            ] = data
        except AttributeError:
            viewer_instance.storage.superfamily_data = {
                (sfamid, family_id, subset_id): data
            }

        return data

    @staticmethod
    def determine_numbering_method(num_scheme, target_numbering):
        """Determines and returns the appropriate numbering method that for a specific scheme and
        target numbering.

        :param num_scheme: A numbering scheme instance.
        :param target_numbering: String, the type of numbering you would like to get. For example,
                                 align, pdb or a custom scheme name/identifier (must be integer).
        :returns: A function that takes a residue number as an argument and converts it into the
                  target numbering type.
        """

        # Handle custom conversion schemes
        if target_numbering in getattr(num_scheme, "custom_names", []):

            def res_to_custom(num):
                """Converts a number to the target custom numbering."""
                return num_scheme.res_to_custom(num, target_numbering)

            # Preserve the function name
            renumber_method = res_to_custom
        # Handle custom normal schemes
        else:
            renumber_method_name = "res_to_%s" % (target_numbering)
            renumber_method = getattr(num_scheme, renumber_method_name, None)

        if renumber_method is None:
            raise RuntimeError(
                "Attempted to find a numbering method with target numbering '%s'. "
                "This not a valid numbering scheme, options are: %s"
                % (target_numbering, getattr(num_scheme, "custom_names", []))
            )

        return renumber_method

    @staticmethod
    def _renumber_residues(residues, renumber_method, obj, properties=None):
        """This method performs the renumbering residue steps.

        :param position: A list of residues
        :param renumbering_method: The desired renumbering method
        :param obj: The name of the object where the residues reside.
        :param properties: The properties list
        :returns: A list of renumbered residues, a list of selectors and a list of properties.
        """
        renumbered_residues, selectors, new_properties = [], [], []
        for pos, position in enumerate(residues):
            res_number = renumber_method(position)
            if res_number is None:
                continue
            res_number_raw = res_number
            if "." in str(res_number):  # Deal with GPCR numbering
                res_number = "%s%s" % CompositeUtilities.convert_gpcr_number(res_number)

            if properties is not None:
                prop = properties[pos] if properties[pos] is not None else 0
                new_properties.append(prop)

            renumbered_residues.append(res_number_raw)
            selectors.append(Selection(numbers=res_number, in_object=obj))

        return renumbered_residues, selectors, new_properties

    @staticmethod
    def load_pdb_batch(
        viewer_instance, pdb_list, sfamid=None, specifier="chain", auto_renumber=True
    ):
        """Downloads a batch of  PDB files from the 3DM API and loads it into the viewer.

        :param viewer_instance: An instance of the calling Viewer class.
        :param pdb_list: A list of PDB ids.
        :param sfamid: The superfamily ID.
        :param specifier: The specifier argument for the API. This allows you to filter what parts
                          of the structure are to be retrieved. "chain" is the default, results in
                          the entire chain. "amino" only downloads the protein, without any ligands.
                          "hetatam" only loads ligands.
        :param auto_renumber: Optional, whether or not the PDB should be automatically renumbered
                              to the align numbering scheme. Defaults to True.
        """

        assert isinstance(pdb_list, list), "pdb_list must be a list but is a %s" % type(
            pdb_list
        )

        if sfamid is None:
            sfamid = CompositeUtilities.get_sfamid(viewer_instance)

        pdbids = ",".join(list(set(pdb_list)))

        api = viewer_instance._get_api("persistence3dm")
        response = api.pdbchains(
            sfamid_header=sfamid,
            pdbids=pdbids,
            specifier=specifier,
        )

        CompositeUtilities.handle_pdb_errors(response, sfamid, pdbids)

        tarball = response.content
        tar_handle = StringIO()
        tar_handle.write(tarball)
        tar_handle.seek(0)
        archive = tarfile.TarFile(fileobj=tar_handle)

        archive.extractall(tempfile.tempdir)
        members = archive.getmembers()

        loaded_pdbs = set()

        for member in members:
            member_pdb = member.name.split(".")[0]
            assert member_pdb in pdb_list, "API returned a PDB that was not requested."

            loaded_pdbs.add(member_pdb)

            abs_pdb_path = os.path.abspath(os.path.join(tempfile.tempdir, member.name))

            with open(abs_pdb_path) as pdb_file:
                if not CompositeUtilities.is_valid_pdb(pdb_file.read()):
                    raise ValueError(
                        "The content of the PDB %s is invalid." % member_pdb
                    )

            viewer_instance.load_pdb_file_into_viewer(
                abs_pdb_path, member_pdb, sfamid, auto_renumber
            )

        if len(loaded_pdbs) < len(pdb_list):
            raise ValueError(
                "Was unable to load the following PDBs: %s"
                % ", ".join(list(set(pdb_list).difference(loaded_pdbs)))
            )

    @staticmethod
    def download_pdb_file(
        viewer_instance,
        pdbid,
        sfamid,
        specifier="chain",
        api_endpoint=None,
        *args,
        **kwargs
    ):
        """Downloads a PDB file from the 3DM API and returns the filename.

        :param viewer_instance: An instance of the calling Viewer class.
        :param sfamid: The superfamily ID.
        :param pdbid: The PDB ID.
        :param specifier: The specifier argument for the API. This allows you to filter what parts
                          of the structure are to be retrieved. "chain" is the default, results in
                          the entire chain. "amino" only downloads the protein, without any ligands.
                          "hetatam" only loads ligands.
        :param api_endpoint: Optional, you can use this to use a custom api endpoint to load a PDB.
                             This argument should be a method of the API class that is logged in.
                             All other args and kwargs are pased to the api endpoint call.
        """

        if sfamid is None:
            sfamid = CompositeUtilities.get_sfamid(viewer_instance)

        if api_endpoint is None:
            api = viewer_instance._get_api(
                "persistence3dm"
            )  # Get the plain persistence 3DM API.
            response = api.pdbchain(
                sfamid_header=sfamid, suffix=[pdbid], specifier=specifier
            )
        else:
            response = api_endpoint(sfamid_header=sfamid, *args, **kwargs)

        CompositeUtilities.handle_pdb_errors(response, sfamid, pdbid)

        if not CompositeUtilities.is_valid_pdb(response.text):
            raise ValueError("The content of the PDB %s is invalid." % pdbid)

        pdbfile = tempfile.NamedTemporaryFile(delete=False, suffix=".%s.pdb" % pdbid)
        pdbfile.write(response.content)
        pdbfile.close()  # Close the PDB file so it can be read under Windows.
        abs_pdb_path = os.path.abspath(pdbfile.name)
        return abs_pdb_path

    @staticmethod
    def handle_pdb_errors(response, sfamid, pdbid):
        """Handles any possible errors from a PDB response.

        :param response: A response object from requests containing a PDB file or a PDB batch.
        :param sfamid: The superfamily ID.
        :param pdbid: The pdbid(s) associated with the request.
        :raises: ValueError, RuntimeWarning
        """
        if response.status_code >= 400:  # 400+ codes indicate error status codes
            if response.status_code in [
                406,
                429,
            ]:  # 406/429 error indicates too many requests
                raise RuntimeWarning(
                    "You are requesting too many PDBs. Please wait a few minutes."
                )
            if len(response.content) == 0:
                raise ValueError(
                    "Error from server while downloading %s: Empty response with status code %i"
                    % (pdbid, response.status_code)
                )
            try:
                response_json = response.json()
            except Exception:
                raise ValueError(
                    "Invalid response from server while downloading %s: '%s'. Status code: %i"
                    % (pdbid, response.content, response.status_code)
                )

            # Extract the error message from the JSON
            msg = response_json.get("message", response_json.get("error"))
            if msg:
                if msg == "Forbidden":
                    raise RuntimeWarning(
                        "You are not permitted access to this 3DM system."
                    )

                if "Transformation matrix" in msg:
                    raise RuntimeWarning(
                        (
                            "Transformation matrix not available for structure %s. "
                            "Please contact support@bio-prodict.nl"
                        )
                        % pdbid
                    )

                raise ValueError(
                    "Error from server while downloading %s: '%s' with status code %i"
                    % (pdbid, msg, response.status_code)
                )

            raise ValueError(
                "Could not load structure '%s:%s'. Status code %s. Server reply: %s. %s"
                % (
                    sfamid,
                    pdbid,
                    response.status_code,
                    response.content,
                    response.request.url,
                )
            )

    @staticmethod
    def _clean_object_name(obj, objects=None):
        """Cleans the name of an object (pdbid) to make sure we don't ignore 3DM sourced files.

        :param obj: The name of an object, likely a PDB ID.
        :param objects: Optional, a list of objects in the scene.
        :returns: An object name that matches context of the scene.
        """
        if objects is None or obj in objects:
            return obj

        for suffix in ["chain", "_prot"]:
            if obj + suffix in objects:
                return obj + suffix

    @staticmethod
    def is_valid_pdb(pdb):
        """Checks if the contents of a PDB file are valid, by finding elements that are required
        for a PDB to be loaded properly.

        :param pdb: The contents of a PDB file (string)
        :returns: Boolean representing whether the file contents contain the necessary elements to
                  be loaded by a viewer.
        """
        return "HETATM" in pdb or "ATOM" in pdb

    @staticmethod
    def convert_residues_to_selections(
        viewer_instance, residues, properties=None, objects=None
    ):
        """This method renumbers a list of residues to a list of selections, with the proper
        renumbering strategy for each object in the scene.

        :param viewer_instance: An instance of the calling Viewer class.
        :param residues: A list of residue numbers in alignment numbering.
        :param properties: Optional, a list of properties that will be returned as a list of
                           property values in an order matching the residues.
        :param objects: A list of all the long names of objects in the scene.
        :returns: A list of Selection objects, a list of renumbered residue numbers and (if
                  properties was set) a list of property values."""
        renumbered_residues = []
        selectors = []
        new_properties = []
        api = viewer_instance._get_api("persistence3dm")
        data_api = viewer_instance._get_api("persistence3dmdata")

        for obj, target_numbering in viewer_instance.storage.numbering_schemes.items():
            obj = CompositeUtilities._clean_object_name(str(obj), objects)

            if obj is None or not viewer_instance.object_in_scene(obj):
                continue  # If the object was not specified, even with suffix, continue

            # We need to make sure that the labeling happens with the right positions.
            if target_numbering in [1, "1", "align"]:
                # We're good to go.
                renumber_method = str
            else:
                # Renumbering for models
                accession, pdb = CompositeUtilities.split_structure_accession(obj)

                num_scheme = NumberingScheme(
                    api,
                    data_api,
                    pdb,
                    CompositeUtilities.get_familyid(viewer_instance),
                    CompositeUtilities.get_sfamid(viewer_instance),
                    accession=accession,
                )

                if target_numbering in num_scheme.custom_names:
                    # Handle custom conversion schemes
                    def renumber_method(num):
                        """Converts a number to the target custom numbering."""
                        return num_scheme.align_to_custom(num, target_numbering)

                else:
                    # Handle normal conversion schemes
                    renumber_method_name = "align_to_%s" % (target_numbering)
                    renumber_method = getattr(num_scheme, renumber_method_name)

            renumbered_results = CompositeUtilities._renumber_residues(
                residues, renumber_method, obj, properties
            )
            renumbered_residues += renumbered_results[0]
            selectors += renumbered_results[1]
            new_properties += renumbered_results[2]

        if properties is None:
            return selectors, renumbered_residues
        else:
            return selectors, renumbered_residues, new_properties

    @staticmethod
    def color_residues_by_property(
        viewer_instance, selectors, color, prop, loop=False, start_color=None
    ):
        """Colors residues using a color scale according to a property.

        :param viewer_instance: An instance of the viewer class where the coloring will take place.
        :param selectors: A list of Selection instances.
        :param color: The color you would like to use for the color scale. The scale will go from
                      grey to this fully saturated color.
        :param prop: A list of properties for each residue in `residues`.
        :param loop: If this parameter is specified, the color scale will simply loop around the
                     color wheel, starting at `color`.
        :param start_color: (Optional) defines a custom (non grey) starting color.
        :returns: Returns the color scale used to color the residues.
        """

        if len(prop) == 0:
            raise ValueError(
                "You did not pass any properties to this function. Could not create colors without"
                " a list of properties."
            )

        if start_color is not None:
            color_scale = Colors().create_scale(
                start_color, color, min(prop), max(prop)
            )
        elif loop:
            color_scale = Colors().create_scale(
                color, color, min(prop), max(prop), loop=True
            )
        else:
            color_scale = Colors().create_scale("grey", color, min(prop), max(prop))

        for sel, selector in enumerate(selectors):
            color = color_scale(prop[sel], with_hash=False)
            viewer_instance._color_residue(selector, color)

        return color_scale

    @staticmethod
    def renumber_with_underflow_guard(
        viewer_instance, stable_selection, res_num, extension=None
    ):
        """Renumbers a selection outside the core to a negative number, taking into account
        the limitations on negative numbers (cannot be <-999) in YASARA.

        :param viewer_instance: Instance of YASARA or PyMol viewer.
        :param stable_selection: Stable atom selection
        :param res_num: Target residue number (int)
        :returns: Tuple with the new negative number and extension
        """
        if extension is None:
            extensions = [None, "B", "C", "D", "E", "F"]
        else:
            extensions = [extension, "W", "Y", "Z"]
        extension_idx = 0

        while res_num < -999:
            res_num += 999
            extension_idx += 1
        extension = extensions[extension_idx]

        # Strip weird extensions like `''`
        if isinstance(extension, str) and not extension.strip("' \n\""):
            extension = None

        viewer_instance.renumber_residue(stable_selection, res_num, extension)
        return (res_num, extension)

    @staticmethod
    def renumber_pdb(
        viewer_instance, pdbid, sfamid, family, target_numbering="align", accession=None
    ):
        """Renumbers all the numbers in a PDB. This is a vital but complex method that allows for
        safer renumbering. Instead of straight up replacing the numbers in each object with their
        corresponding number in the lookup table, residues are isolated into groups and Renumbered
        separately, to ensure that there can be no conflict between numbering schemes at runtime.

        :param pdbid: The PDB ID of the protein that needs renumbering.
        :param sfamid: The SFAMID of the protein that needs renumbering.
        :param family: The current family that is being used. (integer)
        :param target_numbering: The target numbering scheme name or identifying integer.
        :param accession: (Optional) if this is provided, a model renumbering will take place.
                          The PDB ID will be the *template* in this case.
        :returns: The nice name of the numbering scheme.
        """

        # Make sure that the correct structure name is used for selections
        if accession is not None:
            obj = ObjectIdentifier("_".join([accession, pdbid]))
        else:
            obj = ObjectIdentifier(pdbid)

        if "_" in pdbid:
            pdbid, ext = pdbid.split("_", 1)
            # Check if this is a full protein so that renumbering can take place on the right
            # chain (molecule).
            if ext == "full":
                obj.chain = pdbid[-1]

        api = viewer_instance._get_api("persistence3dm")
        data_api = viewer_instance._get_api("persistence3dmdata")
        num_scheme = NumberingScheme(
            api, data_api, pdbid, family, sfamid, accession=accession
        )

        if target_numbering == "res":
            viewer_instance.renumber_object_to_resnum(obj)
            # The numbering type must be remembered
            try:
                viewer_instance.storage.numbering_schemes[obj.name] = "res"
            except AttributeError:
                viewer_instance.storage.numbering_schemes = {obj.name: "res"}
            return  # We're done, residue numbering is just numbering from 1-N.

        try:
            renumber_method = CompositeUtilities.determine_numbering_method(
                num_scheme, target_numbering
            )
        except AttributeError as err:
            if accession is not None:
                raise RuntimeError(
                    "You selected a numbering scheme (%s) that was not available for object (%s). "
                    "Please note that models cannot be renumbered to custom numbering schemes."
                    % (target_numbering, obj)
                )
            else:
                raise err

        # Renumber to negative residue numbers, this avoids situations where residues are merged
        # and deleted due to consecutive numbering.
        if obj.chain:
            # Renumber everything to negative first.
            viewer_instance.renumber_object_to_resnum(
                ObjectIdentifier(obj.name), negative=True
            )
        viewer_instance.renumber_object_to_resnum(obj, negative=True)

        renumbering_index = num_scheme.numbering_scheme_res_index
        stable_selections = viewer_instance.get_stable_numbering_selections(obj)

        for i, (stable_selection, res_num) in enumerate(
            zip(stable_selections, range(1, len(renumbering_index) + 1))
        ):
            try:
                new_number = renumber_method(res_num)
            except KeyError:
                # Can't renumber residues that cannot be found (outside the aligned cores.)
                CompositeUtilities.renumber_with_underflow_guard(
                    viewer_instance, stable_selection, -res_num
                )
                continue

            extension = None
            try:
                int(new_number)  # Check if this is a normal residue number.
                if new_number < 0:  # Variable regions
                    new_number, extension = abs(new_number), "V"
            except (TypeError, ValueError):
                if new_number is None or len(new_number.strip()) == 0:
                    new_number = -i
                elif new_number == "x":
                    # Apply gap numbering
                    new_number, extension = 0, "X"
                elif "." in new_number:
                    # GPCR Numbering
                    new_number, extension = CompositeUtilities.convert_gpcr_number(
                        new_number
                    )
                else:
                    # Normal string type numbering
                    try:
                        new_number, extension = re.findall(
                            r"(\d+)([a-z]+)?", str(new_number)
                        )[0]
                    except IndexError:
                        raise RuntimeError(
                            "Could not renumber %i to %s, new number has unconventional"
                            " formatting." % (res_num, new_number)
                        )
                    extension = extension.upper()

            new_number = int(new_number)
            CompositeUtilities.renumber_with_underflow_guard(
                viewer_instance, stable_selection, new_number, extension
            )

        try:
            viewer_instance.storage.numbering_schemes[obj.name] = target_numbering
        except AttributeError:
            viewer_instance.storage.numbering_schemes = {obj.name: target_numbering}

        return num_scheme.numbering_scheme_nice_names[target_numbering]

    @staticmethod
    def convert_gpcr_number(number):
        """Converts a GPCR number to a format that can be handled by YASARA. For example,

            x.26 ~> 26X
            4.61b ~> 461B
            2.39f ~> 239F

        Returns this numbering as a tuple of the number (integer) and the extension (string).
        This will be removed when Elmar has implemented handling of GPCR numberings.
        """
        if number[0] == "x":
            ext, num = number.split(".")
            return int(num), ext.upper()

        num, ext = re.findall(r"(\d+)([a-z]+)?", number.replace(".", ""))[0]
        return int(num), ext.upper()

    @staticmethod
    def split_structure_accession(obj):
        """Splits an object name into an accession number and a structure id if the name matches
        the model pattern."""
        if "_" in obj and obj.rsplit("_", 1)[1] not in ["prot", "drug", "full"]:
            accession, pdb = obj.rsplit("_", 1)
        else:
            accession, pdb = None, obj
        return accession, pdb

    @staticmethod
    def filter_position_data(data, prop, maxvalue, cutoff):
        """Shows the user a dialog that allows them to filter the results when showing superfamily
        data. The filtering is performed using an amount and a cutoff. If `maxvalue` or `cutoff`
        are specified the window will not be shown.

        The window shown will be adjusted to the type of property that is being loaded.

        :param data: The positionInfo data
        :param prop: The positionInfo property you would like to access
        :param maxvalue: The maximum amount of residues you would like to show.
        :param cutoff: The cutoff for the property value.
        :returns: A dict containing the position numbers as keys and the value for the specified
                  property as values.
        """

        position_data = {}
        print(data)
        for pos in data:
            try:
                if float(pos[prop]) >= float(cutoff):
                    position_data[pos["alignmentPosition"]] = float(pos[prop])
            except TypeError:
                continue

        position_data = sorted(
            position_data.items(), key=lambda x: (x[1], [0]), reverse=True
        )
        if len(position_data) == 0:
            raise RuntimeWarning("Didn't find any residues matching your requirements.")

        if maxvalue is not None:
            return dict(position_data[:maxvalue])
        else:
            return dict(position_data)

    @staticmethod
    def get_literature_hotspot_data(viewer_instance, keywords, cutoff=10, limit=None):
        """Returns literature hotspot data based on keywords and a cutoff on the minimum amount of
        articles that are result from the search.

        :param keywords: The search keywords as a list of tokens.
        :param cutoff: The minimum amount of results to be used as a filter. Defaults to 10.
        :param limit: Optional, the maximum amount of results to return. Results are sorted before
                      limiting.
        :returns: A list of positions and corresponding amount of mutations, and a list of articles.
        """
        api = viewer_instance._get_api("persistence3dm")

        response = api.literatureHotspots(
            suffix=[
                str(CompositeUtilities.get_familyid(viewer_instance)),
                str(CompositeUtilities.get_subset_id(viewer_instance)),
            ],
            searchterms=keywords,
            sfamid_header=CompositeUtilities.get_sfamid(viewer_instance),
        )
        data = response.json()
        articles = data["literatureHotspotsArticlesList"]

        # Flatten nested API response into list.
        articles_per_position = []
        for article in articles:
            for mutation in article["literatureHotspotMutationList"]:
                articles_per_position.append(
                    {
                        "pubmed_id": article["identifier"],
                        "title": article["articletitle"],
                        "mutation": mutation["mutant"],
                        "organism_id": mutation["identifier"],
                        "position": mutation["alignmentPosition"],
                        "url": "https://www.ncbi.nlm.nih.gov/pubmed/"
                        + article["identifier"],
                    }
                )

        articles_per_position.sort(key=lambda art: art["position"])

        mutations_index = sorted(
            data["literatureHotspotsPositions"].items(),
            key=lambda item: item[1],
            reverse=True,
        )

        mutations_index = [
            (int(pos), mut_count)
            for pos, mut_count in mutations_index
            if mut_count >= cutoff
        ]

        if limit is not None and isinstance(limit, int):
            mutations_index = mutations_index[:limit]

        # Filter articles so that only those corresponding to shown positions are included.
        articles_per_position = [
            art
            for art in articles_per_position
            if art["position"] in dict(mutations_index).keys()
        ]

        return mutations_index, articles_per_position

    @staticmethod
    def get_custom_numbering_schemes(viewer_instance):
        """Returns a list of numbering schemes that are specific for the current family and
        superfamily."""
        return NumberingScheme.list_numbering_schemes_for_sfamid(
            viewer_instance._get_api("persistence3dmdata"),
            CompositeUtilities.get_familyid(viewer_instance),
            CompositeUtilities.get_sfamid(viewer_instance),
        )

    @staticmethod
    def get_user_groups(viewer_instance):
        """Returns the group IDs that the currently logged in user belongs to."""
        access_api = viewer_instance._get_api("access")
        response = access_api.public(suffix=["groupmemberships"])
        try:
            groups = response.json()
        except ValueError:
            raise RuntimeError("Unable to parse API response: '%s'" % response.text)
        group_ids = [group["groupid"] for group in groups]
        return group_ids

    # Storage retrieval methods

    @staticmethod
    def get_sfamid(viewer_instance):
        """Returns the superfamily ID that has been selected by the user, or raises a RuntimeError
        if none can be found."""
        try:
            return viewer_instance.storage.system
        except AttributeError:
            # No system selected by the user, check the objects in the scene for clues.
            if viewer_instance.check_scene_objects():
                return viewer_instance.storage.system
        raise RuntimeWarning("Please load an object from 3DM to continue.")

    @staticmethod
    def determine_systems_db(system_name):
        """Returns a system database based on a system id/SFAMID. Defaults to postgres-3dm-systems."""
        if system_name.startswith("pdb_"):
            return "pdb_clusters_v2"
        if system_name.startswith("scop_v2"):
            return "scop_v2"

        return "postgres-3dm-systems"

    @staticmethod
    def get_nice_name_for_system_id(viewer_instance, system_id):
        """Returns the 'nice' name of a system based on a SFAMID/system id, if no nice name exists,
        return the original name."""
        access_api = viewer_instance._get_api("access")
        systems_db = CompositeUtilities.determine_systems_db(system_id)

        systems = CompositeUtilities.load_accessible_systems(
            access_api, systems_db=systems_db
        )
        systems = {v: k for k, v in systems.items()}  # Invert dictionary
        return systems.get(system_id, system_id)

    @staticmethod
    def get_nice_name_for_num_scheme_id(viewer_instance, num_scheme_id):
        """Returns the 'nice' name of a system based on a numbering scheme id, if no nice name
        exists, return the original name."""

        data_api = viewer_instance._get_api("persistence3dmdata")

        (
            _,
            numbering_scheme_nice_names,
        ) = NumberingScheme.retrieve_custom_numbering_schemes(
            data_api,
            CompositeUtilities.get_familyid(viewer_instance),
            CompositeUtilities.get_sfamid(viewer_instance),
        )

        return numbering_scheme_nice_names.get(num_scheme_id, num_scheme_id)

    @staticmethod
    def get_sfamid_nice_name(viewer_instance):
        """Returns the 'proper' name for the superfamily that is currently selected. This is for
        display purposes only."""
        try:
            return viewer_instance.storage.system_nice_name
        except AttributeError:
            if viewer_instance.check_scene_objects():
                nice_name = CompositeUtilities.get_nice_name_for_system_id(
                    viewer_instance, viewer_instance.storage.system
                )
                viewer_instance.storage.system_nice_name = nice_name
                return nice_name

        raise RuntimeWarning("Please load an object from 3DM to continue.")

    @staticmethod
    def get_familyid(viewer_instance):
        """Returns the family ID that has been selected by the user, or raises a RuntimeError if
        none can be found."""
        try:
            stored_family_id = viewer_instance.storage.family
            return stored_family_id
        except AttributeError:
            # No system selected by the user, check the objects in the scene for clues.
            print(
                "# No system selected by the user, check the objects in the scene for clues."
            )
            if viewer_instance.check_scene_objects():
                return viewer_instance.storage.family
        raise RuntimeWarning("Please load an object from 3DM to continue.")

    @staticmethod
    def get_family_nice_name(viewer_instance):
        """Returns the 'proper' name for the family/alignment that is currently selected. This is
        for display purposes only."""
        try:
            return viewer_instance.storage.family_nice_name
        except AttributeError:
            if viewer_instance.check_scene_objects():
                families = CompositeUtilities.load_alignment_families(viewer_instance)
                families = {v: k for k, v in families.items()}  # Invert dictionary

                try:
                    family_nice_name = families[viewer_instance.storage.family]
                except KeyError:
                    family_nice_name = "%s (no name)" % viewer_instance.storage.family

                viewer_instance.storage.family_nice_name = family_nice_name

                return family_nice_name

        raise RuntimeWarning("Please load an object from 3DM to continue.")

    @staticmethod
    def get_subset_id(viewer_instance):
        """Returns the subset ID that has been selected by the user, or raises a RuntimeError if
        none can be found."""
        try:
            return viewer_instance.storage.subset
        except AttributeError:
            # No system selected by the user, check the objects in the scene for clues.
            if viewer_instance.check_scene_objects():
                return int(viewer_instance.storage.subset)
        raise RuntimeWarning("Please load an object from 3DM to continue.")
