""" This module serves to provide methods that allows the plugin to send automated error messages
to the Bio-Prodict bug-reporting services."""
# pylint:disable=E1101,W0212,R0913
import sys
import time
import json
import pprint
import traceback
from ...shared.vendors import requests
from ..api.APIAuth import APIAuth


class ErrorHandling(object):
    """This class serves to provide methods that allows the plugin to send automated error messages
    to the Bio-Prodict bug-reporting services."""

    @staticmethod
    def compile_error_artifacts(viewer_instance):
        """Find information relevant to the error and compile it."""
        full_traceback = "\n".join(traceback.format_exception(*sys.exc_info()))
        try:
            # PyMol handling
            full_storage = viewer_instance.storage.as_json()
        except AttributeError:
            # YASARA handling
            full_storage = pprint.pformat(viewer_instance.storage.__dict__)
        return {"traceback": full_traceback, "storage": full_storage}

    @staticmethod
    def send_feedback(
        title, message, assignee, project, issue_type, user_details, send=False
    ):
        """Reports a bug using the feedback endpoint for Jira.

        :param title: The title/subject for the email
        :param message: The plaintext body message for the email.
        :param assignee: Jira username (bvroling, dbrouwer, jlange, rkuipers, tvdbergh,
                                        sheijl, smaduro)
        :param project: Jira Project Key (AER, INT, MISC, SUPP, TDMA)
        :param issue_type: (BUG, EPIC, IMPROVEMENT, NEW_FEATURE, STORY, TASK)
        :param access_token: A logged in APIAuth object.
        :param send: Whether to actually send the message or not, defaults to False
        """
        data = {
            "assignee": assignee,
            "title": title,
            "body": message,
            "project": project,
            "issueType": issue_type,
        }

        headers = user_details.get_request_headers()
        headers["Content-Type"] = "application/json; charset=utf-8"

        if send:
            response = requests.post(
                "https://api.bio-prodict.nl/access/feedback",
                data=json.dumps(data),
                headers=headers,
            ).content
            print(str(response))

            assert "exception" not in response
        return data

    @staticmethod
    def create_message(user_details, viewer, version, _traceback, _storage):
        """Creates a body message for the email.

        :param user_details: A logged in APIAuth object.
        :param viewer: The name of the viewer being user (YASARA/PyMol...)
        :param version: The version of the plugin being used.
        :param traceback: Traceback as a string
        :param storage: Storage dictionary.
        """
        msg = """On {datetime} user {user} encountered an error in the {viewer} plugin ({version})

    {traceback}

"""
        return msg.format(
            datetime=time.strftime("%b %d %Y %H:%M:%S", time.localtime()),
            user=user_details._APIAuth__username,
            viewer=viewer,
            traceback=_traceback,
            storage=json.dumps(_storage, indent=True),
            version=version,
        )

    @staticmethod
    def create_title(exception, viewer, user_details):
        """Creates the title/subject line for the email."""
        return "Molviewer Plugin: Exception {exception} for user {user} in {viewer} plugin".format(
            exception=exception, user=user_details._APIAuth__username, viewer=viewer
        )

    @staticmethod
    def should_report(exception):
        """Determines if the exception should be reported or if it should be returned to the user.

        :param exception: Exception object.
        """
        return not isinstance(exception, RuntimeWarning)

    @staticmethod
    def handle(exception, viewer_instance, viewer, send=False):
        """Tries to handle an exception. Returns the error if this shouldn't be handled by this
        class.

        :param exception: The exception that needs to be handled.
        :param viewer_instance: A Viewer instance
        :param viewer: String, the viewer in use (YASARA/PyMol)
        :param send: Whether to actually send the message or not, defaults to False
        :returns: Returns the original exception if it will not be handled by ErrorHandling.
                  If ErrorHandling will process the error, a boolean will be returned: True if an
                  error report has been sent, False if not.
        """

        if not ErrorHandling.should_report(exception):
            return exception

        artifacts = ErrorHandling.compile_error_artifacts(viewer_instance)

        if viewer_instance._check_user_access():
            user_details = viewer_instance.storage.api_auth
        else:
            user_details = APIAuth(username="Anonymous")

        message = ErrorHandling.create_message(
            user_details,
            viewer,
            viewer_instance.get_version(),
            artifacts["traceback"],
            artifacts["storage"],
        )

        title = ErrorHandling.create_title(exception, viewer, user_details)

        if not viewer_instance.show_error_feedback_dialog(artifacts["traceback"]):
            # Do not send an email, but print the traceback, this is valuable during development.
            print("Report :" + artifacts["traceback"])
            return False

        ErrorHandling.send_feedback(
            title, message, "sheijl", "TDMA", "BUG", user_details, send
        )
        return True
