""" This module allows you to create a selection in a platform agnostic way. """
# pylint: disable=R0902
try:
    from collections import Iterable
except ImportError:
    from collections.abc import Iterable
from ...shared.vendors.six import string_types
from ...shared.vendors import six


class ObjectIdentifier(object):
    """Identifier class for an object in a viewer. This is used to couple an object name
    with a molecule or chain identifier."""

    def __init__(self, object_name, chain=None):
        self.name = object_name
        self.chain = chain

    def __eq__(self, other):
        if isinstance(other, ObjectIdentifier):
            return self.name == other.name and self.chain == other.chain
        elif isinstance(other, (str, six.text_type)):
            return self.name == other
        raise NotImplementedError(
            "Could not check equality against type %s." % type(other)
        )

    def __repr__(self):
        return str((self.name, self.chain))

    def __str__(self):
        return self.name


class Selection(object):
    """This class creates Selection strings for different molecular viewers."""

    def __init__(self, **kwargs):
        """Creates a selector, all of the following arguments are optional. Empty Selections will
        match everything."""
        self.numbers = kwargs.get("numbers", [])
        self.in_group = kwargs.get("in_group")
        self.not_in_group = kwargs.get("not_in_group")
        self.in_object = kwargs.get("in_object")
        self.not_in_object = kwargs.get("not_in_object")
        self.in_protein = kwargs.get("in_protein")
        self.in_selected = kwargs.get("in_selected")
        self.only_ca = kwargs.get("only_ca")
        self.in_molecule = kwargs.get("in_molecule")

    @staticmethod
    def detect_range(numbers):
        """Detects ranges of numbers, so 1,2,3,4,5 ~> (1, 5). Requires a list of numbers."""
        current_ranges = []
        starting_number = None
        latest_number = None
        recording_range = False
        for num in numbers:
            num = int(num)

            if starting_number is None:
                starting_number = num
                continue

            if starting_number == num - 1 or latest_number == num - 1:
                # This may be a range
                latest_number = num
                recording_range = True
                continue

            else:
                if recording_range and latest_number - starting_number > 2:
                    current_ranges.append((starting_number, latest_number))

                starting_number = num
                recording_range = False
                latest_number = None
                continue

        if recording_range and latest_number - starting_number > 2:
            current_ranges.append((starting_number, latest_number))

        return current_ranges

    def parse_yasara_numerics(self, selector):
        """This method handles parsing the numerical portion of the selection string for yasara."""
        if isinstance(self.numbers, Iterable) and len(self.numbers) == 0:
            selector = "all " + selector
        elif isinstance(self.numbers, Iterable) and len(self.numbers) == 1:
            selector = str(self.numbers[0]) + " " + selector
        elif isinstance(self.numbers, string_types):
            selector = str(self.numbers) + " " + selector
        elif isinstance(self.numbers, Iterable):
            self.numbers = list(self.numbers)
            selector_list = []
            possible_ranges = self.detect_range(self.numbers)

            if len(possible_ranges) == 1:
                _range = possible_ranges[0]
                return ("%s-%s %s" % (_range[0], _range[1], selector)).strip()
            elif len(possible_ranges) > 1:
                for _range in possible_ranges:
                    selector_list.append(
                        ("%s-%s %s" % (_range[0], _range[1], selector)).strip()
                    )
            else:
                for num in self.numbers:
                    selector_list.append((str(num) + " " + selector).strip())
            return selector_list
        else:
            selector = str(self.numbers) + " " + selector

        return selector

    def yasara(self):
        """Outputs a valid yasara selection string based on the initial settings."""
        selector = ""

        # Objects
        if self.in_object is not None:
            selector += "in obj %s " % self.in_object

        if self.not_in_object is not None:
            selector += "not in obj %s " % self.not_in_object

        # Groups
        if self.in_group is not None:
            selector += "%s " % self.in_group

        if self.in_selected:
            selector += "selected "

        if self.not_in_group is not None:
            selector += "!%s " % self.not_in_group

        # Protein
        if self.in_protein is not None:
            if self.in_protein:
                selector += "Protein "
            else:
                selector += "!Protein "

        # Number prefix
        selector = self.parse_yasara_numerics(selector)
        if isinstance(selector, list):
            return selector
        else:
            return selector.strip()

    def parse_pymol_numerics(self, selector):
        """This method handles parsing the numerical portion of the selection string for pymol."""
        resi = None
        if isinstance(self.numbers, Iterable) and len(self.numbers) == 0:
            if len(selector) == 0:
                selector = "all"
            else:
                selector = selector
        elif isinstance(self.numbers, Iterable) and len(self.numbers) == 1:
            resi = "resi %s" % (self.numbers[0])
        elif isinstance(self.numbers, string_types):
            resi = "resi %s" % (self.numbers)
        elif isinstance(self.numbers, Iterable):
            self.numbers = list(self.numbers)
            possible_ranges = self.detect_range(self.numbers)
            if len(possible_ranges) == 1:
                _range = possible_ranges[0]
                resi = ("resi %s-%s" % (_range[0], _range[1])).strip()
            elif len(possible_ranges) > 1:
                nums = ", ".join(["%s-%s" % _range for _range in possible_ranges])
                resi = ("resi %s" % (nums)).strip()
            else:
                nums = ", ".join([str(_range) for _range in self.numbers])
                resi = ("resi %s" % (nums)).strip()
        else:
            resi = "resi %s" % (self.numbers)

        if selector == "":
            selector = resi
        elif resi is not None:
            selector = " and ".join([selector, resi])
        return selector

    def pymol(self):
        """Outputs a valid pymol selection string based on the initial settings."""
        selector = []

        # Objects
        if self.in_object is not None:
            selector.append("object %s" % self.in_object)

        if self.not_in_object is not None:
            selector.append("not object %s" % self.not_in_object)

        # Groups
        if self.in_group is not None:
            selector.append(self.in_group)

        if self.in_selected:
            selector.append("sele")

        if self.not_in_group is not None:
            selector.append("not %s" % self.not_in_group)

        # C-Alpha
        if self.only_ca is not None and self.only_ca:
            selector.append("name ca")

        # Protein
        if self.in_protein is not None:
            if self.in_protein:
                selector.append("polymer")
            else:
                selector.append("not polymer")

        # Number prefix
        selector_str = " and ".join(selector)
        selector_str = self.parse_pymol_numerics(selector_str)
        return "(%s)" % selector_str.strip()
