"""
Module provides classes related to 3DM that are not concerned with the JSON API.
"""
# pylint:disable=R0903,R0913,E1101
try:
    from urllib import quote_plus
except ImportError:
    from urllib.parse import quote_plus
from ..Config import Config


class ThreeDMWebsite(object):
    """This class provides methods for interaction with the 3DM website that do not belong to the
    JSON API."""

    def __init__(self):
        self.base_url = Config().website_base_url

    def create_3dm_residue_link(
        self, position, sfamid="NR_demo_2008", familyid=1, subsetid=1, numschemeid=-1
    ):
        """Creates a link to a alignment position in a 3DM system.

        :param position: The alignment position.
        :param sfamid: The Superfamily ID you'd like to refer to.
        :param familyid: The ID of the family you are referring to. This will default to 1.
        :param subsetid: The ID of the subset you are referring to. This will default to 1.
        :param numschemeid: The numbering scheme ID. Defaults to -1.
        :returns: String, a URL to a alignment position in a 3DM system.
        """
        res_base = "system/{sfamid}/alignment/{familyid}/subset/{subsetid}" \
                   "/numberingscheme/{numschemeid}/alpos/{position}/distribution"

        url = "/".join([self.base_url, res_base])

        formatted_url = url.format(
            sfamid=sfamid,
            familyid=familyid,
            subsetid=subsetid,
            numschemeid=numschemeid,
            position=position,
        )

        return formatted_url

    def create_3dm_literature_link(
        self,
        position,
        keywords,
        sfamid="NR_demo_2008",
        familyid=1,
        subsetid=1,
        numschemeid=-1,
    ):
        """Creates a link to literature details on a position in a 3DM system.

        :param position: The alignment position.
        :param keywords: A list of keywords for searching.
        :param sfamid: The Superfamily ID you'd like to refer to.
        :param familyid: The ID of the family you are referring to. This will default to 1.
        :param subsetid: The ID of the subset you are referring to. This will default to 1.
        :param numschemeid: The numbering scheme ID. Defaults to -1.
        :returns: String, a URL to a alignment position in a 3DM system.
        """
        lit_base = "system/{sfamid}/alignment/{familyid}/subset/{subsetid}/" \
                   "numberingscheme/{numschemeid}/alpos/{position}/literature?keyword={keywords}"

        url = "/".join([self.base_url, lit_base])

        formatted_keywords = "%20AND%20".join([quote_plus(kw) for kw in keywords])

        formatted_url = url.format(
            sfamid=sfamid,
            familyid=familyid,
            subsetid=subsetid,
            numschemeid=numschemeid,
            position=position,
            keywords=formatted_keywords,
        )

        return formatted_url

    @staticmethod
    def get_login_scene_url(client):
        """Returns the URL for the login scene retrieval page, with the client ID included."""
        _type = {"yasara": "loginscene", "pymol": "loginsession"}[client]
        return Config().login_scene_url.format(client=client, type=_type)
