""" This module includes a class for managing plugin updates. """
import os
import tempfile
from .Proxies import Proxies
from ...shared.Config import Config
from ...shared.vendors import requests


class Updates(object):
    """The updates class contains methods that allow viewers to perform version checks."""

    @staticmethod
    def tupelize_version(version):
        """Takes a version string and returns it as a proper tuple with integers."""
        if version is None:
            return None
        return tuple([int(i) for i in version.split(".")])

    @staticmethod
    def new_version_available(version, client, check_updates_force=False):
        """Checks if a new version of the plugin is available. Checks if users have turned off
        the checking feature in the configuration and if the update has already been nagged about.

        :param version: The version to check against (probably the current version.)
        :param client: The name of the client.
        :param check_updates_force: Force checking updates
        :returns: False if no update is available or if no notification should be shown. The version
                  number otherwise as a string.
        """
        if "." in version:
            version = Updates.tupelize_version(version)

        if not check_updates_force:
            try:
                check_updates = Config().check_updates
            except AttributeError:
                check_updates = True
            if not check_updates:
                return False

        latest_stable = Updates.tupelize_version(
            Updates.get_latest_stable_version(client)
        )
        if version >= latest_stable:
            return False

        latest_notified = Updates.tupelize_version(
            Updates.check_notified_version(client)
        )
        if latest_notified is not None and latest_notified >= latest_stable:
            return False

        latest_string = ".".join([str(i) for i in latest_stable])
        Updates.store_notified_version(latest_string, client)

        return latest_string

    @staticmethod
    def _get_temp_path(client):
        """Returns the path where notification states should be stored."""
        tmp = tempfile.gettempdir()
        filename = "%s-latest-version-update-notified.txt" % client
        return os.path.join(tmp, filename)

    @staticmethod
    def store_notified_version(version, client):
        """Stores the version number in a temporary file, intended as a method of checking which
        version was the last one where a user was nagged with an update message.

        :param version: The version number to store.
        :param client: The name of the client where an update is available."""
        with open(Updates._get_temp_path(client), "w+") as _file:
            _file.write(str(version))

    @staticmethod
    def check_notified_version(client):
        """Returns the version where the user was last notified about an update. If the user was
        not notified of any updates before, return None.

        :param client: The name of the client checking for updates.
        :returns: A string containing the version of the plugin where the user was last notified
                  about an update.
        """
        try:
            with open(Updates._get_temp_path(client), "r") as _file:
                notified_version = _file.read().strip()
        except IOError:
            return None
        return notified_version

    @staticmethod
    def get_latest_stable_version(client):
        """Finds out which version is the latest released for a given client.

        :param client: yasara or pymol.
        :returns: The latest released version according to our website, as a string."""

        verify_https = getattr(Config(), "verify_https", True)

        try:
            req = requests.get(
                "https://plugin.bio-prodict.nl/%s/stable/VERSION" % client,
                proxies=Proxies.get_proxies(),
                verify=verify_https,
            )
        except Exception as e:
            if "SSLError" in str(e):
                raise RuntimeWarning(
                    "Unable to retrieve version information. Your need to be connected to the internet "
                    "and have your proxy set up correctly to work with 3DM. Please consult the "
                    "3DM installation guide for more help."
                )
            else:
                raise e
        version = req.text.strip()
        return version
